Package: Dimodal
Version: 1.0.4
Date: 2026-04-23
Title: Spacing Tests for Multi-Modality
Authors@R: c(person("Greg", "Kreider", role=c("aut","cre"),
                    email="support@primachvis.com"),
             person("Melissa", "O'Neill", role=c("cph")),
             person("Primordial Machine Vision Systems, Inc.", role=c("cph"),
                    email="support@primachvis.com"))
Copyright: Primordial Machine Vision Systems, Inc.
Maintainer: Greg Kreider <support@primachvis.com>
Description: Tests for modality of data using its spacing.  The main approach
  evaluates features (peaks, flats) using a combination of parametric models
  and non-parametric tests, either after smoothing the spacing by a low-pass
  filter or by looking over larger intervals.
License: BSD_3_clause + file LICENSE
NeedsCompilation: yes
Depends: statmod
Note: This is the CRAN version of the package.  Due to limitations of
        that platform it does not support the changepoint analysis of
        the spacing.  The full definitive version of the package is
        available at
        https://www.primachvis.com/data/Dimodal_latest.tar.gz
Packaged: 2026-05-01 18:14:49 UTC; kreider
Author: Greg Kreider [aut, cre],
  Melissa O'Neill [cph],
  Primordial Machine Vision Systems, Inc. [cph]
Repository: CRAN
Date/Publication: 2026-05-01 18:50:11 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-02 01:26:56 UTC; unix
Archs: Dimodal.so.dSYM

useDynLib(Dimodal, .registration=TRUE)
importFrom(stats, filter, mad, median, optimize, pnorm, quantile, runif)
importFrom(stats, sd, symnum, uniroot)
importFrom(statmod, pinvgauss, qinvgauss)
importFrom(grDevices, adjustcolor, col2rgb, rgb)
importFrom(grDevices, hcl.colors, hcl.pals, palette.colors, palette.pals)
importFrom(graphics, abline, axis, axTicks, box, hist, layout, legend, lines)
importFrom(graphics, mtext, par, points, rect, rug, text, title)
S3method(print, Dimodal)
S3method(summary, Dimodal)
S3method(plot, Dimodal)
S3method(print, Ditest)
S3method(summary, Ditest)
S3method(print, Diopt)
S3method(print, Dipeak)
S3method(summary, Dipeak)
S3method(print, Diflat)
S3method(summary, Diflat)
S3method(print, Didata)
S3method(summary, Didata)
S3method(print, Ditrack)
S3method(plot, Ditrack)
export(Dimodal, Diopt, Diopt.local, Ditrack)
export(mark.Dipeak, mark.Diflat)
export(Dipeak.test, Dipeak.critval, Diflat.test, Diflat.critval)
export(Dinrun.test, Dirunlen.test, Diexcurht.test, Dipermht.test)
export(midquantile, runs.as.rle, select.peaks, center.diw)
export(match.features, shiftID.place)

Tests for each command in the package.

There is one top level test to run
> source('test_all.R')
The other files have .Rtest suffixes so the package check tests don't run
them separately.  The top level script sources each of these silently - we
do NOT want the individual test results to print and compare, only the
pass message at the end.  The individual test scripts are all stand-alone
and can be sourced from the shell.  They have separate branches if run
interactively from the shell or by the package check, for loading the
commands or the library as needed.  The files can be sourced from any
sister directory but not the top.

test_all.Rout.save has the expected result.

The file test_library.Rtest is for development.  It (re)loads the shared
library and source files from the console, but contains no tests.


TEST STRATEGY

Detectors
There are separate files for the semi-public top-level feature detector
commands: test_find_runs.Rtest, test_find_peaks.Rtest, test_find_flats.Rtest,
and test_find_lvlsec.Rtest.  THe test benches in general contain four types
of checks.  First, we use data available in an R distribution as the source
for interval spacing.  The expected features are based on the prototype
versions of these detectors, which differ substantially from the Dimodal
version.  Second, we construct artificial data where we can control which
features appear.  Examples include piecewise-linear waveforms for peaks and
collections of random variates with small or large variance to define flats.
Third, we define a common set of stress tests to present to each detector.
A table follows listing the tests and corresponding test cases.  Finally,
we create Break The Part (BTP) tests that mostly stress the detector's
interface, checking for bad parameters or unusual inputs.  Some of the
stress tests could also be considered BTP.

   Stress Tests
                   find_runs       find_peaks   find_flats      find_lvlsec
   const val       i TC16 r TC18   BTP13        TC16            i TC12 r TC11
     diff ramp     i TC17 r TC19   BTP19        TC17            TC14,15
   ramp            i TC32 r TC34   TC20         TC18            TC13
   sawtooth        i TC33 r TC35   TC10         TC19            TC16
   all NA          i BTP1 r BTP2   BTP17        TC20            BTP1-2
   NULL vector     BTP8            BTP18        BTP14           BTP3
   0 len vector    i BTP6 r BTP7   BTP16        BTP15           BTP4
   1 len vector    i TC31 r TC34   BTP14        BTP16           BTP5
   2 len vector
     const         i TC36 r TC39   BTP15        TC21            BTP6
     ramp          i TC37 r TC41   TC21         TC22            BTP7
   non-vector      i TC42 r TC44   TC32         TC23            TC17
   NA values
     at start      i TC17 r TC23   TC24         i TC25 r TC24   TC19
     at end        i TC19 r TC24   TC25         i TC27 r TC26   TC20
     in middle     i TC20 r TC12   TC26         i TC29 r TC28   TC21
     mixed         i TC7  r TC27   TC27         i TC31 r TC30   TC22
   NaN values
     at start             r TC28   TC28                r TC31   TC23
     at end               r TC29   TC29                r TC32   TC24
     in middle            r TC30   TC30                r TC33   TC25
     mixed                r TC14   TC31                r TC34   TC26
   Inf values             r TC14
   outliers                                     TC10-12
                                                BTP2-3,5-6
   one flat < minlen                            TC36

Significance Tests
test_sigtests.Rtest does quality checks on the feature evaluators in the
sigtests.R file.  The first group tests the models, peak height and flat
length.  The test cases combine back-and-forth conversions of a large
number of variations of the three parameters, for example generating the
probability of a feature and then converting that back to a critical value,
comparing the result to original.  A large set of test cases finds the
probability or critical value for a handful of given setups, comparing the
result to values from the prototype version; these mostly check that the
models were ported correctly.  The second group tests runs, either Dinrun.test
or Dirunlen.test.  The expected probabilities come from the prototype version.
The third group contains the excursion tests.  Finally we have a number of
BTP checks for each of the eight tests, per this table.

   BTP/Stress tests
                      pk tst/cv    ft tst/cv    nrun   runlen  perm   excur
   no value           PT1    PC1   FT1    FC1   KR1    RL1     PH1    EH1
   NA value           PT2    PC2   FT2    FC2   KR9    RL9     PH2    EH2
   NaN value          PT2    PC2   FT2    FC2   KR10   RL10    PH3    EH3
   NULL vector        PT3    PC3   FT3    FC3   KR2,3  RL2,3   PH7    EH7
   -Inf value         PT4    PC4   FT4    FC4                  PH4    EH4
   +Inf value         PT4    PC4   FT4    FC4                  PH5    EH5
   not numeric value  PT2    PC2   FT2    FC2   KR11   RL11    PH6    EH6
   NA n/stID/ht       PT7    PC7   FT7    FC7   KR4    RL4     PH8    EH8
   NaN n              PT7    PC7   FT7    FC7   KR4    RL4     PH8    EH8
   bad n              PT7    PC7   FT7    FC7   KR6    RL6     PH10   EH10
   non-integer n      PT6    PC6   FT6    FC6   KR6    RL6     PH9    EH9
   multiple n         PT6    PC6   FT6    FC6   KR6    RL6     PH9    EH8
   non-numeric n      PT7    PC7   FT7    FC7   KR6    RL6     PH10   EH10
   NA flp/endID/nperm PT9    PC9   FT9    FC9   KR4    RL4     PH11   EH11
   NaN flp            PT9    PC9   FT9    FC9   KR4    RL4     PH11   EH11
   flp <= 0           PT9    PC9   FT8    FC9   KR7    RL7     PH11   EH11
   flp real > 1.0     PT8    PC8   FT8    FC8   KR7    RL7            
   multiple flp       PT8    PC8   FT8    FC8   KR7    RL7     PH12   EH12
   flp > n            PT9    PC9   FT8    FC9   KR7    RL7            
   non-numeric flp    PT9    PC9   FT9    FC9   KR7    RL7     PH11   EH11
	 NULL stID/endID                              KR5    RL5            
   NA filter/feps     PT10   PC10  FT10   FC10  KR8    RL8            
   alt filters        PT10   PC10  FT10   FC10  KR8    RL8            
   bad filter         PT10   PC10  FT10   FC10  KR8    RL8            
   multiple filters   PT10   PC10  FT10   FC10  KR8    RL8            
   alt basedist                    FT11   FC11                        
   bad basedist                    FT11   FC11                        
   multiple basedist               FT11   FC11                        
   NA basedist                     FT11   FC11                        
   non-bool l.tail    PT11         FT12                        PH13   EH13

Top Level Functions
Each top level function, Dimodal, Diopt, and Ditrack, has a test bench in
the file test_<function>.Rtest.  For Diopt the tests focus on correct and
incorrect parameter values and the main modes of accessing and changing
options.  The Ditrack tests are not included in test_all because of the run
time; these are more stress tests for Dimodal than checks of the Ditrack
code.

Changepoints
The changepoint tests do not require any detectors be installed; they only
check the fusion of the results by creating a mock-up of library results.


CHARACTERIZATION

Files ending in .Rchar contain code used in evaluating Dimodal and its
tests.  They are sourced from the command line to characterize the analysis,
for example the stability of the excursion test.  Like the .Rtest files,
these will not be run during the R installation check.

# Manually run test_Ditrack.Rtest because of run time.

source('../tests/test_Diopt.Rtest', echo=FALSE)
source('../tests/test_Dimodal.Rtest', echo=FALSE)
source('../tests/test_find_runs.Rtest', echo=FALSE)
source('../tests/test_find_peaks.Rtest', echo=FALSE)
source('../tests/test_find_flats.Rtest', echo=FALSE)
source('../tests/test_find_lvlsec.Rtest', echo=FALSE)
source('../tests/test_sigtests.Rtest', echo=FALSE)
source('../tests/test_utility.Rtest', echo=FALSE)

# methods.R:
# S3 class methods for the data structures built in this library.
#
# c 2025-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
#   


### Dimodal objects.

# Print the contents of Dimodal object x, depending on which analyses have
# been done and the feature(s) selected.
print.Dimodal <- function(x, feature=c("peaks", "flats"), ...) {

  feature <- match.arg(feature, several.ok=TRUE)

  cat("\n")
  
  if ("data" %in% names(x)) {
    cat("Setup\n")
    print(x$data)
  }

  if (("peaks" %in% feature) && ("lp.peaks" %in% names(x))) {
    cat("Peaks in Low-Pass Spacing\n")
    cat("feature positions at middle of filter\n")
    print(x$lp.peaks)
  }
  if (("flats" %in% feature) && ("lp.flats" %in% names(x))) {
    cat("Flats in Low-Pass Spacing\n")
    cat("feature positions at middle of filter\n")
    print(x$lp.flats)
  }

  if (("peaks" %in% feature) && ("diw.peaks" %in% names(x))) {
    cat("Peaks in Interval Spacing\n")
    cat("feature positions at end of interval\n")
    print(x$diw.peaks)
  }
  if (("flats" %in% feature) && ("diw.flats" %in% names(x))) {
    cat("Flats in Interval Spacing\n")
    cat("feature positions at end of interval\n")
    print(x$diw.flats)
  }

  invisible(x)
}

# Summarize the contents of Dimodal object, depending on which analyses have
# been done and the feature(s) selected.
summary.Dimodal <- function(object, feature=c("peaks", "flats"), ...) {

  feature <- match.arg(feature, several.ok=TRUE)

  x <- object

  summary(x$data)

  if (("peaks" %in% feature) && ("lp.peaks" %in% names(x))) {
    cat("Peaks in Low-Pass Spacing\n")
    cat("feature positions at middle of filter\n")
    summary(x$lp.peaks)
  }
  if (("flats" %in% feature) && ("lp.flats" %in% names(x))) {
    cat("Flats in Low-Pass Spacing\n")
    cat("feature positions at middle of filter\n")
    summary(x$lp.flats)
  }

  if (("peaks" %in% feature) && ("diw.peaks" %in% names(x))) {
    cat("Peaks in Interval Spacing\n")
    cat("feature positions at end of interval\n")
    summary(x$diw.peaks)
  }
  if (("flats" %in% feature) && ("diw.flats" %in% names(x))) {
    cat("Flats in Interval Spacing\n")
    cat("feature positions at end of interval\n")
    summary(x$diw.flats)
  }

  invisible(object)
}

# Return a dummy object of class Dimodal, with mockups for all elements.
mockup.Dimodal <- function() {

  res <- list(data=mockup.Didata(), opt=NULL, 
              lp.peaks=mockup.Dipeak(), lp.flats=mockup.Diflat(),
              diw.peaks=mockup.Dipeak(), diw.flats=mockup.Diflat())
  class(res) <- "Dimodal"
  res
}


# Create a graph of Dimodal object x, plotting the low-pass spacing, a
# histogram of the data, and/or interval spacing per show.  Mark feature(s),
# which if 'none' or NULL or NA adds no annotations.  The features to mark in
# the histogram come from the first shown (ie. leftmost) analysis, lp or diw.
# Diopt options opt provides significance levels.  Uses layout to order the
# plots.
# Returns x invisibly.
plot.Dimodal <- function(x, show=c("lp", "histogram", "diw"),
                         feature=c("peaks", "flats"), opt=Diopt(), ...) {

  show <- unique(tolower(show))
  show <- match.arg(show, several.ok=TRUE)
  i <- match("lp", show)
  if (((1 < length(i)) || !is.na(i)) && !("lp" %in% rownames(x$data))) {
    # If we have only generated changepoints, then plot spacing i.p.v. LP.
    if (!("Diw" %in% rownames(x$data)) &&
        ("cpt" %in% feature) && ("cpt" %in% names(x))) {
      show[i] <- "cpt"
    } else {
      show <- show[-i]
    }
  }
  i <- match("diw", show)
  if (((1 < length(i)) || !is.na(i)) && !("Diw" %in% rownames(x$data))) {
    show <- show[-i]
  }
  if (0 == length(show)) {
    stop("no data to show exists in Dimodal result")
  }

  feature <- tolower(feature)
  if ((0 == length(feature)) ||
      ((1 == length(feature)) && (is.na(feature) || ("none" == feature)))) {
    feature <- NULL
  } else {
    feature <- match.arg(feature, several.ok=TRUE)
    if (("lp" %in% show) && ("diw" %in% show)) {
      feat.src <- show[min(match(c("lp", "diw"), show))]      
    } else if ("lp" %in% show) {
      feat.src <- "lp"
    } else if ("diw" %in% show) {
      feat.src <- "diw"
    } else {
      feat.src <- "none"
    }
  }
  
  if ("Diopt" != class(opt)) {
    opt <- Diopt(opt)
  }

  layout(matrix(1:length(show), nrow=1))
  # Layout for a 3 grid forces cex to 0.66.  Force that even if fewer plots.
  oldpar <- par(cex=0.66)
  on.exit(par(oldpar))

  add.decile <- "histogram" %in% show
  for (i in seq_along(show)) {
    if ("histogram" == show[i]) {
      h <- plot_Dimodal_hist(x, opt, feat.src)
      if ("peaks" %in% feature) {
        if (("lp" == feat.src) && ("lp.peaks" %in% names(x))) {
          mark.Dipeak(x$lp.peaks, h, opt)
        } else if (("diw" == feat.src) && ("diw.peaks" %in% names(x))) {
          mark.Dipeak(x$diw.peaks, h, opt)
        }
      }
      if ("flats" %in% feature) {
        if (("lp" == feat.src) && ("lp.flats" %in% names(x))) {
          mark.Diflat(x$lp.flats, h, opt)
        } else if (("diw" == feat.src) && ("diw.flats" %in% names(x))) {
          mark.Diflat(x$diw.flats, h, opt)
        }
      }
    }
  
    if ("lp" == show[i]) {
      plot_Dimodal_lp(x, opt, add.decile)
      if (("peaks" %in% feature) && ("lp.peaks" %in% names(x))) {
        mark.Dipeak(x$lp.peaks, NULL, opt)
      }
      if (("flats" %in% feature) && ("lp.flats" %in% names(x))) {
        mark.Diflat(x$lp.flats, NULL, opt)
      }
    }
  
    if ("diw" == show[i]) {
      plot_Dimodal_diw(x, opt, add.decile)
      if (("peaks" %in% feature) && ("diw.peaks" %in% names(x))) {
        mark.Dipeak(x$diw.peaks, NULL, opt)
      }
      if (("flats" %in% feature) && ("diw.flats" %in% names(x))) {
        mark.Diflat(x$diw.flats, NULL, opt)
      }
    }
  }

  invisible(x)
}

# Plot the histogram of the raw data in Dimodal class object x.  Use graphics
# options in Diopt opt.  Plot a CDF atop, with axes, and jittered data below.
# The quantile axis label comes from feat.src, either 'lp', 'diw', or 'none'.
# Returns the histogram.
plot_Dimodal_hist <- function(x, opt, feat.src) {

  qs <- quantile(x$data["x",], c(0.05, 0.5, 0.95), names=FALSE)
  dmin <- qs[2] + (3 * (qs[1] - qs[2]))
  dmax <- qs[2] + (3 * (qs[3] - qs[2]))
  dhist <- x$data["x",(dmin<=x$data["x",]) & (x$data["x",]<=dmax)]
  # Want at least 5 points per bucket, on average.
  nbreaks <- length(dhist) %/% 5

  clr <- get.colors(opt$palette)
  hclr <- adjustcolor(clr[opt$colID.hist], 1, 0.9, 0.9, 0.9)
  hbdr <- adjustcolor(hclr, 1, 1.15, 1.15, 1.15)

  h <- hist(dhist, breaks=nbreaks, col=hclr, border=hbdr,
            xlab="data", ylab="count", main="Raw Data")
  # Draw the cumulative count and an axis.
  hcum <- max(h$counts) * cumsum(h$counts) / sum(h$counts)
  # ecdf aligns to end of breaks, not to mids.
  lines(h$breaks[-1L], hcum, col=clr[opt$colID.cdf], lwd=2)
  axis(4, line=NA, at=max(h$counts)*seq(0,1,by=0.1), labels=FALSE,
       col=clr[opt$colID.cdf])
  feat.src <- switch(feat.src, lp="LP   ", diw="Diw   ", none="none")
  if ("none" != feat.src) {
    mtext(feat.src, 4, at=0, adj=1, las=0, cex=par("cex")*par("cex.axis"))
  }
  rug(jitter(dhist))

  dhist <- sort(dhist)
  # Something big that will be off the graph to the right.
  xr <- 2 * dhist[length(dhist)]

  invisible(h)
}

# Plot the raw data and low-pass spacing in Dimodal class object x.  Use
# graphic options in Diopt opt.  If add.decile is TRUE draw the decile axis
# underneath.
plot_Dimodal_lp <- function(x, opt, add.decile) {

  raw <- x$data["Di",-1L]
  stID <- attr(x$data, "lp.stID")
  endID <- attr(x$data, "lp.endID")
  lp <- x$data["lp",stID:endID]

  clr <- get.colors(opt$palette)
  clr[opt$colID.data] <- adjustcolor(clr[opt$colID.data], 1, 0.9, 0.9, 0.9)
  
  ymax <- max(1.1*max(lp), quantile(raw, 0.98, names=FALSE))
  plot(0, 0, col=NA, xlim=c(0,ncol(x$data)), ylim=c(0,ymax),
       xlab="index", ylab="spacing", main="Low-Pass Spacing")
  points(2:ncol(x$data), raw, col=clr[opt$colID.data], pch=20,
         cex=ifelse(500 < ncol(x$data), 0.1, 0.5))

  raw[raw < (1.04 * ymax)] <- NA
  raw[!is.na(raw)] <- 1.03 * ymax
  points(raw, col=clr[opt$colID.data], pch=4)

  lines(stID:endID, lp, col=clr[opt$colID.filter], lwd=2, lty=1)
  if (add.decile) {
    axis(1, line=4, at=ncol(x$data)*seq(0,1,by=0.1), labels=FALSE,
         col=clr[opt$colID.cdf])
  }
}

# Plot the raw data and interval spacing in Dimodal class object x.  Use
# graphic options in Diopt opt.  If add.decile is TRUE draw the decile axis
# underneath.
plot_Dimodal_diw <- function(x, opt, add.decile) {

  raw <- x$data["Di",-1L]
  stID <- attr(x$data, "diw.stID")
  endID <- ncol(x$data)
  diw <- x$data["Diw",stID:endID]

  clr <- get.colors(opt$palette)
  clr[opt$colID.data] <- adjustcolor(clr[opt$colID.data], 1, 0.9, 0.9, 0.9)

  ylbl <- sprintf("spacing, interval = %.0f", attr(x$data,"wdiw"))

  # For large intervals the spacing will be too small to see differences.
  # Use the range of the interval spacing to increase the spacing for plotting.
  rscl <- max(1, min(attr(x$data,"wdiw")/2, floor(max(diw) / max(1,min(diw)))))
  raw <- raw * rscl
  ymax <- 1.1 * max(diw)

  plot(0,0, col=NA, xlim=c(0,ncol(x$data)), ylim=c(0,ymax),
       xlab="index", ylab=ylbl, main="Interval Spacing")
  points(2:ncol(x$data), raw, col=clr[opt$colID.data], pch=20,
         cex=ifelse(500 < ncol(x$data), 0.1, 0.5))
  if (1.1 < rscl) {
    mtext(sprintf("spacing (dots) scaled by %.0f", rscl), 4, 0.5,
          cex=par("cex"))
  }

  raw[raw < (1.04 * ymax)] <- NA
  raw[!is.na(raw)] <- 1.03 * ymax
  points(raw, col=clr[opt$colID.data], pch=4)

  lines(stID:endID, diw, col=clr[opt$colID.filter], lwd=2, lty=1)
  if (add.decile) {
    axis(1, line=4, at=ncol(x$data)*seq(0,1,by=0.1),
         labels=FALSE, col=clr[opt$colID.cdf])
  }
}

# Return the colors for palette pal, using HCL colors if it has prefix 'hcl:'.
get.colors <- function(pal) {

  if ("hcl:" == substr(pal, 1, 4)) {
    hcl.colors(8, substring(pal, 5))
  } else {
    palette.colors(8, pal)
  }
}


### Ditest objects.

# Pretty-print the Ditest class object x.  Other arguments are ignored.
# Returns x invisibly.
print.Ditest <- function(x, ...) {

  arglist <- c(x$statname, names(x$parameter))
  is.excur <- !is.null(x$xbase) && !is.null(x$base)
  is.markov <- !is.null(x$tmat) && !is.null(x$wt)
  if (is.markov) {
    arglist <- c(arglist, "tmat", "wt")
  }

  pnames <- NULL
  fixarg <- NULL
  if (0 == length(x$parameter)) {
    p <- cbind(x$statistic, x$p.value)
  } else {
    p <- matrix(x$statistic, ncol=1)
    for (i in seq_along(x$parameter)) {
      if (nrow(p) == length(x$parameter[[i]])) {
        p <- cbind(p, x$parameter[[i]])
        pnames <- c(pnames, names(x$parameter)[i])
      } else {
        fixarg <- c(fixarg,
                    paste0(names(x$parameter)[[i]], "=", format(x$parameter[[i]])))
      }
    }
    p <- cbind(p, x$p.value)
  }
  colnames(p) <- c(x$statname, pnames, "p.value")
  # The result may contain invalid entries.  Do not print those.
  p <- p[is.finite(p[,1]) & is.finite(p[,ncol(p)]), ,drop=FALSE]

  cat("\n")
  cat(strwrap(x$method, prefix="  "), sep="\n")
  cat("  evaluated with ", x$statfn, "(",
      paste0(arglist, collapse=", "), ")\n", sep="")
  # This prevents R from printing [1,] before the row.
  rownames(p) <- rep("  ", nrow(p))
  print(p, print.gap=3, max=ncol(p)*nrow(p))
  if (!is.null(fixarg)) {
    cat("    with ", paste0(fixarg, collapse=", "), "\n", sep="")
  }
  if (!is.null(x$alternative)) {
    cat("    for features ", x$alternative, " than or equal to ",
        x$statname, "\n", sep="")
  }
  if (is.markov) {
    cat("  Markov chain variables tmat, wt stored with result\n")
  }
  if (is.excur) {
    if ("signed" == x$base) {
      cat("  signed difference basis for draws stored with result\n")
    } else {
      cat("  raw data basis for heights stored with result\n")
    }
  }
  if (!is.null(x$model)) {
    params <- sapply(seq_along(x$model),
                     function(i) {
                       paste0(names(x$model)[i], "=", format(x$model[i])) })
    cat("  model parameters ", paste0(params, collapse=", "), "\n", sep="")
  }
  cat("\n")

  invisible(x)
}

# Print the test statistic and p values for Ditest class object.
# Other arguments are ignored.  Returns object invisibly.
summary.Ditest <- function(object, ...) {

  p <- cbind(object$statistic, object$p.value)
  colnames(p) <- c(object$statname, "p.value")
  rownames(p) <- rep("  ", nrow(p))
  cat("\n  ", object$method, "\n", sep="")
  print(t(p))
  cat("\n")

  invisible(object)
}


### Diopt objects.

# Pretty-print Diopt class object x.  Returns x invisibly.
print.Diopt <- function(x, ...) {

  wtag <- max(nchar(sub("(.+?)\\.", "", names(diopt.dflt)))) + 2
  wval <- wtag + 10

  cat("\nSpacing to Analyze:\n")
  cat(sprintf("  %s\n", printopt(x, "analysis", wtag)))
  cat("\nGeneral Data Preparation (data.*):\n")
  cat(sprintf("  %-*s\n", wval, printopt(x, "data.midq", wtag)))
  cat("Low-Pass Filter Options (lp.*):\n")
  cat(sprintf("  %-*s    %-*s\n",
              wval, printopt(x, "lp.kernel", wtag),
              wval, printopt(x, "lp.window", wtag)))
  cat(sprintf("  %s\n", printopt(x, "lp.tests", wtag)))
  cat(sprintf("  %s\n", printopt(x, "lp.param", wtag)))
  
  cat("Interval Spacing Options (diw.*):\n")
  cat(sprintf("  %-*s\n", wval, printopt(x, "diw.window", wtag)))
  cat(sprintf("  %s\n", printopt(x, "diw.tests", wtag)))
  cat(sprintf("  %s\n", printopt(x, "diw.param", wtag)))
  
  cat("Peak Detector Options (peak.*):\n")
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "peak.fht", wtag),
              wval, printopt(x, "peak.frelht", wtag),
              wval, printopt(x, "peak.fhtie", wtag)))
  cat(sprintf("  %-*s\n",
              wval, printopt(x, "peak.fhsupp", wtag)))
  
  cat("Flat Detector Options (flat.*):\n")
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "flat.fripple", wtag),
              wval, printopt(x, "flat.minlen", wtag),
              wval, printopt(x, "flat.fminlen", wtag)))
  cat(sprintf("  %-*s    %-*s\n",
              wval, printopt(x, "flat.noutlier", wtag),
              wval, printopt(x, "flat.distrib", wtag)))
  
  cat("Excursion Test Options (excur.*):\n")
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "excur.nrep", wtag),
              wval, printopt(x, "excur.ntop", wtag),
              wval, printopt(x, "excur.seed", wtag) ))
  
  cat("Permutation (runht) Test Options (perm.*):\n")
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "perm.nrep", wtag),
              wval, "",
              wval, printopt(x, "perm.seed", wtag) ))
  
  cat("Significance Thresholds (alpha.*):\n")
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "alpha.ht", wtag),
              wval, printopt(x, "alpha.pkexcur.lp", wtag),
              wval, printopt(x, "alpha.pkexcur.diw", wtag)))
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "alpha.len", wtag),
              wval, printopt(x, "alpha.ftexcur.lp", wtag),
              wval, printopt(x, "alpha.ftexcur.diw", wtag)))
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "alpha.runht", wtag),
              wval, printopt(x, "alpha.nrun", wtag),
              wval, printopt(x, "alpha.runlen", wtag)))
  
  cat("Tracking Options (track.*):\n")
  cat(sprintf("  %-*s\n",
              wval, printopt(x, "track.maxwindow", wtag)))

  cat("Display Options:\n")
  cat(sprintf("  %-*s    %-*s\n",
              wval, printopt(x, "palette", wtag),
              wval, printopt(x, "digits", wtag)))
  cat("Display Options - Annotations (mark.*):\n")
  cat(sprintf("  %-*s    %-*s\n",
              wval, printopt(x, "mark.alpha", wtag),
              wval, printopt(x, "mark.flat", wtag)))
  cat("Display Options - Color (colID.*):\n")
  cat(sprintf("  %-*s    %-*s\n",
              wval, printopt(x, "colID.peak", wtag),
              wval, printopt(x, "colID.flat", wtag)))
  cat(sprintf("  %-*s    %-*s    %-*s\n",
              wval, printopt(x, "colID.data", wtag),
              wval, printopt(x, "colID.filter", wtag),
              wval, printopt(x, "colID.hist", wtag)))
  cat(sprintf("  %-*s\n",
              wval, printopt(x, "colID.cdf", wtag)))
  
  cat("\n")

  invisible(x)
}

# Create a string from Diopt class x element tag, of format <tag>  <value>
# where tag is forced to a width of wtag.
printopt <- function(x, tag, wtag) {

  shorttag <- sub("(.+?)\\.", "", tag)
  if (("lp.tests" == tag) || ("diw.tests" == tag)) {
    paste0(sprintf("%-*s", wtag, shorttag), paste(x[[tag]], collapse=", "))
  } else if (("lp.param" == tag) || ("diw.param" == tag)) {
    if (is.null(x[[tag]])) {
      paste(sprintf("%-*s", wtag, shorttag), "-none-", sep="")
    } else {
      lval <- sapply(seq_along(x[[tag]]),
                     function(i) {
                       paste0(names(x[[tag]])[i], "=", x[[tag]][i], sep="")
                     })
      paste0(sprintf("%-*s", wtag, shorttag), paste(lval, collapse=", "))
    }
  } else if (1 == length(x[[tag]])) {
    paste(sprintf("%-*s", wtag, shorttag), x[[tag]], sep="")
  } else {
    paste0(sprintf("%-*s", wtag, shorttag), paste(x[[tag]], collapse=", "))
  }
}


### Dipeak objects.

# Pretty-print the Dipeak class object x.  Other arguments are ignored.
# Returns x invisibly.
print.Dipeak <- function(x, ...) {

  if (0 == nrow(x)) {
    cat("\n  no extrema found\n\n")
    return(invisible(x))
  }

  pkID <- which(!is.na(x$lminID) & !is.na(x$rminID))
  if (0 == length(pkID)) {
    cat("\n  no maxima found\n\n")
    return(invisible(x))
  }

  opt <- Diopt()
  dgt <- opt$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  hcol <- c("ht", "hexcur", "nrun", "runlen", "runht", "", "")
  pcol <- c("pht", "pexcur", "pnrun", "prunlen", "prunht", "ppeak", "pass")
  
  # This is to cleanly print the find.peaks example, where the raw position
  # and probabilities are not included.
  show.raw <- "x" %in% colnames(x)
  show.stats <- any(pcol %in% colnames(x))
  show.support <- any(((x$lminID != x$lsuppID) | (x$rminID != x$rsuppID))[pkID])
  
  spos <- format(x$pos[pkID])
  wpos <- max(nchar(spos))
  sx   <- format(x$x[pkID], digits=dgt)
  wx <- max(nchar(sx))
  slm  <- format(x$lminID[pkID])
  sxlm <- format(x$x[pkID-1], digits=dgt)
  srm  <- format(x$rminID[pkID])
  wmin <- max(nchar(slm), nchar(srm))
  sxrm <- format(x$x[pkID+1], digits=dgt)
  wxmin <- max(nchar(sxlm), nchar(sxrm))
  if (show.support) {
    sls  <- format(x$lsupp[pkID])    
    srs  <- format(x$rsupp[pkID])    ; wsupp <- max(nchar(sls), nchar(srs))
  } else {
    wsupp <- 0
  }

  cat("\n  location of", ifelse(1==length(pkID), "maximum", "maxima"), "\n")

  if (show.raw) {
    minhdr <- "minima pos (raw)"
  } else {
    minhdr <- "minima pos"
  }
  supphdr <- "support pos"

  if (show.raw) {
    wmintop <- 2 * (wmin + wxmin) + 9
  } else {
    wmintop <- 2 * wmin + 3
  }
  wmingap <- ceiling(max(0, nchar(minhdr) - wmintop) / 2)
  if (show.support) {
    wsupptop <- (2 * wsupp) + 3
    wsuppgap <- ceiling(max(0, nchar(supphdr) - wsupptop) / 2)
    if (show.raw) {
      posraw <- sprintf("%-*s    %*s", wpos, "pos", wx, "raw")
    } else {
      posraw <- sprintf("%-*s", wpos, "pos")
    }
    cat(sprintf("  %s     %*s    %*s\n",
                posraw, wmintop, format(minhdr, justify="c", width=wmintop),
                wsupptop, format(supphdr, justify="c", width=wsupptop)))
    for (i in seq_along(pkID)) {
      if (show.raw) {
        posraw <- sprintf("%*s   (%*s)", wpos, spos[i], wx, sx[i])
        minraw <- sprintf("%*s - %*s (%*s - %*s)", wmin, slm[i], wmin, srm[i],
                          wxmin, sxlm[i], wxmin, sxrm[i])
      } else {
        posraw <- sprintf("%*s", wpos, spos[i])
        minraw <- sprintf("%*s - %*s", wmin, slm[i], wmin, srm[i])
      }
      cat(sprintf("  %s    %*s%s    %*s%*s - %*s\n",
                  posraw, wmingap, "", minraw,
                  wsuppgap, "", wsupp, sls[i], wsupp, srs[i]))
    }
  } else {
    if (show.raw) {
      posraw <- sprintf("%*s    %-*s", wpos, "pos", wx, "raw")
    } else {
      posraw <- sprintf("%*s    %-*s", wpos, "pos", wx, "")
    }
    cat(sprintf("  %s     %*s\n",
                posraw, wmintop, format(minhdr, justify="c", width=wmintop)))
    for (i in seq_along(pkID)) {
      if (show.raw) {
        posraw <- sprintf("%*s   (%*s)", wpos, spos[i], wx, sx[i])
        minraw <- sprintf("%*s - %*s (%*s - %*s)", wmin, slm[i], wmin, srm[i],
                  wxmin, sxlm[i], wxmin, sxrm[i])
      } else {
        posraw <- sprintf("%*s    %*s ", wpos, spos[i], wx, "")
        minraw <- sprintf("%*s - %*s", wmin, slm[i], wmin, srm[i])
      }
      cat(sprintf("  %s    %s\n", posraw, minraw))
    }
  }

  cstat <- c("pos", hcol)
  cpval <-  c("pos", pcol)
  sstat <- matrix("", nrow=length(pkID), ncol=length(cpval))
  spval <- matrix("", nrow=length(pkID), ncol=length(cpval))
  sok <- rep("", length(cpval))
  keep <- rep(FALSE, length(cpval))

  if (show.stats) {
    sstat[,1] <- format(x$pos[pkID])
    spval[,1] <- format(x$pos[pkID])
    keep[1] <- TRUE
    if ("pht" %in% colnames(x)) {
      sstat[,2] <- format(pmax(x$lht[pkID], x$rht[pkID]), digits=dgt)
      spval[,2] <- mark.p(x$pht[pkID], opt$alpha.ht)
      sok[2] <- mark.p(opt$alpha.ht, NA)
      keep[2] <- TRUE
    }
    if ("pexcur" %in% colnames(x)) {
      sstat[,3] <- format(x$hexcur[pkID], digits=dgt)
      if ("Diw" == attr(x, "source")) {
        spval[,3] <- mark.p(x$pexcur[pkID], opt$alpha.pkexcur.diw)
        sok[3] <- mark.p(opt$alpha.pkexcur.diw, NA)
      } else {
        spval[,3] <- mark.p(x$pexcur[pkID], opt$alpha.pkexcur.lp)
        sok[3] <- mark.p(opt$alpha.pkexcur.lp, -1)
      }
      keep[3] <- TRUE
    }
    if ("pnrun" %in% colnames(x)) {
      sstat[,4] <- format(x$nrun[pkID])
      spval[,4] <- mark.p(x$pnrun[pkID], opt$alpha.nrun)
      sok[4] <- mark.p(opt$alpha.nrun, NA)
      keep[4] <- TRUE
    }
    if ("prunlen" %in% colnames(x)) {
      sstat[,5] <- format(x$runlen[pkID])
      spval[,5] <- mark.p(x$prunlen[pkID], opt$alpha.runlen)
      sok[5] <- mark.p(opt$alpha.runlen, NA)
      keep[5] <- TRUE
    }
    if ("prunht" %in% colnames(x)) {
      sstat[,6] <- format(x$runht[pkID])
      spval[,6] <- mark.p(x$prunht[pkID], opt$alpha.runht)
      sok[6] <- mark.p(opt$alpha.runht, NA)
      keep[6] <- TRUE
    }
    if (any(!is.na(x$ppeak[pkID]))) {
      spval[,7] <- mark.p(x$ppeak[pkID], NA)
      sel <- 0 < x$naccept[pkID]
      spval[sel,8] <- sprintf("T (%.0f)", x$naccept[pkID])[sel]
      keep[c(7,8)] <- TRUE
    }

    if (any(keep)) {
      sstat <- sstat[,keep, drop=FALSE]
      spval <- spval[,keep, drop=FALSE]
      cstat <- cstat[keep]
      cpval <- cpval[keep]
      sok <- sok[keep]

      w <- apply(nchar_marked(rbind(sstat, spval, cstat, cpval)), 2, max)
      # Extra space after position, before total probability, pass columns
      w[1] <- w[1] + 2
      if (keep[7]) {
        w[length(w)-2] <- w[length(w)-2] + 2
        w[length(w)-1] <- w[length(w)-1] + 1
      }
      for (i in 1:ncol(sstat)) {
        sstat[,i] <- format(sstat[,i], width=w[i])
        spval[,i] <- pad.marked(spval[,i], width=w[i])
        cstat[i] <- format(cstat[i], width=w[i])
        cpval[i] <- format(cpval[i], width=w[i])
        sok[i] <- format(sok[i], width=w[i])
      }

      cat("\n  statistics\n")
      cat("  ", paste0(cstat, sep="  "), "\n", sep="")
      for (i in seq_along(pkID)) {
        cat("  ", paste0(sstat[i,], sep="  "), "\n", sep="")
      }
      cat("\n  probabilities\n")
      cat("  ", paste0(cpval, sep="  "), "\n", sep="")
      for (i in seq_along(pkID)) {
        cat("  ", paste0(spval[i,], sep="  "), "\n", sep="")
      }
      cat("\n  accept at\n")
      cat("  ", paste0(sok, sep="  "), "\n", sep="")
    }
  }   # show.stats
  cat("\n")

  invisible(x)
}

# Print the location of the peaks with test results.  Other arguments are
# ignored.  Uses default alpha values from Diopt to judge significance.
# Returns object invisibly.
summary.Dipeak <- function(object, ...) {

  # Stupid check requirement that changes the argument between print, summary.
  x <- object

  if (0 == nrow(x)) {
    cat("\n  no extrema found\n\n")
    return(invisible(object))
  }

  pkID <- which(!is.na(x$lminID) & !is.na(x$rminID))
  if (0 == length(pkID)) {
    cat("\n  no maxima found\n\n")
    return(invisible(object))
  }

  opt <- Diopt()
  dgt <- opt$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  cat("\n")

  if ("Diw" == attr(x, "source")) {
    aexcur <- opt$alpha.pkexcur.diw
  } else {
    aexcur <- opt$alpha.pkexcur.lp
  }
  spec <- data.frame(var=c('pht', 'pnrun', 'prunlen', 'prunht', 'pexcur'),
                     alpha=c(opt$alpha.ht, opt$alpha.nrun, opt$alpha.runlen,
                             opt$alpha.runht, aexcur),
                     name=c('height model', 'nrun', 'runlen',
                            'runht (permutation)', 'excursion'))
  passed <- list.passes(x[pkID,], spec)

  cat("  summary of ", ifelse(1==length(pkID), "maximum", "maxima"), "\n",
      sep="")
  
  spos <- format(x$pos[pkID])            ; wpos <- max(nchar(spos), 3)
  sx   <- format(x$x[pkID], digits=dgt)  ; wx <- max(nchar(sx), 3)
  if ("ppeak" %in% colnames(x)) {
    spval <- mark.p(x$ppeak[pkID], NA)   ;  wpval <- max(nchar(spval), 5)
    cat(sprintf("  %*s    %*s    %*s     passing tests\n",
                wpos, "pos", wx, "raw", wpval, "ppeak"))
  } else {
    cat(sprintf("  %*s    %*s      passing tests\n",
                wpos, "pos", wx, "raw"))
  }
  for (i in seq_along(pkID)) {
    if ("ppeak" %in% colnames(x)) {
      cat(sprintf("  %*s   (%*s)   %*s     %s\n",
                  wpos, spos[i], wx, sx[i], wpval, spval[i], passed[i]))
    } else {
      cat(sprintf("  %*s   (%*s)     %s\n",
                  wpos, spos[i], wx, sx[i], passed[i]))
    }
  }
  
  cat("\n")
  
  invisible(object)
}

# Build a list of passing tests in Dipeak or Diflat object x, where pspec is
# a data frame with columns $var the name of the variable, $alpha the
# acceptance elvel, and $name the full test name.  The passing test names are
# gathered in a comma-separated list, or "none" if empty.  Returns a vector
# of these results per row in x.
list.passes <- function(x, pspec) {

  passed <- matrix("", nrow=nrow(x), ncol=nrow(pspec))
  for (i in 1:nrow(pspec)) {
    if (pspec$var[i] %in% colnames(x)) {
      pass <- !is.na(x[,pspec$var[i]]) & (x[,pspec$var[i]] < pspec$alpha[i])
      passed[,i] <- ifelse(pass, pspec$name[i], "")
    }
  }
  keep <- 0 < nchar(passed)
  sapply(1:nrow(passed),
         function(i) {
           k <- keep[i,]
           ifelse(!any(k), "none", paste0(passed[i,k], collapse=", "))
         })
}

# Annotate a plot with the location of local peaks and minima in Dipeak
# object x.  Pass the histogram hist if drawing atop one, otherwise assume
# the plot is for spacing.  The Diopt options opt provide graphical parameters
# and other arguments are ignored.  Returns x invisibly.
mark.Dipeak <- function(x, hist=NULL, opt=Diopt(), ...) {

  if (nrow(x) <= 2) {
    return()
  }

  clr <- get.colors(opt$palette)[opt$colID.peak]
  
  # Automatically ignore the first and last entry in the table.
  x <- x[-c(1,nrow(x)),]

  if (is.null(hist)) {
    if ("ppeak" %in% colnames(x)) {
      sel <- x$ismax & (0 < x$naccept)
      abline(v=x$pos[sel], col=clr, lty="62", lwd=2)
      sel <- x$ismax & (is.na(x$naccept) | (0 == x$naccept))
      abline(v=x$pos[sel], col=clr, lty="26", lwd=1)
    } else {
      abline(v=x$pos[x$ismax], col=clr, lty="26", lwd=1)
    }
    abline(v=x$pos[!x$ismax], col=clr, lty=3, lwd=1)
  } else {
    # Something big off the graph to right.
    xr <- 2 * max(hist$breaks)
    y <- max(hist$counts) * x$pos / sum(hist$counts)
    if ("ppeak" %in% colnames(x)) {
      sel <- x$ismax & (0 < x$naccept)
      for (i in which(sel)) {
        lines(c(x$x[i], xr), c(y[i], y[i]),  col=clr, lty="62", lwd=2)
        lines(c(x$x[i], x$x[i]), c(0, y[i]), col=clr, lty="62", lwd=2)
      }
      sel <- x$ismax & (is.na(x$naccept) | (0 == x$naccept))
      for (i in which(sel)) {
        lines(c(x$x[i], xr), c(y[i], y[i]),  col=clr, lty="26", lwd=1)
        lines(c(x$x[i], x$x[i]), c(0, y[i]), col=clr, lty="26", lwd=1)
      }
    } else {
      for (i in which(x$ismax)) {
        lines(c(x$x[i], xr), c(y[i], y[i]),  col=clr, lty="26",  lwd=1)
        lines(c(x$x[i], x$x[i]), c(0, y[i]), col=clr, lty="26",  lwd=1)
      }
    }
  }

  invisible(x)
}

# Create an empty Dipeak object with 0 rows, which should be used as the
# check if there were any features found.
mockup.Dipeak <- function() {

  # These come from find.peaks().
  peakcols <- c("pos", "ismax", "valsd", "lht", "rht", "lminID", "rminID",
                "lsuppID", "rsuppID")

  m <- matrix(0, nrow=0, ncol=length(peakcols), dimnames=list(NULL, peakcols))
  df <- as.data.frame(m)
  class(df) <- c("Dipeak", class(df))
  df
}


### Diflat objects.

# Pretty-print the Diflat class object x.  Other arguments are ignored.
# Returns x invisibly.
print.Diflat <- function(x, ...) {

  if (0 == nrow(x)) {
    cat("\n  no flats found\n\n")
    return(invisible(x))
  }

  opt <- Diopt()
  dgt <- opt$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  hcol <- c("len", "hexcur", "", "")
  pcol <- c("plen", "pexcur", "pflat", "pass")
  show.raw <- ("xst" %in% colnames(x)) && ("xend" %in% colnames(x))
  show.stats <- any(pcol %in% colnames(x))

  sst <- format(x$stID)
  send <- format(x$endID)
  wpos <- max(nchar(sst), nchar(send))
  sst <- format(sst, width=wpos)
  send <- format(send, width=wpos)

  sxst <- format(x$xst, digits=dgt)
  sxend <- format(x$xend, digits=dgt)
  # Uniform widths.  format may treat st, end differently so per column.
  sxst <- format(sxst, width=max(nchar(sxst)))
  sxend <- format(sxend, width=max(nchar(sxend)))

  cstat <- c("endpoints", "raw", hcol, "")
  cpval <- c("endpoints", "", pcol, "")
  sstat <- matrix("", nrow=nrow(x), ncol=length(cpval))
  spval <- matrix("", nrow=nrow(x), ncol=length(cpval))
  sok <- rep("", length(cpval))
  keep <- rep(FALSE, length(cpval))

  sstat[,1] <- sapply(seq_along(sst),
                      function(i) { paste0(sst[i], " - ", send[i]) })
  spval[,1] <- sstat[,1]
  keep[1] <- TRUE
  if (show.raw) {
    sstat[,2] <- sapply(seq_along(sxst),
                        function(i) {
                          paste0("(", sxst[i], " ", sxend[i], ")") })
    keep[2] <- TRUE
  }
  sok[1] <- "accept at"

  if ("plen" %in% colnames(x)) {
    sstat[,3] <- format(x$len)
    spval[,3] <- mark.p(x$plen, opt$alpha.len)
    sok[3] <- mark.p(opt$alpha.len, NA)
    keep[3] <- TRUE
  }

  if ("pexcur" %in% colnames(x)) {
    sstat[,4] <- format(x$hexcur, digits=dgt)
    if ("Diw" == attr(x, "source")) {
      spval[,4] <- mark.p(x$pexcur, opt$alpha.ftexcur.diw)
      sok[4] <- mark.p(opt$alpha.ftexcur.diw, NA)
    } else {
      spval[,4] <- mark.p(x$pexcur, opt$alpha.ftexcur.lp)
      sok[4] <- mark.p(opt$alpha.ftexcur.lp, NA)
    }
    keep[4] <- TRUE
  }

  if (any(!is.na(x$pflat))) {
    spval[,5] <- mark.p(x$pflat, NA)
    sel <- 0 < x$naccept
    spval[sel,6] <- sprintf("T (%0.f)", x$naccept)[sel]
    keep[c(5,6)] <- TRUE
  }

  sstat <- sstat[,keep, drop=FALSE]
  spval <- spval[,keep, drop=FALSE]
  cstat <- cstat[keep]
  cpval <- cpval[keep]
  sok <- sok[keep]

  w <- apply(nchar_marked(rbind(sstat, spval, cstat, cpval, sok)), 2, max)
  # Extra space after position, before total probability, pass columns.
  if (show.raw) {
    rawcol <- 2
    w[rawcol] <- w[rawcol] + 2
  } else {
    rawcol <- 0
    w[1] <- w[1]
  }
  if (keep[5]) {
     w[length(w)-2] <- w[length(w)-2] + 2
     w[length(w)-1] <- w[length(w)-1] + 1
  }
  for (i in 1:ncol(sstat)) {
    sstat[,i] <- format(sstat[,i], width=w[i])
    spval[,i] <- pad.marked(spval[,i], width=w[i])
    cstat[i] <- format(cstat[i], width=w[i], justify=ifelse(rawcol==i,"c","l"))
    cpval[i] <- format(cpval[i], width=w[i], justify=ifelse(rawcol==i,"c","l"))
    sok[i] <- format(sok[i], width=w[i])
  }

  cat("\n")
  if (show.raw) {
    shdr1 <- format("", width=w[1]+w[2]+2)
  } else {
    shdr1 <- format("", width=w[1])
  }
  if (show.stats) {
    cat("  ", shdr1, "  statistics\n", sep="")
  }
  cat("  ", paste0(cstat, sep="  "), "\n", sep="")
  for (i in 1:nrow(x)) {
    cat("  ", paste0(sstat[i,], sep="  "), "\n", sep="")
  }
  if (show.stats) {
    cat("\n  ", shdr1, "  probabilities\n", sep="")
    cat("  ", paste0(cpval, sep="  "), "\n", sep="")
    for (i in 1:nrow(x)) {
      cat("  ", paste0(spval[i,], sep="  "), "\n", sep="")
    }
    cat("\n  ", paste0(sok, sep="  "), "\n", sep="")
  }
  cat("\n")

  invisible(x)
}

# Print the location of flats and their probabilities.  Other arguments are
# ignored.  Returns object invisibly.  Uses default alpha values from Diopt
# to judge significance.
summary.Diflat <- function(object, ...) {

  x <- object

  if (0 == nrow(x)) {
    cat("\n  no flats found\n\n")
    return(invisible(object))
  }

  opt <- Diopt()
  dgt <- opt$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  if ("Diw" == attr(x, "source")) {
    aexcur <- opt$alpha.ftexcur.diw
  } else {
    aexcur <- opt$alpha.ftexcur.lp
  }
  spec <- data.frame(var=c("plen", "pexcur"),
                     alpha=c(opt$alpha.len, aexcur),
                     name=c("length model", "excursion"))
  passed <- list.passes(x, spec)
  
  sst <- format(x$stID)
  send <- format(x$endID)
  wpos <- max(nchar(sst), nchar(send), ceiling((nchar("endpoints")-1)/2))
  sst <- format(sst, width=wpos)
  send <- format(send, width=wpos)
  spts <- sapply(seq_along(sst),
                 function(i) { paste0(sst[i], " - ", send[i]) })
  sposlbl <- format("endpoints", width=max(nchar(spts)), justify="l")

  sxst <- format(x$xst, digits=dgt)
  sxend <- format(x$xend, digits=dgt)
  # Uniform widths.  format may treat st, end differently so per column.
  sxst <- format(sxst, width=max(nchar(sxst)))
  sxend <- format(sxend, width=max(nchar(sxend)))
  sloc <- sapply(seq_along(sxst),
                 function(i) { paste0("(", sxst[i], " ", sxend[i], ")") })
  sloclbl <- format("raw", width=max(nchar(sloc)), justify="c")

  if ("pflat" %in% colnames(x)) {
    spflat <- mark.p(x$pflat, NA)
    wpflat <- max(nchar(spflat), nchar("pflat"))
    spflat <- format(spflat, width=wpflat)
    splbl <- format("pflat", width=wpflat)
  } else {
    wpflat <- 0
    spflat <- rep("", nrow(x))
    splbl <- ""
  }

  cat("\n")
  cat("  summary of ", ifelse(1==nrow(x), "flat", "flats"), "\n", sep="")
  cat("  ", sposlbl, "  ", sloclbl, "  ", splbl, "    passing tests\n", sep="")
  for (i in 1:nrow(x)) {
    cat("  ", spts[i], "  ", sloc[i], "  ", spflat[i], "    ", passed[i], "\n",
        sep="")
  }
  cat("\n")
  
  invisible(object)
}

# Annotate a plot with the location of local flats in Diflat object x.  Pass
# the histogram hist if drawing atop one, otherwise assume the plot is for
# spacing.  The Diopt options opt provide graphical parameters and other
# arguments are ignored.  Returns x invisibly.
mark.Diflat <- function(x, hist=NULL, opt=Diopt(), ...) {

  if (0 == nrow(x)) {
    return()
  }

  if ("pflat" %in% colnames(x)) {
    is.sig <- 0 < x$naccept
  } else {
    is.sig <- rep(FALSE, nrow(x))
  }

  clr <- get.colors(opt$palette)[opt$colID.flat]

  if (is.null(hist)) {
    if ("box" == opt$mark.flat) {
      # Make color transparent so it doesn't obscure the filtered value.
      rgbval <- col2rgb(clr)
      clr <- rgb(rgbval[1], rgbval[2], rgbval[3], 0x8f, maxColorValue=255)
    } else if ("bar" == opt$mark.flat) {
      tics <- axTicks(2)
      off <- 0.1 * diff(range(tics))
    } else {
      stop("internal error - unsupported flat marking")
    }
    
    for (i in 1:nrow(x)) {
      y <- x$srcval[i]
      dy <- x$ht[i] / 2
      if ("box" ==opt$mark.flat) {
        lwd <- ifelse(!is.na(is.sig[i]) & is.sig[i], 2.5, 1.5)
        rect(x$stID[i], y-dy, x$endID[i], y+dy, border=clr, lwd=lwd)
      } else if ("bar" == opt$mark.flat) {
        xends <- c(x$stID[i], x$endID[i])
        lwd <- ifelse(!is.na(is.sig[i]) & is.sig[i], 3, 1.5)
        if ((y - dy - tics[1]) < (tics[length(tics)] - (y + dy))) {
          lines(xends, c(y+dy+off,y+dy+off), col=clr, lwd=lwd)
        } else {
          lines(xends, c(y-dy-off,y-dy-off), col=clr, lwd=lwd)
        }
        points(xends, c(y, y), col=clr, cex=1.5)
      } else {
        stop("internal error - unsupported flat marking")
      }
    }
  } else {
    yflat <- 1.02 * max(hist$counts)
    for (i in 1:nrow(x)) {
      yflat <- max(hist$counts) * ifelse(is.sig[i], 1.025, 1.01)
      lwd <- ifelse(!is.na(is.sig[i]) & is.sig[i], 3, 1.5)
      lines(c(x$xst[i], x$xend[i]), c(yflat, yflat), col=clr, lwd=lwd)
    }
  }

  invisible(x)
}

# Create an empty Diflat object with 0 rows, which should be used as the
# check if there were any features found.
mockup.Diflat <- function() {

  # These come from find.flats().
  flatcols <- c("src", "stID", "endID", "len", "srcval", "ht", "htsd")

  m <- matrix(0, nrow=0, ncol=length(flatcols), dimnames=list(NULL, flatcols))
  df <- as.data.frame(m)
  class(df) <- c("Diflat", class(df))
  df
}


### Didata objects.

# Print tables with the setup of the analysis data (source, filters applied)
# and statistics of the raw data, spacing, and low-pass spacing and interval
# spacing if generated.  Extra arguments are ignored.  Returns x invisibly.
print.Didata <- function(x, ...) {

  if ((0 == nrow(x)) || (ncol(x) < 2)) {
    cat("\n  no data present\n\n")
    return(invisible(x))
  }

  opt <- Diopt()
  dgt <- opt$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  a <- attributes(x)

  have.lp <- "lp.kernel" %in% names(a)
  have.diw <- "diw.window" %in% names(a)

  xstat <- NULL
  idstat <- NULL
  wid = nchar(format(ncol(x)))
  sID <- paste0(format(1, width=wid), " - ", format(ncol(x), width=wid), sep="")
  slbl <- c("row", format("valid", width=nchar(sID), justify="c"),
            "range", "sd")
  xstat <- rbind(xstat, slbl)
  xstat <- rbind(xstat,
                 c("x", sID, format(diff(range(x["x",])), digits=dgt),
                   format(sd(x["x",]), digits=dgt)))
  sID <- paste0(format(2, width=wid), " - ", format(ncol(x), width=wid), sep="")
  xdi <- x["Di",2:ncol(x)]
  xstat <- rbind(xstat,
                 c("Di", sID, format(diff(range(xdi)), digits=dgt),
                   format(sd(xdi), digits=dgt)))

  cat("  data source       ", a$x, "\n")

  if (have.lp) {
    if (a$lp.window < 1.0) {
      flp <- a$lp.window
    } else {
      flp <- a$lp.window / ncol(x)
    }
    cat(sprintf("  low-pass spacing   with %3d (%5.3f) %s filter\n",
               a$wlp, flp, a$lp.kernel))
    sID <- paste0(format(a$lp.stID, width=wid), " - ",
                  format(a$lp.endID, width=wid), sep="")
    xlp <- x["lp",a$lp.stID:a$lp.endID]
    xstat <- rbind(xstat,
                   c("LP Di", sID, format(diff(range(xlp)), digits=dgt),
                     format(sd(xlp), digits=dgt)))
  }
  
  if (have.diw) {
    if (a$diw.window < 1.0) {
      fiw <- a$diw.window
    } else {
      fiw <- a$diw.window / ncol(x)
    }
    cat(sprintf("  interval spacing   with %3d (%5.3f) interval\n",
                a$wdiw, fiw))
    sID <- paste0(format(a$diw.stID, width=wid), " - ",
                  format(ncol(x), width=wid), sep="")
    xdiw <- x["Diw",a$diw.stID:ncol(x)]
    xstat <- rbind(xstat,
                   c("Diw", sID, format(diff(range(xdiw)), digits=dgt),
                     format(sd(xdiw), digits=dgt)))
    if (have.lp) {
      cat(sprintf("     positions at interval end; shift by %+d vs. low-pass\n",
                  ((a$lp.stID+a$lp.endID-1) - (a$diw.stID+ncol(x)-1)) %/% 2))
    } else {
      cat("     positions at interval end\n")
    }
  }
  
  if (!have.lp && !have.diw) {
    cat("  no filters applied\n")
  }

  xstat[,1] <- format(xstat[,1], width=max(nchar(xstat[,1])))
  xstat[,3] <- format(xstat[,3], width=max(nchar(xstat[,3])))
  xstat[,4] <- format(xstat[,4], width=max(nchar(xstat[,4])))

  cat("\nInformation\n")
  for (i in 1:nrow(xstat)) {
    cat("  ", paste(xstat[i,], collapse="    "), "\n", sep="")
  }
  cat("\n")

  invisible(x)
}

# Describe the setup used to generate the data: the data source and filters
# applied.  Extra arguments ignored.  Returns object invisibly.
summary.Didata <- function(object, ...) {

  x <- object

  if (0 == nrow(x)) {
    cat("\n  no data present\n\n")
    return(invisible(object))
  }

  a <- attributes(x)
  have.lp <- "lp.kernel" %in% names(a)
  have.diw <- "diw.window" %in% names(a)

  cat("\n")
  cat("Setup\n")
  cat("  data source       ", a$x, "\n")
  
  if (have.lp) {
    if (a$lp.window < 1.0) {
      flp <- a$lp.window
    } else {
      flp <- a$lp.window / ncol(x)
    }
    cat(sprintf("  low-pass spacing   with %3d (%5.3f) %s filter\n",
               a$wlp, flp, a$lp.kernel))
  }

  if (have.diw) {
    if (a$diw.window < 1.0) {
      fiw <- a$diw.window
    } else {
      fiw <- a$diw.window / ncol(x)
    }
    cat(sprintf("  interval spacing   with %3d (%5.3f) interval\n",
                a$wdiw, fiw))
  }

  if (!have.lp && !have.diw) {
    cat("  no filters applied\n")
  }
  cat("\n")

  invisible(object)
}

# Create an empty matrix with a fixed number of columns and no rows (test
# nrow to check if this is a dummy value).
mockup.Didata <- function() {

  m <- matrix(0, nrow=0, ncol=10, dimnames=list(NULL, NULL))
  attributes(m) <- c(attributes(m), list(x="mockup"))
  class(m) <- c("Didata", class(m))
  m
}


### Common code

# Common formatting of p values, with the potential to mark those at or
# under the significance threshold alpha if the option mark.alpha is TRUE.
# Use a negative number (not zero) or NA for alpha to not mark and just
# format probabilities consistently.  Returns a vector of strings per p.
mark.p <- function(p, alpha) {

  pout <- sprintf("%.*f", 4, p)
  if (!is.na(alpha) && (0 <= alpha) && Diopt()$mark.alpha) {
    for (i in which(p<=alpha)) {
      pout[i] <- sprintf("\033[4m%s\033[24m", pout[i])
    }
  }
  pout
}

# Count the characters in character object x (vector, matrix), adjusting for
# any marked p values.  Returns the counts.
nchar_marked <- function(x) {

  nx <- nchar(x)

  # Do before nx changes.
  at.end <- which('\033[24m' == substring(x, nx-4))
  nx[at.end] <- nx[at.end] - 5
  
  at.st <- which('\033[4m' == substr(x, 1, 4))
  nx[at.st] <- nx[at.st] - 4

  nx
}

# Increase the width of character vector s elements to be width, by adding
# spaces at the end.  If an element is already longer than width, do nothing.
pad.marked <- function(s, width) {

  sapply(s,
         function(s2) {
           w <- nchar(s2)
           if ('\033[24m' == substring(s2, w-4)) {
             w <- w - 5
           }
           if ('\033[4m' == substr(s2, 1, 4)) {
             w <- w - 4
           }
           if (w < width) {
             paste0(s2, format(' ', width=width-w))
           } else {
             s2
           }
         })
}

# sigtests.R -
# Feature significance tests.
#
# c 2024-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
# - 
#

#### Public Interface

# Determine the quantile/significance level of the local maximum with height(s)
# ht in a sample with n data points and after low-pass filtering the spacing
# with the kernel type filter of size flp, which may be the total window size
# or a fraction of the data.  lower.tail determines whether the probability
# of the null distribution is less than/equal (TRUE) or greater than ht.
# Returns a list of class 'Ditest' with elements:
#    $method            descriptive text
#    $statfn            name (text) of function used to evaluate statistic
#    $statistic         what is tested, the adjusted standardized peak height
#    $statname          text name of the statistic
#    $parameter         inverse Gaussian/Wald parameters corrht, mu, lambda
#    $p.value           probability of height
#    $alternative       direction of test
#    $model             model parameters ht, n, flp, filter from arguments
# The statistic and p.value will have the length of the ht argument.
Dipeak.test <- function(ht, n, flp, filter, lower.tail=TRUE) {

  n <- n[1L]
  flp <- flp[1L]
  filter <- filter[1L]
  
  check.modelargs(n, flp, lower.tail)
  
  if (1 <= flp) {
    flp <- flp / n
  }

  # NULL ht returns empty vector.  Other invalid values to NA, coercing the
  # vector out of its original type in the process.
  if (0 < length(ht)) {
    ht[!is.numeric(ht)] <- NA_real_
    ht <- as.numeric(ht)
  }

  frat <- peak.filterratio(filter)

  corr <- peak.correction(flp)

  x <- 10^(((ht / frat) - corr["badj"]) / corr["madj"])
  names(x) <- NULL

  mu <- peak.qwald.param(1, n, flp)
  lambda <- peak.qwald.param(2, n, flp)
  p <- statmod::pinvgauss(x, mu, lambda, lower.tail=lower.tail)

  p[is.na(ht)] <- NA_real_
  p[is.nan(ht)] <- NaN

  res <- list(method="low-pass spacing height model",
              statfn="pinvgauss", statistic=ht, statname="ht",
              parameter=list(corrht=x, mu=mu, lambda=lambda), p.value=p,
              alternative=ifelse(lower.tail, "less", "greater"),
              model=c(n=n, flp=flp, filter=filter))
  class(res) <- "Ditest"
  res
}

# Return the local maxima height(s) at significance level(s) pval in a sample
# with n data points and after low-pass filtering with kernel filter of size
# flp, which may be the vector length or a fraction of the data.  Invalid pval
# evaluate to NaN.
# NOTE: when building and verifying models based on quantiles, use pval=1-q.
Dipeak.critval <- function(pval, n, flp, filter) {

  n <- n[1L]
  flp <- flp[1L]
  filter <- filter[1L]

  check.modelargs(n, flp, TRUE)
  
  if (1 <= flp) {
    flp <- flp / n
  }

  if (0 < length(pval)) {
    pval[!is.numeric(pval)] <- NA_real_
    pval <- as.numeric(pval)
  }
  # This is the behavior of the quantile functions.
  pgood <- (pval >= 0) & (pval <= 1)
  pval[!pgood] <- NaN

  frat <- peak.filterratio(filter)

  # Model of the Weibull scale=2, shape=4 null distribution.
  mu <- peak.qwald.param(1, n, flp)
  lambda <- peak.qwald.param(2, n, flp)
  htpred <- statmod::qinvgauss(1-pval, mu, lambda)

  # Could also modify prediction for different distributions.
  #   beta    = 5.76 - (4.62 * htpred)
  #   normal  = 0.93 * htpred
  #   gumbel, gamma, chisq, logistic  = -1.40 + (2.25 * htpred)
  corr <- peak.correction(flp)
  cv <- frat * (corr["badj"] + (corr["madj"] * log10(htpred)))
  names(cv) <- NULL
  cv
}

# Return the quantile/significance level of a flat of length(s) len in a
# sample of n data points drawn from distribution basedist, then low-pass
# filtered with kernel filter of size flp (may be fractional or the actual
# size).  lower.tail determines whether the probability of the null
# distribution is less than/equal (TRUE) or greater than (FALSE) the length.
# Returns a list of class 'Ditest' with elements
#    $method             descriptive text
#    $statfn             name (text) of function used to evaluate statistic
#    $statistic          what is tested, the flat length
#    $statname           text name of the statistic
#    $p.value            probability of length in model
#    $alternative        direction of test
#    $model              model parameters n, flp, filter, basedist from args
# The statistic and p.value will have the length of the len argument.  The
# p-value will be NaN if there's a problem setting up the test, otherwise
# a value between 0 and 1.
Diflat.test <- function(len, n, flp, filter, basedist, lower.tail=TRUE) {

  n <- n[1L]
  flp <- flp[1L]
  filter <- filter[1]
  basedist <- basedist[1L]

  check.modelargs(n, flp, lower.tail)

  if (1 <= flp) {
    flp <- flp / n
  }

  if (0 < length(len)) {
    len[!is.numeric(len)] <- NA_real_
    len <- as.numeric(len)
  }

  pcf <- flat.pcoef(n, flp, filter, basedist)

  optfn <- function(p, l, pcf) {
    sum( c(1, p, p^2, log(p/(1-p))) * pcf ) - l
  }
  
  # Needed for optimize, not uniroot.
  sefn <- function(p, l, pcf) {
    (sum( c(1, p, p^2, log(p/(1-p))) * pcf ) - l)^2
  }

  pmin <- flat.pmin(pcf)
  if (is.na(pmin)) {
    # logit blows up at 0 and 1, so offset a little for finite boundaries.
    prng <- c(.Machine$double.eps, 1-.Machine$double.eps)
  } else {
    # We've built the model only for q >= 0.9.  Assume a linear drop below
    # this, although it doesn't match the actual behavior.
    prng <- c(pmin, 1-.Machine$double.eps)
  }
  lmin <- optfn(prng[1L], 0, pcf)
  lmax <- optfn(prng[2L], 0, pcf)

  p <- sapply(len,
              function(l) {
                if (is.infinite(l)) {
                  ifelse(l < 0, 0, 1)
                } else if (!is.finite(l)) {
                  NaN
                } else if (l < lmin) {
                  pmin * (l / lmin)
                } else if (lmax < l) {
                  1.0
                } else {
                  pval <- safe.uniroot(optfn, prng, l, pcf, maxiter=2500,
                                       tol=.Machine$double.eps)$root
                  if (is.na(pval)) {
                    pval <- optimize(sefn, prng, l, pcf,
                                     lower=prng[1], upper=prng[2],
                                     tol=.Machine$double.eps)$minimum
                  }
                  pval
                }
              })

  if (!lower.tail && (0 < length(p))) {
    p <- 1 - p
  }
  
  res <- list(method="low-pass spacing height model",
              statfn="Diflat.test", statistic=len, statname="len",
              p.value=p,
              model=c(n=n, flp=flp, filter=filter, basedist=basedist),
              alternative=ifelse(lower.tail, "less", "greater"))
              
  class(res) <- "Ditest"
  res
}

# Return the flat length at significance level pval for a draw of size n from
# distribution basedist and smoothed by a low-pass kernel of type filter and
# size flp (may be fractional or the actual size).  Invalid p, or p values
# that are outside the valid region of our model, evaluate to NaN.
# NOTE: when building and verifying models based on quantiles, use pval=1-q.
Diflat.critval <- function(pval, n, flp, filter, basedist) {

  n <- n[1L]
  flp <- flp[1L]
  filter <- filter[1L]
  basedist <- basedist[1L]

  check.modelargs(n, flp, TRUE)

  if (1 <= flp) {
    flp <- flp / n
  }

  if (0 < length(pval)) {
    pval[!is.numeric(pval)] <- NA_real_
    pval <- as.numeric(pval)
  }
  pgood <- (pval >= 0) & (pval <= 1)
  pval[!pgood] <- NaN

  pcf <- flat.pcoef(n, flp, filter, basedist)
  pmin <- flat.pmin(pcf)
  if (is.na(pmin)) {
    pmin <- 0
  }

  # Internally we work with quantiles.
  p <- 1 - pval

  pmat <- matrix(c(rep(1, length(p)), p, p^2, log(p/(1-p))),
                 ncol=4, byrow=FALSE)

  # Replace the model with a linear drop from lmin to 0 over pmin.
  lmin <- (c(1, pmin,  pmin^2, log(pmin/(1-pmin))) %*% pcf)[1]
  # Remove NA values from the selection.
  ltmin <- p < pmin
  ltmin[is.na(ltmin)] <- FALSE
  pmat[ltmin,c(1,3,4)] <- 0
  pmat[ltmin,2] <- p[ltmin] * lmin / (pmin * pcf[2])

  cv <- as.vector( pmat %*% pcf)
  cv[(cv < 0) | !is.finite(cv)] <- NaN
  cv
}

# Perform the Kaplansky-Riordan test on the number of runs in the data vector
# x, using the find.runs detector to count runs whose values are within feps
# of their average (ignored for character, logical, or integer values).
# Runs are tested within each interval between start point(s) stID and
# endpoint(s) endID (incl.).  NA and NaN indices generate NA test values,
# real values are rounded to integers, and non-numeric arguments raise
# errors.  lower.tail if TRUE the probability is that the run count is not
# more than the observed.  Returns a list with class 'Ditest' with elements:
#    $method             descriptive text
#    $statfn             name (text) of function used to evaluate statistic
#    $statistic          what is tested, the number of runs
#    $statname           text name of statistic
#    $parameter          Gaussian distribution parameters Erun, Vrun; feps
#    $p.value            probability of number of runs is <= actual
#    $alternative        direction of test
# The parameters are used to standardize the statistic for the test, with
# Erun being the expected number of runs and Vrun the variance in the count.
# The statistic and p.value will have the same length, one per start/end
# segment, which may be 0 if stID and endID are empty.
Dinrun.test <- function(x, stID, endID, feps, lower.tail=TRUE) {

  check.endargs(stID, endID, length(x))

  if (0 == length(stID)) {
    stats <- matrix(NA_real_, nrow=4, ncol=0)
  } else {
    stID <- round(stID)
    endID <- round(endID)

    # A straightforward implementation of the equations in Example 1 of
    # Section 6 of Kaplansky and Riordan.

    stats <- sapply(seq_along(stID),
                    function(s) {
                      if (is.na(stID[s]) || is.nan(stID[s]) ||
                          is.na(endID[s]) || is.nan(endID[s])) {
                         return(c(NA_real_, NA_real_, NA_real_, NA_real_))
                      }

                      sel <- x[stID[s]:endID[s]]
                      runs <- find.runs(sel, feps)

                      n <- table(sel)
                      # Something changed between R 4.0.1 and 4.4.1 where
                      # n became integers that would overflow on large sets.
                      # Making real fixes, and also wipes the element names.
                      n <- as.double(n)
                      nuniq <- length(n)

                      a1 <- sum(n)
                      if (2 == nuniq) {
                        # Actually the Wald/Wolfowitz test.
                        a2 <- n[1] * n[2]
                        a3 <- 0
                      } else if (3 == nuniq) {
                        a2 <- (n[1]*n[2]) + (n[1]*n[3]) + (n[2]*n[3])
                        a3 <- n[1] * n[2] * n[3]
                      } else {
                        a2 <- 0
                        a3 <- 0
                        for (i in 1:(nuniq-1)) {
                          for (j in (i+1):nuniq) {
                            aij <- n[i] * n[j]
                            a2 <- a2 + aij
                            if ((i < (nuniq - 1)) && (j < nuniq)) {
                              for (k in (j+1):nuniq) {
                                a3 <- a3 + (aij * n[k])
                              }
                            }
                          }
                        }
                      }

                      if (nuniq < 2) {
                        erun <- 0
                        vrun <- 0
                      } else {
                        erun <- 1 + ((2 * a2) / a1)
                        vrun <- (((2*a2) * ((2*a2) - a1)) -
                                 (6*a1*a3)) / (a1 * a1 * (a1-1))
                      }

                      # Per Wald/Wolfowitz (last paragraph Sec. 5),
                      # p is probability of nrun or fewer runs.
                      p <- pnorm(runs$stats$nrun, erun, sqrt(vrun),
                                 lower.tail=lower.tail)
                      c(runs$stats$nrun, erun, vrun, p)
                    })
  }

  res <- list(method="Kaplanksy-Riordan test on number of runs",
              statfn="pnorm", statistic=stats[1,], statname="nrun",
              parameter=list(Erun=stats[2,], Vrun=stats[3,], feps=feps),
              p.value=stats[4,],
              alternative=ifelse(lower.tail, "less", "greater"))
  class(res) <- "Ditest"
  res
}

# Determine the probability of the longest run   in the data vector x, using
# the find.runs detector to count runs whose values are within feps of the
# average (ignored for character, logical, or integer values).  Runs are
# checked within each each interval between start point(s) stID and endpoint(s)
# endID (incl.).  NA and NaN indices generate NA test values, reals are
# rounded to integers, and non-numeric or OOB endpoints raise erros.  Returns
# a list of class 'Ditest' with elements
#    $method             descriptive text
#    $statfn             name of function used to evaluate statistic
#    $statistic          what is tested, the maximum run length
#    $statname           text name of statistic
#    $parameter          test function parameter featlen (endID-stID+1)
#    $p.value            probability of having a run of the given length
#    $tmat               Markov chain transition matrix
#    $smat               Markov chain stationary state
# The statistic, parameter, and p.value will have the same length, one per
# start/end segment.
Dirunlen.test <- function(x, stID, endID, feps) {

  check.endargs(stID, endID, length(x))

  if (0 == length(x)) {
    runlen <- numeric(0)
    featlen <- endID - stID + 1
    p <- numeric(0)
    tmat <- NULL
    smat <- NULL
  } else {
    tmat <- mkv.transition(x)
    smat <- mkv.stationary(tmat)

    if (0 == length(stID)) {
      runlen <- numeric(0)
      featlen <- endID - stID + 1
    } else {
      rl <- sapply(seq_along(stID),
                   function(s) {
                     if (is.na(stID[s]) || is.nan(stID[s]) ||
                         is.na(endID[s]) || is.nan(endID[s])) {
                        c(NA_real_, NA_real_)
                     } else {
                       sel <- x[stID[s]:endID[s]]
                       c(find.runs(sel, feps)$stats$maxrun,
                         sum(!is.na(sel) & !is.nan(sel)))
                     }
                   })
      runlen <- rl[1,]
      featlen <- rl[2,]
    }

    if (0 == length(stID)) {
      p <- numeric(0)
    } else  {
      p1 <- sapply(seq_along(stID),
                   function(s) {
                     test.markov.runlen(runlen[s], featlen[s], tmat, smat)
                   })

      p2 <- sapply(seq_along(stID),
                   function(s) {
                     test.markov.runlen(runlen[s]-1, featlen[s], tmat, smat)
                   })

      p <- p1 - p2
    }
  }
  
  res <- list(method="Markov chain model of longest run",
              statfn="test.markov.runlen", statistic=runlen, statname="runlen",
              parameter=list(featlen=featlen), p.value=p, tmat=tmat, wt=smat)
  class(res) <- "Ditest"
  res
}

# Build a distribution of heights by nexcur draws from data xbase, each draw
# of ndraw points, reconstruct the signal and height depending on whether
# is.peak is TRUE or FALSE for flats, and determine the probability of each
# height ht.  ht and ndraw may be vectors and must have the same length.  If
# lower.tail is TRUE count the number of simulations whose reconstructed
# height is smaller than ht, if FALSE larger; in other words use TRUE for
# flats and FALSE for peaks.  seed if a positive integer is used to set up
# the RNG before the draws.  Returns a list of class 'Ditest' with elements
#    $method             descriptive text
#    $statfn             name (text) of function used to evaluate statistic
#    $statistic          what is tested, the number of runs
#    $statname           text name of statistic
#    $parameter          function arguments stID, endID, permute, nexcur, seed
#    $p.value            probability of number of runs
#    $alternative        direction of test
#    $critval            critical values at quantiles 0.10, 0.05, 0.01, 0.005
#                          (or 1-q if lower.tail TRUE)
#    $xbase              base of samples
Diexcurht.test <- function(ht, ndraw, xbase, nexcur, is.peak, lower.tail=TRUE,
                           seed=0) {

  nexcur <- round(nexcur[1])
  check.htargs(ht, ndraw, xbase, nexcur, lower.tail)

  if (0 < seed) {
    set.seed(seed)
  }

  if (lower.tail) {
    cvID <- round(c(0.005, 0.01, 0.05, 0.10) * nexcur)
    cvnames <- c("cv005", "cv01", "cv05", "cv10")
  } else {
    cvID <- round(c(0.90, 0.95, 0.99, 0.995) * nexcur)
    cvnames <- c("cv90", "cv95", "cv99", "cv995")
  }
  cvID <- pmax(cvID, 1)
  resnames <- c("ht", "pval", cvnames)

  if ((0 == length(xbase)) || (0 == length(ht))) {
    pe <- matrix(NA_real_, nrow=length(ht), ncol=length(resnames),
                 dimnames=list(NULL, resnames))
  } else {
    pe <- sapply(seq_along(ndraw),
                 function(i) {
                   # Minimum draw size to ensure a peak goes to either side.
                   if (!is.finite(ht[i]) || 
                       !is.finite(ndraw[i]) || (ndraw[i] < 3)) {
                     p <- c(ht[i], NA_real_, rep(NA_real_, length(cvID)))
                   } else {
                     excht <- .Call("C_excursion", xbase, nexcur,ndraw[i],
                                    is.peak, TRUE, PACKAGE="Dimodal")
                     pval <- .Call("C_htprob", excht, ht[i], lower.tail,
                                   PACKAGE="Dimodal")
                     p <- c(ht[i], pval, excht[cvID])
                   }
                   names(p) <- resnames
                   p
                 })
    pe <- t(pe)
  }

  res <- list(method="excursion height test",
              statfn="Diexcurht.test", statistic=pe[,"ht"],
              statname="ht",
              parameter=list(ndraw=ndraw, nexcur=nexcur, seed=seed),
              p.value=pe[,"pval"],
              alternatives=ifelse(lower.tail, "less", "greater"),
              critval=pe[,-c(1,2), drop=FALSE], xbase=xbase)
  names(res$statistic) <- NULL
  names(res$p.value) <- NULL
  class(res) <- "Ditest"
  res
}

# Carry out nperm permutations of the values in signed difference vector
# xbase and collect the height (difference of the resulting range) for each.
# We forbid the permutation to have adjacent same-sign symbols.  The
# probability of height(s) ht against this distribution are the number of
# permutations smaller than ht if lower.tail is TRUE, or larger if FALSE.
# seed if a positive integer is used to set the RNG before the draws.
# Returns a list of class 'Ditest' with elements
#    $method             descriptive text
#    $statfn             name (text) of function used to evaluate statistic
#    $statistic          what is tested, the number of runs
#    $statname           text name of statistic
#    $parameter          function arguments stID, endID, permute, nperm, seed
#    $p.value            probability of number of runs
#    $alternative        direction of test
#    $critval            critical values at quantiles 0.10, 0.05, 0.01, 0.005
#                          (or 1-q if lower.tail TRUE)
#    $xbase              base of samples
# Statistic, p.value, critval will be NA if there are two or fewer xbase points
# or no lengths.
Dipermht.test <- function(ht, xbase, nperm, lower.tail=TRUE, seed=0) {

  nperm <- round(nperm[1])
  check.htargs(ht, FALSE, xbase, nperm, lower.tail)

  if (lower.tail) {
    qcv <- c(0.005, 0.01, 0.05, 0.10)
    cvnames <- c("cv005", "cv01", "cv05", "cv10")
  } else {
    qcv <- c(0.90, 0.95, 0.99, 0.995)
    cvnames <- c("cv90", "cv95", "cv99", "cv995")
  }

  if (0 < seed) {
    set.seed(seed)
  }

  # Minimum base size to ensure a peak goes up and down.
  # Does it make sense to impose a stronger minimum, because the resolution
  # will be too low for only a few runs, ie. 5! = 120 ~ 1%?
  if ((length(xbase) < 2) || (0 == length(ht))) {
    cval <- numeric(0)
    pe <- matrix(NA_real_, nrow=length(ht), ncol=2,
                 dimnames=list(NULL, c("ht", "pval")))
  } else {
    xgrp <- sign(xbase)
    # The alternating requirement knocks out a large fraction of the possible
    # permutations.
    if (lfactorial(length(xbase)) < (log(nperm) - log(0.05))) {
      permht <- apply(permute.all.alt(xbase), 1,
                      function(r) {
                        cp <- cumsum(r)
                        max(cp) - min(0, cp[length(cp)]) })
    } else {
      gspec <- setup.permute(xgrp)
      permht <- replicate(nperm, alt.permute(xbase, gspec))
    }
    permht <- sort(permht)

    cval <- permht[pmax(round(qcv * length(permht)), 1)]
    names(cval) <- cvnames

    pval <- sapply(ht,
                   function(h) {
                     if (is.na(h) || is.nan(h)) {
                       NA_real_
                     } else {
                       .Call("C_htprob", permht, h, lower.tail,
                             PACKAGE="Dimodal")
                     }
                   })
    pe <- cbind(ht=ht, pval=pval)
  }

  res <- list(method="run height permutation test",
              statfn="Dipermht.test", statistic=pe[,"ht"],
              statname="ht",
              parameter=list(nperm=nperm, seed=seed),
              p.value=pe[,"pval"],
              alternatives=ifelse(lower.tail, "less", "greater"),
              critval=cval, xbase=xbase)
  names(res$statistic) <- NULL
  names(res$p.value) <- NULL
  class(res) <- "Ditest"
  res
}


#### Implementation Details

# Validity check on the draw and window size arguments or lower.tail flag to
# any of the model test/critical value functions.  Bad values raise an error.
# filter and basedist have defaults.
check.modelargs <- function(n, flp, lower.tail) {

  # This is consistent with R's behavior for pnorm - non-numeric arguments
  # raise errors, logical are coerced to integers.
  if ((0 == length(n)) || !is.finite(n) ||
      (!is.numeric(n) && !is.logical(n))) {
    stop("non-numeric argument n", call.=FALSE)
  }
  if ((0 == length(flp)) || !is.finite(flp) ||
      (!is.numeric(flp) && !is.logical(flp))) {
    stop("non-numeric argument flp", call.=FALSE)
  }
  if (n < 1) {
    stop("argument n must be at least 1", call.=FALSE)
  }
  # This fails if the filter size was given absolutely and is more than
  # half the data.  You probably want to pass the fractional size.
  if ((flp <= 0) || (n < flp)) {
    stop("argument flp must be positive and less than n", call.=FALSE)
  }

  if ((0 == length(lower.tail)) || !is.logical(lower.tail)) {
    stop("non-logical argument lower.tail", call.=FALSE)
  }

  if (((flp < 1.0) && ((flp < 0.05) || (0.30 < flp))) ||
      ((n < 60) || (500 < n))) {
    warning("models used outside recommended range (flp: 0.05-0.30, n 60-500)")
  }
}

# Validity checks on the start indices stID and endpoints endID of intervals
# in a vector of length nx.  Bad values (not numeric, not within vector,
# endID < stID) raise an error.  NA and NaN values are allowed.
check.endargs <- function(stID, endID, nx) {

  if (length(stID) != length(endID)) {
    stop("start and endpoint vectors do not have same length")
  }

  if (any(is.na(stID) | is.nan(stID) | is.na(endID) | is.nan(endID))) {
    return()
  }

  if (0 < length(stID)) {
    if (any(!is.numeric(stID))) {
      stop("not all start points numeric")
    }   
    if (any((stID < 1) | (nx < stID))) {
      stop("some start points lie outside data")
    }
  }

  if (0 < length(endID)) {
    if (any(!is.numeric(endID))) {
      stop("not all endpoints numeric")
    }
    if (any((endID < 1) | (nx < endID))) {
      stop("some endpoints lie outside data")
    }
  }

  if (any(endID < stID)) {
    stop("have endpoints before start points")
  }
}

# Validity checks on the Diexcurht.test and Dipermht.test arguments.  Pass
# FALSE for ndraw for the permutation tests where it is not used.  Bad values
# (empty/non-finite/non-numeric xdraw, negative nperm, ) raise an error.
# ht and ndraw may have NA, NaN values or be an empty vector.
check.htargs <- function(ht, ndraw, xbase, nperm, lower.tail) {

  if ((0 < length(xbase)) &&
      (any(!is.integer(xbase) & !is.numeric(xbase)) ||
       any(!is.finite(xbase)))) {
    stop("xbase set is empty or has non-numeric or non-finite values")
  }

  if ((!is.numeric(nperm) && !is.integer(nperm)) || !is.finite(nperm) ||
      (nperm <= 0)) {
    stop("nperm must be positive")
  }

  if ((1 != length(ndraw)) || !is.logical(ndraw) || (FALSE != ndraw)) {
    if (length(ndraw) != length(ht)) {
      stop("draw sizes do not match heights")
    }

    if ((0 < length(ndraw)) && any(!is.integer(ndraw) & !is.numeric(ndraw))) {
      stop("non-numeric draw sizes")
    }
    if (0 < length(ndraw)) {
      valid <- is.finite(ndraw)
      if (0 < sum(valid)) {
        if (any(ndraw[valid] > length(xbase), na.rm=TRUE)) {
          stop("excursion test samples more points than available")
        }
        if (any(ndraw[valid] < 0)) {
          stop("draw sizes cannot be negative")
        }
      }
    }
  }

  if ((0 < length(ht)) && any(!is.integer(ht) & !is.numeric(ht))) {
    stop("non-numeric heights")
  }
  if (0 < length(ht)) {
    valid <- is.finite(ht)
    if (0 < sum(valid)) {
      if (any(ht[valid] < 0)) {
        stop("feature heights cannot be negative")
      }
    }
  }

  if ((0 == length(lower.tail)) || !is.logical(lower.tail)) {
    stop("non-logical argument lower.tail", call.=FALSE)
  }
}

## Height Models

# Model the correction of the local maxima height relative the FIR standard
# deviation, htsd, for the named filter.
peak.filterratio <- function(filter) {

  # There's a small variation in the correction for the window size and
  # sample draw - not modeling, 10% for n=100/500, win=0.10/0.20 combinations
  # for the first three, 25% for gauss and blackman, 40% for the last two.
  # Also a small variation in quantile.  We're picking the ratio to the Kaiser
  # filter at q=0.999; other quantiles will give larger ratios which
  # under-predicts the ht/sd.
  if ((0 == length(filter)) || is.na(filter) || !is.character(filter)) {
    warning("filter not a character string, using Kaiser")
    frat <- 1.0
  } else if (("triangular" == filter) || ("bartlett" == filter)) {
    frat <- 1.086
  } else if ("hamming" == filter) {
    frat <- 1.122
  } else if ("hanning" == filter) {
    frat <- 1.155
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    frat <- 1.239
  } else if ("blackman" == filter) {
    frat <- 1.248
  } else {
    if ("kaiser" != filter) {
      warning("unsupported filter for peak test, using Kaiser")
    }
    frat <- 1.0
  }

  frat
}

# Calculate the Wald/inverse gaussian parameters param (1 for the mu/mean,
# 2 for the lambda/shape; or by name) for the sample draw size n and
# FIR window width flp (as a fraction of n).
peak.qwald.param <- function(param, n, flp) {

  # Modeled at flp=0.15 only.
  #  p1 <- 3.8356 - (1.1459 * log10(n))
  #  p2 <- 4.762

  # Our model for either parameter is the same: linear against the log of
  # the draw size, where the slope and intercept are linear with the width.
  if ((1 == param) || ("mu" == param) || ("mean" == param)) {
    b <-  5.8158 + (2.4152 * log10(flp))
    m <- -1.9704 - (1.0131 * log10(flp))
  } else if ((2 == param) || ("lambda" == param) || ("shape" == param)) {
    b <- -2.0204 + (49.7357 * flp)
    m <-  2.6034 - (19.5195 * flp)
  }

  # Clip invalid (negative) values to 0.  Returning NA will not work
  # with pinvgauss, if the values also contain NAs.
  max(0, b + (m * log10(n)))
}

# Correction of height quantile in peak model, which depends on the
# fraction filter kernel size flp.  Returns a pair with elements $badj, $madj,
# the intercept and slope of the (linear) correction.
peak.correction <- function(flp) {

  c(badj=-0.2305 + (11.8716 * flp) - (46.9360 * (flp^2)) + (85.0096 * (flp^3)),
    madj=4.2412 - (7.2054 * flp) + (0.1547 / flp))
}

## Length Models

# Get the flat length model coefficients for a parametric draw of size n from
# distribution basedist and smoothed by an FIR kernel filter of fractional
# width flp (must be < 1).  Returns a 4 element vector that must be multiplied
# by the p terms (1, p, p^2, log(p/(1-p))) and summed to get the final
# quantile.
flat.pcoef <- function(n, flp, filter, basedist) {

  if ((0 == length(basedist)) || is.na(basedist) || !is.character(basedist)) {
    warning("basedist not a character string, using logistic")
    pm <- pcoef.length.logistic(filter)
  } else if ("weibull" == basedist) {
    pm <- pcoef.length.weibull(filter)
  } else if (("normal" == basedist) || ("gaussian" == basedist)) {
    pm <- pcoef.length.normal(filter)
  } else if ("gumbel" == basedist) {
    pm <- pcoef.length.gumbel(filter)
  } else {
    if ("logistic" != basedist) {
      warning(paste0("unsupported base distribution ", basedist,
                     ", using logistic"))
    }
    pm <- pcoef.length.logistic(filter)
  }

  nw <- c(1, n, n^2, flp, flp*n, flp*(n^2))

  pcf <- matrix(pm, nrow=4, byrow=TRUE)
  # Sanity check.
  if (ncol(pcf) != length(nw)) {
    stop("internal error - pcoef matrix does not conform to n/flp combinations")
  }

  # Collapse the model by column, leaving a 4 element vector to be multiplied
  # by the p terms.
  pcf %*% nw
}

# Return a vector with the flat length model coefficients based on a logistic
# draw smoothed by FIR kernel filter.  Columns are combinations of draw size
# and fractional kernel window size terms.  Can be cast into a 4x6 matrix by
# row.
pcoef.length.logistic <- function(filter) {

  if ((0 == length(filter)) || is.na(filter) || !is.character(filter)) {
    warning("filter not a character string, using Kaiser")
    filter <- "kaiser"
  }
  if ("hanning" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  895.07,  -6.4915,  0.050202,   -2914.0,   39.755,  -0.12859,
      -1945.1,   14.621,  -0.10957,     6365.8,  -87.586,   0.28017,
       1068.6,   -8.4033,  0.060307,   -3524.0,   49.338,  -0.15443,
      -0.80887, 0.078324, -6.8829e-5,   3.8649, -0.16948,   1.9127e-4)
  } else if ("hamming" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -1.3561,  -2.6245,   0.055274,   -1175.4,   41.627,  -0.15507,
      -45.672,    6.7682,  -0.12187,     2713.6,  -93.135,   0.34047,
       61.9459,  -4.4633,   0.067877,   -1604.3,   53.304,  -0.18934,
       -1.0098, 0.086396,  -9.7872e-5,  4.5358,  -0.19984,   2.8048e-4)
  } else if (("triangular" == filter) || ("bartlett" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -731.78,   20.902,   9.4621e-3,  1129.6,  -29.898, -0.025898,
       1552.6,   -44.270, -0.023055,   -2347.6,   62.174,  0.062428,
       -814.69,   23.292,  0.014510,    1185.1,  -31.308, -0.039256,
      -0.11633, 0.071427, -7.8624e-5,   1.9172, -0.15436,  2.168e-4)
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  2248.3,  -20.929,  0.058887,   -7656.4,   82.141,  -0.15955,
       -4859.7,   45.630, -0.12765,    16555,   -178.30,    0.34485,
        2639.1,  -25.057,  0.069560,   -8993.5,   97.708,  -0.18757,
      -0.83352, 0.074709, -4.7317e-5,   3.5018, -0.15093,   1.2987e-4)
  } else if ("blackman" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  1616.7,  -18.638,  0.051899,   -5458.4,   73.494,   -0.13673,
       -3490.0,   40.584, -0.11238,    11782.1, -159.31,     0.29516,
        1894.7,  -22.256,  0.061167,   -6397.7,   87.202,   -0.16035,
      -0.41984, 0.067871, -3.3585e-05,  2.0423,   -0.12768,  8.6115e-5)
  } else {
    if ("kaiser" != filter) {
      warning("unsupported filter ", filter, ", using kaiser")
    }
    #   const         n        n^2        flp     flp*n     flp*n^2
    c(-2389.1,   39.048, -0.025085,   5.9153e+3,  -72.984,  0.047613,
       5122.2,  -83.133,  0.050001,  -1.2677e+4,  154.64,  -0.094133,
      -2739.2,   44.077, -0.023815,   6.7609e+3,  -80.650,  0.042964,
      0.43973, 0.069628, -9.5534e-5,    0.41845, -0.15918,  2.807e-4)
  }
}

# Return a vector with the flat length model coefficients based on a Weibull
# draw smoothed by FIR kernel filter.  Can be cast into a 4x6 matrix by row.
pcoef.length.weibull <- function(filter) {

  if ((0 == length(filter)) || is.na(filter) || !is.character(filter)) {
    warning("filter not a character string, using Kaiser")
    filter <- "kaiser"
  }
  if ("hanning" == filter) {
    #   const         n        n^2        flp     flp*n      flp*n^2
    c( -4.9840,  -3.2583,   1.8028e-3,  330.12,   19.067,    0.033660,
       17.469,    7.1345,  -3.8490e-3, -711.49,  -41.233,   -0.073340,
       -6.5314,  -3.9661,   2.1845e-3,  360.79,   22.817,    0.039462,
       0.72389, 0.026799,   2.1729e-5, -2.1478,  8.0694e-3, -8.6371e-5)
  } else if ("hamming" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  86.658,  -6.2163,   9.3797e-3,  -700.22,    37.300,   9.2935e-3,
      -175.64,   13.469,   -0.020101,    1489.3,   -80.287,  -0.021620,
        93.357,  -7.3396,   0.010898,    -804.34,   43.671,   0.011922,
       0.58325, 0.027359,   2.5845e-5,  -2.0106,   0.00681,  -9.7996e-5)
  } else if (("triangular" == filter) || ("bartlett" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  133.22,  -6.5424,  0.012183,   -570.49,    36.029,  -7.5517e-4,
       -281.46,  14.191,  -0.026233,   1223.8,    -77.590,   5.4783e-4,
        154.47,  -7.7411,  0.014236,   -669.89,    42.186,   2.2254e-5,
       0.36176, 0.034808,  1.585e-5,   -1.3107, -0.015858,  -8.0323e-5)
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  81.895,  -5.3860,  4.3516e-3,  -16.150,    21.258,   0.014945,
      -160.95,   11.623,  -9.2230e-3,   24.492,   -45.817,  -0.032756,
        83.962,  -6.3206,  4.9779e-3,  -22.006,    25.059,   0.017657,
        1.3213, 0.019862,  2.0665e-5,   -3.9419, 0.026294,  -6.4009e-5)
  } else if ("blackman" == filter) {
    #    const         n        n^2        flp     flp*n    flp*n^2
    c(  -254.86,  -2.3780,  5.5634e-5,  1055.6,    12.453,   0.026062,
         551.58,   5.2308, -1.2819e-4, -2256.2,   -26.961,  -0.056475,
        -292.19,  -2.9261,  1.6688e-4,  1189.1,    14.961,   0.030318,
         1.1371, 0.019393,  1.9808e-5, -3.5074,  0.030359,  -6.5452e-5)
  } else {
    if ("kaiser" != filter) {
      warning("unsupported filter ", filter, ", using kaiser")
    }
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  185.83,  -8.9368,  0.024456,   -1864.9,    61.922,  -0.063500,
       -392.38,  19.341,  -0.052597,    3978.0,  -132.97,    0.13362,
        211.76, -10.507,   0.028442,   -2133.5,    71.896,  -0.070919,
      0.010027, 0.038499,  1.4777e-5,  0.26152, -0.035014,  -6.5996e-5)
  }
}

# Return a vector with the flat length model coefficients based on normal
# draws smoothed by FIR kernel filter.  Can be cast into a 4x6 matrix by row.
pcoef.length.normal <- function(filter) {

  if ((0 == length(filter)) || is.na(filter) || !is.character(filter)) {
    warning("filter not a character string, using Kaiser")
    filter <- "kaiser"
  }
  if ("hanning" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  209.32,  -7.3078,  0.011596,   -711.87,    36.242,  -2.1480e-3,
       -448.08,  15.925,  -0.025068,   1529.6,    -78.318,   3.8716e-3,
        246.65,  -8.7548,  0.013716,   -846.01,    42.923,  -2.2864e-3,
       0.59752, 0.036907,  1.4542e-5,   -1.604, -0.026236,  -6.7247e-5)
  } else if ("hamming" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  307.04,  -10.703,  0.022635,   -1405.5,    54.554,  -0.032498,
       -658.02,   23.252, -0.048878,    3028.0,  -117.83,    0.068855,
        356.80,  -12.685,  0.026553,   -1644.8,    64.186,  -0.037172,
       0.54728, 0.036049,  2.1316e-5,  -1.7339, -0.024355,  -8.694e-5)
  } else if (("triangular" == filter) || ("bartlett" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  443.26,  -10.033,  0.025407,    -1897.6,    55.841, -0.056668,
       -946.30,   21.779, -0.054899,     4064.4,  -120.38,   0.121160,
        510.96,  -11.885,  0.029831,    -2192.6,    65.403, -0.065230,
       0.17459, 0.046524,  4.2899e-6,  -0.34232, -0.059278, -4.144e-5)
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  262.78,  -7.8587,  8.9202e-3,  -986.59,     33.510,  -4.2118e-3,
       -550.32,  16.984,  -0.019065,   2093.2,     -72.079,   7.9891e-3,
        292.97,  -9.2384,  0.010298,  -1123.7,      39.192,  -4.0579e-3,
        1.6729, 0.025024,  2.2439e-5,  -4.9313,   8.3653e-3, -7.3441e-5)
  } else if ("blackman" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  74.431,  -6.1651,  6.0550e-3,   7.4568,     24.212,  0.011533,
      -145.40,   13.308,  -0.012853,  -34.403,    -52.049,  -0.025881,
        76.036,  -7.2398,  6.9213e-3,  12.338,     28.388,   0.014190,
        1.2774, 0.024892,  2.1922e-5,  -3.7843,  0.011591,  -7.6688e-5)
  } else {
    if ("kaiser" != filter) {
      warning("unsupported filter ", filter, ", using kaiser")
    }
    #   const         n        n^2        flp     flp*n    flp*n^2
    c ( 741.06,  -17.064,  0.053711,   -3807.4,    93.135,  -0.16180,
      -1593.3,    36.962, -0.11593,     8194.4,   -200.65,   0.34606,
        861.61,  -20.087,  0.062820,   -4425.2,    108.73,  -0.18599,
      -0.56429, 0.055043, -1.137e-5,    2.2923, -0.089751,   8.0502e-6)
  }
}

# Return a vector with the flat length model coefficients based on a Gumbel
# draw smoothed by FIR knerel filter.  Can be cast into a 4x6 matrix by row.
pcoef.length.gumbel <- function(filter) {

  if ((0 == length(filter)) || is.na(filter) || !is.character(filter)) {
    warning("filter not a character string, using Kaiser")
    filter <- "kaiser"
  }
  if ("hanning" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -1598.0,   47.873, -0.032367,    4161.1,  -106.21,  0.073324,
        3410.9, -102.175,  0.065916,   -8822.6,   225.65, -0.147592,
       -1804.3,   54.258, -0.032486,    4613.0,  -118.40,  0.071224,
      -0.23631, 0.075773, -9.6954e-5,   2.6738, -0.16307,  2.5834e-4 )
  } else if ("hamming" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -4048.0,   77.083,  -0.10121,      11980,  -197.31,   0.27878,
        8685.7, -164.802,   0.21272,     -25613,   420.27,  -0.58524,
       -4646.0,   87.778,  -0.11039,      13626,  -221.98,   0.30297,
        1.0524,  0.06882,  -1.0146e-4, -0.56980, -0.16052,   2.9109e-4 )
  } else if (("triangular" == filter) || ("bartlett" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -2278.6,   74.222,  -0.11808,    6609.5,   -188.44,   0.32789,
        4964.3, -160.127,   0.25141,  -14335.1,    405.65,  -0.69687,
       -2698.3,   86.257,  -0.13276,    7743.4,   -217.24,   0.36735,
       1.6654, 0.044069,  -5.5019e-5, -2.9821, -0.079026,   1.3708e-4 )
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  18.961, -0.26832,   0.066611,   -1454.5,   39.102,  -0.21431,
       -79.873,  1.68360,  -0.146120,    3278.1,  -87.435,   0.46803,
        80.357, -1.79288,   0.080824,   -1900.2,   50.104,  -0.25722,
       -1.2499, 0.095639,  -1.0463e-4,   4.8565, -0.20499,   2.6783e-4 )
  } else if ("blackman" == filter) {
    #   const         n        n^2        flp     flp*n    flp*n^2
    c(  148.96,   13.913,   0.038246,   -1735.7,  -6.8326,  -0.12234,
       -381.39,  -28.733,  -0.085554,    3944.8,  11.1196,   0.27155,
        256.95,   14.512,   0.048553,   -2302.8,  -2.6911,  -0.15256,
       -1.4974, 0.093077,  -1.0534e-4,   6.0483,  -0.20031,  2.74553e-4 )
  } else {
    if ("kaiser" != filter) {
      warning("unsupported filter ", filter, ", using kaiser")
    }
    #   const         n        n^2        flp     flp*n    flp*n^2
    c( -2286.5,   77.418,  -0.14717,    7699.7,   -209.25,   0.42424,
        5091.1, -168.237,   0.31626,  -17056.9,    454.52,  -0.91247,
       -2834.9,   91.323,  -0.16862,    9425.3,   -245.53,   0.48656,
       3.0400, 0.031533,  -4.2934e-5, -6.9384, -0.056694,   1.24383e-4 )
  }

}

# Find the location of the minimum of our flat model.  pcf is the matrix
# returned from flat.pcoef.  Return NA if there is no minimum, or the largest
# value below 1 if there are multiple zeroes.
flat.pmin <- function(pcf) {

  # Our model is
  #  pcf[1] * (p * pcf[2]) + (p^2 * pcf[3]) * (log(p/(1-p)) * pcf[4])
  # and the gradient is
  #  pcf[2] + (2 * p * pcf[3]) + ((1/p) * (1/(1-p)) * pcf[4])
  # Extrema are located where gradient == 0, which is a cubic equation and
  # can be solved directly rather than by a root finder.  The solution is
  # based on the Wikipedia cubic equation page.

  if ((0 == pcf[2]) && (0 == pcf[3])) {
    exp(-pcf[1]/pcf[4]) / (1.0 + exp(-pcf[1]/pcf[4]))
  } else if (0 == pcf[3]) {
    (1.0 + sqrt(1.0 + (4.0 * pcf[4] / pcf[2]))) / 2.0
  } else {
    B <- pcf[2] / (2.0 * pcf[3])
    D <- pcf[4] / (2.0 * pcf[3])
    E <- -((B^2) + B + 1.0) / 3.0
    F <- ((2.0*(B^3)) + (3.0*(B^2)) - (3.0*B) - (27.0*D) - 2.0) / 27.0
    delta <- ((E^3) / 27.0) + ((F^2) / 4.0)
    # Treat values close to zero as matching
    eps <- 1.0e-10
    if (eps < delta) {
      # Other two roots are imaginary.
      t0 <- cbrt((-F/2.0) * sqrt(delta)) + cbrt((-F/2.0) - sqrt(delta))
    } else if (delta < -eps) {
      phis <- (acos(((3 * F) / (2 * E)) * sqrt(-3/E)) - (2 * (0:2) * pi)) / 3
      t0 <- 2.0 * sqrt(-E/3.0) * cos(phis)
    } else {
      t0 <- c((3.0 * F) / E, (-3.0 * F) / (2.0 * E))
    }
  }

  # Shift back to p.
  t0 <- t0 - ((B - 1.0) / 3.0)
  t0 <- t0[is.finite(t0)]
  if ((0 == length(t0)) || (0 == sum((0 <= t0) & (t0 < 1)))) {
    NA_real_
  } else {
    max(t0[t0<1])
  }
}

# Wrapper around uniroot that returns error.value for $root instead of stopping
# if there's a problem finding a root.  Arguments passed through to uniroot.
safe.uniroot <- function(..., error.value=NA_real_) {
  dummy <- list(root=error.value, f.root=NULL, iter=NA, init.it=NA,
                estim.prec=NA)
  
  tryCatch(uniroot(...),
           error=function(e) { dummy }, warning=function(w) { dummy })
}

# Return the cube root of each x, which may be negative, unlike x^(1/3).
cbrt <- function(x) {

  sign(x) * (abs(x)^(1/3))
}

## Markov Chains

# Build the transition matrix for a first order Markov chain along data x
# for the unique values lvls, which can be a superset of what is present in
# x (in case some values are missing in a subset).  If rel is TRUE normalize
# each row to sum to 1, except if there are no transitions to the final
# state along it, where the sum will be 0, otherwise return the raw
# transition counts.  The transition matrix is #lvls^r x #lvls.
mkv.transition <- function(x, lvls=sort(unique(x), na.last=NA), rel=TRUE) {

  x <- x[is.finite(x)]

  nlvl <- length(lvls)
  
  # Guarantee that each possible combination is counted at least once.
  # The vectors are length nlvl^2.
  pairst <- sapply(lvls, rep, nlvl, simplify="vector")
  pairend <- sapply(1:nlvl, function(i) { lvls }, simplify="vector")
  tmat <- table(start=c(x[-length(x)], NA, pairst),
                  end=c(x[-1L], NA, pairend))
  # The pairs mean tmat is nlvl x nlvl.  Values in x not in lvls increase
  # the size.
  if ((nlvl != nrow(tmat)) || (nlvl != ncol(tmat))) {
    stop("data contains unrecognized levels")
  }
  tmat <- tmat - 1

  if (rel) {
    tmat <- tmat / rowSums(tmat)
    tmat[!is.finite(tmat)] <- 0
  }

  tmat
}

# Return the steady-state state vector for Markov transition matrix tmat.
# If none is found, return a vector of NAs.  
mkv.stationary <- function(tmat) {

  ttmat <- t(tmat)
  ev <- eigen(ttmat)

  # Stationary state is not valid if largest left eigenvalue is not 1.
  if (1e-4 < abs(Mod(ev$values[1]) - 1)) {
    dummy <- rep(NA_real_, ncol(tmat))
    names(dummy) <- colnames(tmat)
    dummy
  } else {
    smat <- ttmat %*% ev$vectors
    scl <- Mod(colSums(smat))

    # Or if we cannot normalize.  In case of tied eigenvalues we need to
    # check all corresponding eigenvectors to see if any are good =
    # eigenvalue at 1 and sum of stationary state not 0.  Note that the
    # scale is unsigned but the final normalization is signed.
    valid <- (Mod(ev$values) - 1) < 1e-4
    i <- which.max(scl)
    if (!valid[i] || (scl[i] < 1e-4)) {
      dummy <- rep(NA_real_, ncol(tmat))
      names(dummy) <- colnames(tmat)
      dummy
    } else {
      Re(smat[,i] / sum(smat[,i]))
    }
  }
}

# Return the probability that all runs are <= runlen in length for
# trajectories of featlen steps through a Markov chain with transition
# matrix tmat.  Consider all paths starting from state(s) i; if more than
# one, scale by weight wt (stationary matrix or 1/nrow(tmat) for uniform
# starts).  If transition matrix is bad (b.v. only one state) return NA.
test.markov.runlen <- function(runlen, featlen, tmat, wt, i=1:nrow(tmat)) {

  runlen <- runlen[1L]
  featlen <- featlen[1L]

  if (!is.finite(runlen) || !is.finite(featlen)) {
    return(NA_real_)
  }

  if ((ncol(tmat) != nrow(tmat)) ||
      !all(abs(rowSums(tmat) - 1.0) < 2*.Machine$double.eps)) {
    warning("bad transition matrix in longest runs test")
    return(NA_real_)
  }
  if (nrow(tmat) != length(wt)) {
    warning("weight/stationary vector incompatible with transition matrix")
    return(NA_real_)
  }

  # Comment out this if you want the original, slower R implementation.
  return( .Call("C_test_runlen", runlen, featlen, tmat, wt, PACKAGE="Dimodal") )

  if ((runlen < 1) || (featlen < 1)) {
    return(0)
  } else if (featlen <= runlen) {
    return(1)
  }

  # A advances to the next run length.  R resets back to len=1.
  A <- diag(diag(tmat))
  R <- tmat - A
  
  recur <- array(0, c(runlen, nrow(tmat), ncol(tmat)))
  recur[runlen,,] <- A

  if (2 <= featlen) {
    for (t in 2:featlen) {
      top <- R %*% recur[1,,]
      prev <- recur[runlen,,]
      recur[runlen,,] <- top + A
      if (1 < runlen) {
        for (j in (runlen-1):1) {
          curr <- top + (A %*% prev)
          prev <- recur[j,,]
          recur[j,,] <- curr
        }
      }
    }
  }

  plong <- rowSums(recur[1,,])[i]

  if (1 < length(i)) {
    1 - sum(plong * wt)
  } else {
    1 - plong
  }
}

## Runs Permutations

# Separate the group identifiers grp (any class, we take unique values) in
# preparation for a permutation that separates them.  Returns a list with
#   $grpval      unique identifiers
#   $grpID       position in grp of each identifier, in grpval order
#   $ngrp        population of each symbol
setup.permute <- function(grp) {

  # Doing this once before the permutations, rather than at the top of
  # alt.permute, cuts the run time roughly in half, from 4.2 s to 2.2 s
  # for the test bench.
  
  grpval <- sort(unique(grp))
  grpID <- lapply(grpval, function(g) { which(grp == g) })
  ngrp <- sapply(grpID, length)

  list(grpval=grpval, grpID=grpID, ngrp=ngrp)
}

# Return the height of a permutation, or the permutation itself if raw TRUE,
# of the values in x, with the added condition that no two adjacent will
# belong to the same group, described by the setup.permute result grpspec.
alt.permute <- function(x, grpspec, raw=FALSE) {

  grpval <- grpspec$grpval
  # Scramble order within each group.
  grpID <- lapply(grpspec$grpID, function(g) { g[sample.int(length(g))] })
  ngrp <- grpspec$ngrp
  
  perm <- integer(length(x))
  
  if (1 == length(grpval)) {
    perm <- x[grpID[[1]]]
  } else {
    sID <- .Call("C_altperm_symbols", ngrp, PACKAGE="Dimodal")
    for (i in seq_along(grpval)) {
      perm[sID==i] <- x[grpID[[i]]]
    }
  }

  if (raw) {
    perm
  } else {
    # This is the old height, suitable for flats.
    # diff(range(cumsum( perm )))
    cp <- cumsum(perm)
    max(cp) - min(0, cp[length(cp)])
  }
}

# Exhaustively permute the elements of vector x, deleting any where the sign
# of adjacent elements is the same.  Returns a matrix with length(x) columns
# and permutations in rows.
permute.all.alt <- function(x) {

  perm <- permute.all(x)
  dropcol <- -ncol(perm)
  keep <- apply(perm, 1, function(r) { all(sign(r[-1L]) != sign(r[dropcol])) })
  perm[keep,,drop=FALSE]
}

# Exhaustively permute the elements of vector x.  Returns a matrix with
# length(x)! rows and length(x) columns.
permute.all <- function(x) {

  if (4 == length(x)) {
    matrix(c(x[1],x[2],x[3],x[4], x[1],x[2],x[4],x[3], x[1],x[3],x[2],x[4],
             x[1],x[3],x[4],x[2], x[1],x[4],x[2],x[3], x[1],x[4],x[3],x[2],
             x[2],x[1],x[3],x[4], x[2],x[1],x[4],x[3], x[2],x[3],x[1],x[4],
             x[2],x[3],x[4],x[1], x[2],x[4],x[1],x[3], x[2],x[4],x[3],x[1],
             x[3],x[1],x[2],x[4], x[3],x[1],x[4],x[2], x[3],x[2],x[1],x[4],
             x[3],x[2],x[4],x[1], x[3],x[4],x[1],x[2], x[3],x[4],x[2],x[1],
             x[4],x[1],x[2],x[3], x[4],x[1],x[3],x[2], x[4],x[2],x[1],x[3],
             x[4],x[2],x[3],x[1], x[4],x[3],x[1],x[2], x[4],x[3],x[2],x[1]),
             nrow=24, byrow=TRUE)
  } else if (3 == length(x)) {
    matrix(c(x[1],x[2],x[3], x[1],x[3],x[2],
             x[2],x[1],x[3], x[2],x[3],x[1],
             x[3],x[1],x[2], x[3],x[2],x[1]), nrow=6, byrow=TRUE)
  } else if (2 == length(x)) {
    matrix(c(x[1],x[2], x[2],x[1]), nrow=2, byrow=TRUE)
  } else if (1 == length(x)) {
    matrix(x[1], nrow=1, ncol=1)
  } else if (0 < length(x)) {
    # Why does sapply or simplify2array not bind the sub-matrices correctly?
    # Ah well, do it ourselves.
    sub <- lapply(1:length(x),
                  function(i) {
                    sub <- permute.all(x[-i])
                    cbind(rep(x[i], nrow(sub)), sub)
                  })
    perm <- sub[[1]]
    for (i in 2:length(x)) {
      perm <- rbind(perm, sub[[i]])
    }
    perm
  } else {
    stop("no permutations possible for 0 length vector")
  }
}

# Ditrack.R:
# Mode tree / SiZer type tracking of feature significance as the kernel size/
# interval width varies.
#
# c 2025-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
# - 

## Manually Tested With:           LP      Diw
#    faithful$eruptions            p,f     p,f
#    faithful$waiting              p,f     p,f
#    iris$Petal.Length             p,f      p      LP: warns small kernel
#    iris$Petal.Width               p       p      LP: warns small kernel
#    quakes[,3]                    p,f     p,f     Diw graphs slant right
#    chickwts$weight                p       p      LP: warns small kernel
#    tstds1                        p,f     p,f
#    tstds2                        p,f     p,f
#    tstds3                        p,f     p,f
#    tstds4                        p,f     p,f
#    meuse$cadmium  (sp)           p,f     p,f     LP: warns small kernel
#    meuse$copper   (sp)           p,f     p,f     LP: warns small kernel
#    paulsen$y      (boot)         p,f     p,f
#    statfaculty    (diptest)       p       p      LP: warns small kernel
#    Mammals$speed  (quantreg)     p,f      p      LP: warns small kernel
#                                                  LP flat at only one window
#    stamps         (multimode)    p,f     p,f
#    galaxy         (multimode)    p,f     p,f     Diw flat plot crashes scale
#    acicdity       (multimode)    p,f     p,f     LP: warns small kernel
#    enzyme         (multimode)    p,f     p,f     LP: warns small kernel
#              



##### Public Interface

# Track the peaks and flats in data x in smooth spacing while the filter or
# interval size changes from 0.01 t/m minwindow.  All analysis is done with
# the opt options list; the [lp|diw].window value is overwritten and should
# not appear in the param option.  Returns a list of class 'Ditrack' with
# elements
#   $peaks   subset of Dipeak data frame w/ pos, ismax, ppeak, plus winpct
#   $flats   subset of Diflat data w/ stID, endID, pflat, plus winpct
#   $nx      length of the data x
#   $smooth  filter applied, 'lp' or 'diw'
# The winpct column is window/interval size as a percent, ie. 1 t/m 40.
# There is one row in peaks and flats for each feature, thus multiple rows
# per winpct.
Ditrack <- function(x, smooth=c("lp", "diw"), opt=Diopt()) {

  smooth <- match.arg(smooth)
  opt$analysis <- smooth

  pkcol <- c("pos", "ismax", "ppeak", "naccept")
  ftcol <- c("stID", "endID", "pflat", "naccept")

  # Other ideas from development for data and plotting:
  #  Add a minimum-minimum width column (rminID-lminID+1) to pk.
  #  Copy the htsd column to pk.
  #  Generate graph of matching features, ie. trace features as window changes.
  #  Measure depth along graph.
  maxwin <- opt$track.maxwindow
  if (maxwin < 1) {
    maxwin <- round(maxwin * 100)
  }
  feat <- lapply(1:maxwin,
                 function(i) {
                   if ("lp" == smooth) {
                     opt$lp.window <- 0.01 * i
                   } else if ("diw" == smooth) {
                     opt$diw.window <- 0.01 * i
                   } else {
                     stop("unsupported smoothing method")
                   }
                   
                   # We could try to screen specific warnings by using
                   # warning(warningCondition(msg, class='warn.track'))
                   # and suppressing by class, but there are several possible,
                   # including model OOB and kernel too small, so just ignore
                   # everything.
                   m <- suppressWarnings(Dimodal(x, opt))
                   if ("lp" == smooth) {
                     peaks <- m$lp.peaks
                     flats <- m$lp.flats
                   } else if ("diw" == smooth) {
                     peaks <- m$diw.peaks
                     flats <- m$diw.flats
                   } else {
                     stop("unsupported smoothing method")
                   }
                  
                   npk <- nrow(peaks)
                   if (2 < npk) {
                     pk <- peaks[-c(1,npk), pkcol, drop=FALSE]
                   } else {
                     pk <- matrix(0, nrow=0, ncol=length(pkcol),
                                  dimnames=list(NULL, pkcol))
                   }
                   pk <- cbind(winpct=rep(i, nrow(pk)), pk)
                   pk <- as.matrix(pk)
                   # Named by row in [lp|diw].peaks, copied into pk.  Ignore.
                   rownames(pk) <- NULL
                  
                   if (0 < nrow(flats)) {
                     ft <- flats[,ftcol, drop=FALSE]
                   } else {
                     ft <- matrix(0, nrow=0, ncol=length(ftcol),
                                  dimnames=list(NULL, ftcol))
                   }
                   ft <- cbind(winpct=rep(i, nrow(flats)), ft)
                   ft <- as.matrix(ft)
                   rownames(ft) <- NULL
                  
                   list(pk=pk, ft=ft)
                 })

  pk <- NULL
  ft <- NULL
  for (i in seq_along(feat)) {
    pk <- rbind(pk, feat[[i]]$pk)
    ft <- rbind(ft, feat[[i]]$ft)
  }
  
  trk <- list(peaks=pk, flats=ft, nx=length(x), smooth=smooth)
  class(trk) <- c("Ditrack", class(trk))
  trk
}

# Print a table of the number of significant features as the window size
# varies in Ditrack result x.  Extra arguments are ignored.  Returns
# x invisibly.
print.Ditrack <- function(x, ...) {

  if ((0 == nrow(x$peaks)) && (0 == nrow(x$flats))) {
    cat("\nno peaks or flats at any window size\n\n")
    return(invisible(x))
  }
  if ((5 != ncol(x$peaks)) || (5 != ncol(x$flats))) {
    cat("\ninternal error - bad Ditrack contents\n\n")
    return(invisible(x))
  }

  if (0 < nrow(x$peaks)) {
    pkwrng <- range(x$peaks[,"winpct"])
    pkcnt <- sapply(pkwrng[1]:pkwrng[2],
                    function(i) {
                      sel <- x$peaks[,"winpct"] == i
                      c(sum(x$peaks[sel,"ppeak"] <= 0.01, na.rm=TRUE),
                        sum(x$peaks[sel,"ppeak"] <= 0.05, na.rm=TRUE),
                        sum(x$peaks[sel,"naccept"]))
                    }, simplify="array")
  } else {
    pkwrng <- NULL
    pkcnt <- NULL
  }

  if (0 < nrow(x$flats)) {
    ftwrng <- range(x$flats[,"winpct"])
    ftcnt <- sapply(ftwrng[1]:ftwrng[2],
                    function(i) {
                      sel <- x$flats[,"winpct"] == i
                      c(sum(x$flats[sel,"pflat"] <= 0.01, na.rm=TRUE),
                        sum(x$flats[sel,"pflat"] <= 0.05, na.rm=TRUE),
                        sum(x$flats[sel,"naccept"]))
                    })
  } else {
    ftwrng <- NULL
    ftcnt <- NULL
  }

  if ("lp" == x$smooth) {
    wtxt <- "LP kernel size"
  } else if ("diw" == x$smooth) {
    wtxt <- "interval size"
  } else {
    stop("unsupported method generated features")
  }
  wtxt2 <- "at level"
  wwin <- max(nchar(wtxt), nchar(wtxt2))
  wcnts <- 16

  ptxt <- ifelse(is.null(pkcnt), "", format("peaks", width=wcnts, justify="c"))
  ftxt <- ifelse(is.null(ftcnt), "", format("flats", width=wcnts, justify="c"))
  cat("\n")
  cat("Significant Feature Counts\n")
  cat(sprintf("%*s    %*s    %*s\n", wwin, "", wcnts, ptxt, wcnts, ftxt))
  if (is.null(pkwrng)) {
    pk01 <- ""
    pk05 <- ""
  } else {
    pk01 <- "0.01"
    pk05 <- "0.05"
  }
  if (is.null(ftwrng)) {
    ft01 <- ""
    ft05 <- ""
  } else {
    ft01 <- "0.01"
    ft05 <- "0.05"
  }
  cat(sprintf("%*s    %4s  %4s  %4s    %4s  %4s  %4s\n",
              wwin, wtxt2, pk01, pk05, "pass", ft01, ft05, "pass"))
  cat(sprintf("%*s\n", wwin, wtxt))
  
  for (i in min(pkwrng[1],ftwrng[1]):max(pkwrng[2],ftwrng[2])) {
    if (is.null(pkwrng) || (i < pkwrng[1]) || (pkwrng[2] < i)) {
      pk01 <- ""
      pk05 <- ""
      pkpass <- ""
    } else {
      pk01 <- sprintf("%4d", pkcnt[1,i-pkwrng[1]+1])
      pk05 <- sprintf("%4d", pkcnt[2,i-pkwrng[1]+1])
      pkpass <- sprintf("%4d", pkcnt[3,i-pkwrng[1]+1])
    }
    if (is.null(ftwrng) || (i < ftwrng[1]) || (ftwrng[2] < i)) {
      ft01 <- ""
      ft05 <- ""
      ftpass <- ""
    } else {
      ft01 <- sprintf("%4d", ftcnt[1,i-ftwrng[1]+1])
      ft05 <- sprintf("%4d", ftcnt[2,i-ftwrng[1]+1])
      ftpass <- sprintf("%4d", ftcnt[3,i-ftwrng[1]+1])
    }
    
    cat(sprintf("%*d    %4s  %4s  %4s    %4s  %4s  %4s\n",
                wwin,i, pk01,pk05,pkpass, ft01,ft05,ftpass))
  }
  cat("\n")

  invisible(x)
}

# Summarize the tracking results object by printing the ranges of window sizes
# with accepted features, for both peaks and flats.  Extra arguments ignored.
summary.Ditrack <- function(object, ...) {

  cat("\n")
  if ("lp" == object$smooth) {
    cat("Window Sizes (lp.window) With Accepted Features\n")
  } else if ("diw" == object$smooth) {
    cat("Interval Sizes (diw.window) With Accepted Features\n")
  } else {
    stop("unsupported method generated filters")
  }

  cat("  peaks   ", paste0(pass.range(object$peaks), sep="   "), "\n")
  cat("  flats   ", paste0(pass.range(object$flats), sep="   "), "\n")
  cat("\n")

  invisible(object)
}

# Return a vector of strings with the continguous ranges of window sizes
# that have more than one passing (naccept > 0) feature for Ditrack
# peaks/flats tracker x.
pass.range <- function(x) {

  if (0 == nrow(x)) {
    return(c("none"))
  }
  
  npass <- c(0, sign(sapply(split(x[,'naccept'], x[,'winpct']), sum)), 0)
  st <- names(npass)[which(npass[-1L]>npass[-length(npass)])+1]
  end <- names(npass)[which(npass[-1L]<npass[-length(npass)])]

  if (length(st) < length(end)) {
    st <- c(st, rep(NA, length(end)-length(st)))
  } else if (length(end) < length(st)) {
    end <- c(st, rep(NA, length(st)-length(end)))
  }

  st <- as.integer(st)
  end <- as.integer(end)

  sapply(1:length(st),
         function(i) { sprintf("%4.2f-%4.2f", 0.01*st[i],0.01*end[i]) })
}

# Graph the peaks and/or flats as the low-pass filter/interval size changes,
# as stored in Ditrack object x (feature may be a vector), using parameters
# in options opt.  Extra arguments are ignored.  Returns x invisibly.  The
# function uses layout to place the plots, even if there is only one feature.
plot.Ditrack <- function(x, feature=c("peaks", "flats"), opt=Diopt(), ...) {

  feature <- match.arg(feature, several.ok=TRUE)

  layout(matrix(1:length(feature), nrow=1))
  # layout for a 3 grid, as in the Dimodal plot, forces cex to 0.66.
  # Duplicate that here.
  oldpar <- par(cex=0.66)
  on.exit(par(oldpar))
  
  maxwin <- opt$track.maxwindow
  if (maxwin < 1) {
    maxwin <- round(maxwin * 100)
  }

  if ("peaks" %in% feature) {
    plot_pktrack(x$peaks, x$nx, x$smooth, maxwin)
  }
  if ("flats" %in% feature) {
    plot_fttrack(x$flats, x$nx, x$smooth, maxwin)
  }

  invisible(x)
}


##### Implementation

# Plot the location of peaks and minima as the smooth filter width varies, as
# stored in the combined matrix pk over windows from 1 t/m maxwin.  nx is the
# number of points in the original data.
plot_pktrack <- function(pk, nx, smooth, maxwin) {

  if (is.null(pk) || (0 == nrow(pk)) || (5 != ncol(pk))) {
    dummyplot("no peaks found")
    return()
  }

  oldpar <- par(mar=c(5,4,4,4)+0.1)
  on.exit(par(oldpar))

  # Set up coloring of ppeak.  Color nbreak+1 will be for minima.
  nbreak <- 5
  breaks <- c(0, 0.005, 0.01, 0.05, 0.10, 1.0)
  brkID <- cut(pk[,"ppeak"], breaks, labels=FALSE, include.lowest=TRUE)
  brkID[is.na(brkID)] <- nbreak + 1
  ptcex <- seq(1.25, 0.75, len=nbreak)
  ptcex <- c(ptcex, 1.10)

  clrs <- track.pcol(nbreak)
  
  # ASCII 45 is '-'.  19/20 are larger/smaller dots.
  pch <- ifelse(is.na(pk[,"ppeak"]), 45, 20)
  pch <- sapply(1:nrow(pk),
                function(i) {
                  if (is.na(pk[i,"ppeak"])) {
                    45
                  } else if (0 == pk[i,"naccept"]) {
                    1
                  } else {
                    20
                  }
                })

  if ("lp" == smooth) {
    ylbl <- "low-pass kernel size"
  } else if ("diw" == smooth) {
    ylbl <- "interval width"
  } else {
    stop("unsupported smooth generated peaks")
  }
  
  plot(0,0, col=NA, xlim=c(0,nx), ylim=c(0,maxwin+1),
       xlab="index", ylab=ylbl, main="Peak Track\n", yaxt="n")
  abline(h=seq(5, maxwin, by=5), col="grey50", lty=3)

  ytick <- seq(0, maxwin, by=10)
  axis(2, at=ytick, labels=round(0.01 * ytick * nx))
  axis(4, at=ytick, labels=0.01*ytick)

  for (i in 1:nrow(pk)) {
    points(pk[i,"pos"], pk[i,"winpct"], col=clrs[brkID[i]], pch=pch[i],
           cex=ptcex[brkID[i]]*ifelse(1==pch[i],0.75,1.0))
  }

  xl <- c(0.1, 0.3, 0.5, 0.7, 0.9)
  dxl <- 0.06 / 2
  yl <- 1.04 * (maxwin + 1)
  rect(nx*(xl-dxl), yl-1, nx*(xl+dxl), yl, col=clrs, border=NA)
  mtext(c("0.005", "0.01", "0.05", "0.10", "1.0"), at=nx*xl, cex=0.75)
  mtext("ppeak <=", line=1, at=nx*xl[1], cex=0.75)
}

# Plot the spans of flats as the smooth filter window varies, as stored in
# the combined matrix ft over windows from 1 t/m maxwin, coloring by the best
# probability for the feature.  nx is the number of points in the original
# data.
plot_fttrack <- function(ft, nx, smooth, maxwin) {

  if (is.null(ft) || (0 == nrow(ft)) || (5 != ncol(ft))) {
    dummyplot("no flats found")
    return()
  }

  oldpar <- par(mar=c(5,4,4,4)+0.1)
  on.exit(par(oldpar))

  nbreak <- 5
  breaks <- c(0, 0.005, 0.01, 0.05, 0.10, 1.0)
  brkID <- cut(ft[,"pflat"], breaks, labels=FALSE, include.lowest=TRUE)
  lwd <- seq(2.5, 1.25, len=nbreak)

  clrs <- track.pcol(nbreak)
 
  if ("lp" == smooth) {
    ylbl <- "low-pass kernel size"
  } else if ("diw" == smooth) {
    ylbl <- "interval width"
  } else {
    stop("unsupported smooth generated flats")
  }
  
  plot(0,0, col=NA, xlim=c(0,nx), ylim=c(0,maxwin+1),
       xlab="index", ylab=ylbl, main="Flat Track\n", yaxt="n")
  abline(h=seq(5, maxwin, by=5), col="grey50", lty=3)

  ytick <- seq(0, maxwin, by=10)
  axis(2, at=ytick, labels=round(0.01 * ytick * nx))
  axis(4, at=ytick, labels=0.01*ytick)

  if (1 == nrow(ft)) {
    # cbind below would turn ftplt into a vector.
    ftplt <- matrix(c(ft, brkID[1]), nrow=1,
                    dimnames=list(NULL, c(colnames(ft), "clrID")))
  } else {
    # In case of overlapping flats, draw more significant on top = later.
    # Not going to worry about marking accepted flats for this reason.
    o <- order(ft[,"pflat"], decreasing=TRUE)
    # If you pass vectors as x,y, ex. ft[o,c('stID','winpct')] and endID,winpct,
    # then lines will connect between flats, ie. must plot one row at a time.
    ftplt <- cbind(ft[o,], clrID=brkID[o])
  }
  apply(ftplt, 1,
        function(r) {
          y <- r["winpct"]
          lines(r[c("stID","endID")], c(y,y), col=clrs[r["clrID"]],
                lwd=lwd[r["clrID"]]) })

  xl <- c(0.1, 0.3, 0.5, 0.7, 0.9)
  dxl <- 0.06 / 2
  yl <- 1.04 * (maxwin + 1)
  rect(nx*(xl-dxl), yl-1, nx*(xl+dxl), yl, col=clrs, border=NA)
  mtext(c("0.005", "0.01", "0.05", "0.10", "1.0"), at=nx*xl, cex=0.75)
  mtext("pflat <=", line=1, at=nx*xl[1], cex=0.75)
}

# Generate a list of nclr+1 colors for plotting the p values from the tracks.
# The palette choice is fixed (Plasma, with the last color from YlOrRed).
track.pcol <- function(nclr) {

 clrs <- hcl.colors(nclr, "Plasma")
 clrs[nclr] <- hcl.colors(nclr+1, "YlOrRd")[nclr]
 # Last color duplicated so minima at least significant value.
 c(clrs, clrs[length(clrs)])
}

# Create an empty plot with the text message msg in the center.
dummyplot <- function(msg) {

  plot(0,0, col=NA, xlim=c(0,1), ylim=c(0,1), ann=FALSE, axes=FALSE)
  box()
  text(0.5,0.5, msg, adj=c(0.5,0.5), cex=1)
}


# Plot all HCL palettes pals with n colors, to pick a good gradient
# for the p values.  Can pass colorspace::[deu|tri|pro]tan to see how
# they look for the color blind.
plot_hclpals <- function(n, cblindfn=identity, pals=hcl.pals("sequential")) {

  # Interesting choice:  Plasma, Inferno, Viridis, Ballow, ag_Sunset

  pals <- sort(pals)
  ncol <- ceiling(sqrt(length(pals)))
  nrow <- ceiling(length(pals) / ncol)
  clrs <- sapply(pals, function(p) { cblindfn(hcl.colors(p, n=n)) })

  plot(0,0, col=NA, xlim=c(0,ncol), ylim=c(0,nrow), ann=FALSE, axes=FALSE)
  rID <- 0
  cID <- 0
  for (i in seq_along(pals)) {
    xl <- cID + (0:(n-1) / n)
    xr <- cID + ((1:n) / n)
    rect(xl, rID+0.1, xr, rID+0.7, col=clrs[,i], border=clrs[,i])
    rect(xl[1], rID+0.1, xr[n], rID+0.7, col=NA, border="white")
    text(cID+0.5, rID+0.85, pals[i], adj=c(0.5,0.5), cex=1.5)
    cID <- cID + 1
    if (cID == ncol) {
      cID <- 0
      rID <- rID + 1
    }
  }
}

# Diopt:
# Gathering of analysis and test parameters for the Dimodal package.  Modeled
# on the par() options storage.
#
# c 2024-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
# - 
#


##### Public Interface

# Diopt()
#   returns current db
# Diopt(NULL)
#   returns current db, resets db to default
# Diopt(tag1=val1, tag2=val2, ...)
#   changes db with valid values, returns old values
# Diopt(list(tag1=val1, tag2=val2, ...))
#   changes db with values valids, returns old values
# Diopt("tag1", "tag2", ...)
#   returns current db of tags
# Diopt.local(tag1=val1, tag2=val2, ...)
#   returns current db overridden by valid values, no change to db
# Diopt.local(list(tag1=val1, tag2=val2, ...))
#   returns current db overridden by valid values, no change to db


# Access or change the analysis and test options.  If called without arguments
# return a list of all current values.  If called with NULL as the argument,
# reset all values to their default.  If called with strings, return the
# values for those tags, or NULL if the tag does not exist.  If called
# with tag=value pairs, or such pairs wrapped in an unnamed list, change
# the option and return the old value.  Can mix strings and pairs.
# Diopt(Diopt(tag1=val1, tag2=val2)) will restore the original state.
# Maintains the default option values as global state, but the library will
# use a local copy during an analysis.
Diopt <- function(...) {

  args <- list(...)
  if (0 == length(args)) {
    get("diopt", envir=diopt.env)
  } else if ((1 == length(args)) && is.null(args[[1]])) {
    opt <- get("diopt", envir=diopt.env)
    assign("diopt", diopt.dflt, envir=diopt.env)
    opt
  } else {
    if ((1 == length(args)) && is.list(args[[1]]) &&
        (is.null(names(args)[1]) || ("" == names(args)[1]))) {
      args <- args[[1]]
    }

    vals <- lapply(seq_along(args),
                   function(i) {
                     tag <- names(args)[i]
                     val <- args[[i]]
                     opt <- get("diopt", envir=diopt.env)
                     if (is.null(tag) || ("" == tag)) {
                       # If assign a name to this, then the returned list
                       # element is still unnamed.  Must handle separately.
                       opt[[val]]
                     } else if (!(tag %in% names(opt))) {
                       warning(paste0("ignoring invalid Diopt key ", tag))
                       NULL
                     } else {
                       val <- validate.diopt(tag, val)
                       if (is.null(val)) {
                         NULL
                       } else {
                         old <- opt[[tag]]
                         opt[[tag]] <- val
                         assign("diopt", opt, envir=diopt.env)
                         old
                       }
                     }
                   })
    names(vals) <- lapply(seq_along(args),
                          function(i) {
                            tag <- names(args)[i]
                            if (is.null(tag) || ("" == tag)) {
                              args[[i]]
                            } else {
                              tag
                            }
                          })
    vals
  }
}

# Return a list with all current options, overridden by the tag=value pairs
# in the argument.  These pairs may be wrapped in a single unnamed list.
Diopt.local <- function(...) {

  args <- list(...)
  if ((1 == length(args)) && is.list(args[[1]]) &&
      (is.null(names(args)[1]) || ("" == names(args)[1]))) {
    args <- args[[1]]
  }
  
  opt <- get("diopt", envir=diopt.env)
  for (i in seq_along(args)) {
    tag <- names(args)[i]
    val <- args[[i]]
    val <- validate.diopt(tag, val)
    # Do not allow adding extra tags.
    if (!(tag %in% names(opt))) {
      warning(paste0("ignoring invalid Diopt key ", tag))
    } else if (!is.null(val)) {
      opt[[tag]] <- val
    }
  }
  opt
}


##### Internal Implementation

# Default option values.
diopt.dflt <-
  list(
    # Analyses
    analysis=c("lp", "diw"),
                               # which spacing/smoothing to check
    # Data prep
    data.midq=0,               # mid-quantile approximation method
    # Low-pass Filter Setup
    lp.kernel="kaiser",        # filter kernel
    lp.window=0.15,            # kernel size as fraction of data or integer
    lp.tests=c("ht", "pkexcur", "len", "ftexcur"),
                               # default tests on low-pass features
    lp.param=NULL,             # overrides of detector/test parameters
    # Interval Spacing Setup
    diw.window=0.10,           # interval width as fraction of data or integer
    diw.tests=c("pkexcur", "runht", "nrun", "runlen", "ftexcur"),
                               # default tests on interval features
    diw.param=NULL,            # overrides of detector/test parameters
    # Local Extrema Detector
    peak.fht=0.05,             # min peak ht as fraction of data range
    peak.frelht=0.15,          # peak ht as fraction of local range
    peak.fhtie=0.001,          # max relative difference for tied points
    peak.fhsupp=0.9,           # fraction peak height for support (1 min-min)
    # Local Flat Detector
    flat.fripple=0.05,         # ripple specification
    flat.minlen=30,            # min absolute length of flat
    flat.fminlen=0.05,         # min flat length as fraction of data
    flat.noutlier=1,           # number of outliers (outside ripple) w/i flat
    flat.distrib="logistic",   # null distribution model
    # Excursion/Permutation/Runs Tests
    excur.nrep=15000,          # number sample trials peak/flat excursion tests
    excur.ntop=8,              # num large spacings at start/end to skip
    excur.seed=0,              # excursion RNG seed before test, 0 to not set
    perm.nrep=5000,            # number sample trials for the runht test
    perm.seed=0,               # permutation RNG seed before test, 0 to not set
    # Test Significance (Acceptance) Levels
    alpha.ht=0.01,             # peak height model significance level
    alpha.pkexcur.lp=0.05,     # low-pass peak excursion significance level
    alpha.pkexcur.diw=0.05,    # Diw peak excursion significance level
    alpha.len=0.05,            # flat length model significance level
    alpha.ftexcur.lp=0.01,     # low-pass flat excursion significance level
    alpha.ftexcur.diw=0.01,    # Diw flat excursion significance level
    alpha.runht=0.005,         # runs permutation significance level
    alpha.nrun=0.01,           # runs statistics significance level
    alpha.runlen=0.01,         # longest run significance level
    # Tracking Parameters
    track.maxwindow=0.4,       # largest value for [lp|diw].window
    # Display Parameters
    palette="Dark 2",          # plotting color palette, may start w/ 'hcl:'
    colID.data=7,              # index in palette drawing data points/spacing
    colID.filter=4,            # index in palette drawing LP/interval spacing
    colID.hist=6,              # index in palette drawing histogram of data
    colID.cdf=1,               # index in palette drawing distribution
    colID.peak=8,              # index in palette marking peaks
    colID.flat=5,              # index in palette marking flats
    mark.alpha=TRUE,           # whether to underline significant probabilities
    mark.flat="box",           # how to indicate flats in graphs
    digits=4                   # signif digits for raw values, 0 take options()
  )
class(diopt.dflt) <- "Diopt"

# Global state.  Must stash diopt in a new environment because globals are
# locked in a namespace, but the environment contents are not.
diopt.env <- new.env(hash=FALSE)
assign("diopt", diopt.dflt, envir=diopt.env)


# Spacing analysis to generate/run.
analysis.names <- c("lp", "diw")

# Low-pass filter kernels.
fir.names <- c("kaiser", "triangular", "bartlett", "hamming", "hanning",
               "gaussian", "normal", "blackman")

# Null distributions for flat model.
flat.basedist <- c("logistic", "weibull", "gaussian", "normal", "gumbel")

# Feature tests available in the low-pass spacing.
feattest.lp <- c("ht", "pkexcur", "len", "ftexcur")

# Feature tests available in the interval spacing.
feattest.diw <- c("pkexcur", "runht", "nrun", "runlen", "ftexcur")

# Styles for indicating flats in graph.
flat.style <- c("box", "bar")


# Check if the value val is appropriate for the option tag.  Return the valid
# value (may be modified, b.v. for string completion) or NULL if the value
# cannot be used.  If tag is not recognized return NULL.
validate.diopt <- function(tag, val) {

  switch(tag,
         analysis=diopt.isstring(val, analysis.names, TRUE),
         data.midq=diopt.isposint(val, incl0=TRUE, maxval=4),
         lp.kernel=diopt.isstring(val, fir.names, FALSE),
         lp.window=diopt.iswindow(val),
         lp.tests=diopt.isstring(val, feattest.lp, TRUE),
         lp.param=diopt.isparam(val),
         diw.window=diopt.iswindow(val),
         diw.tests=diopt.isstring(val, feattest.diw, TRUE),
         diw.param=diopt.isparam(val),
         peak.fht=diopt.isfraction(val),
         peak.frelht=diopt.isfraction(val),
         peak.fhtie=diopt.isfraction(val),
         peak.fhsupp=diopt.isfraction(val, TRUE),
         flat.fripple=diopt.isfraction(val),
         flat.minlen=diopt.isposint(val),
         flat.fminlen=diopt.isfraction(val),
         flat.noutlier=diopt.isposint(val, incl0=TRUE),
         flat.distrib=diopt.isstring(val, flat.basedist, FALSE),
         excur.nrep=diopt.isposint(val),
         excur.ntop=diopt.isposint(val, incl0=TRUE),
         excur.seed=diopt.isposint(val, incl0=TRUE),
         perm.nrep=diopt.isposint(val),
         perm.seed=diopt.isposint(val, incl0=TRUE),
         alpha.ht=diopt.isfraction(val),
         alpha.pkexcur.lp=diopt.isfraction(val),
         alpha.pkexcur.diw=diopt.isfraction(val),
         alpha.len=diopt.isfraction(val),
         alpha.ftexcur.lp=diopt.isfraction(val),
         alpha.ftexcur.diw=diopt.isfraction(val),
         alpha.runht=diopt.isfraction(val),
         alpha.nrun=diopt.isfraction(val),
         alpha.runlen=diopt.isfraction(val),
         track.maxwindow=diopt.iswindow(val),
         palette=diopt.ispalette(val),
         colID.data=diopt.iscolorID(val),
         colID.filter=diopt.iscolorID(val),
         colID.hist=diopt.iscolorID(val),
         colID.cdf=diopt.iscolorID(val),
         colID.peak=diopt.iscolorID(val),
         colID.flat=diopt.iscolorID(val),
         mark.alpha=diopt.isbool(val),
         mark.flat=diopt.isstring(val, flat.style, FALSE),
         digits=diopt.isposint(val, incl0=TRUE),
         NULL)
}


# Screen the values val, returning the full match to options opt, or NULL if
# there is no partial match.  If several.ok is TRUE allow multiple values and
# return NULL if any fail.
diopt.isstring <- function(val, opts, several.ok=FALSE) {

  val <- tolower(val)
  if (several.ok) {
    optID <- pmatch(val, opts, nomatch=0L)
    if (any(0 == optID)) {
      NULL
    } else {
      opts[optID]
    }
  } else if (1 == length(val)) {
    optID <- pmatch(val, opts, nomatch=0L)
    if (0 == optID) {
      NULL
    } else {
      opts[optID]
    }
  } else {
    NULL
  }
}

# Return the value val if either a fraction or a positive integer, or NULL.
diopt.iswindow <- function(val) {
  fval <- diopt.isfraction(val)
  if (is.null(fval)) {
    diopt.isposint(val, TRUE)
  } else {
    fval
  }
}

# Return the value val if a fraction between 0 and 1, excl unless incl01 is
# TRUE, or NULL.
diopt.isfraction <- function(val, incl01=FALSE) {
  if ((1 == length(val)) && is.numeric(val) &&
     ifelse(incl01, 0.0 <= val, 0.0 < val) &&
     ifelse(incl01, val <= 1.0, val < 1.0)) {
    val
  } else {
    NULL
  }
}

# Return the value val if a positive integer, or NULL.  If incl0 is TRUE
# also allow zero values.
diopt.isposint <- function(val, incl0=FALSE, maxval=Inf) {
  if ((1 == length(val)) && (abs(val - round(val)) < 1e-6) &&
      ifelse(incl0, 0 <= val, 0 < val) && (val <= (maxval + 1e-6))) {
    as.integer(round(val))
  } else {
    NULL
  }
}

# Return the value val if any positive number, or NULL.
diopt.ispos <- function(val) {

  if ((1 == length(val)) && is.numeric(val) && (0.0 < val)) {
    val
  } else {
    NULL
  }
}

# Return the value val if a pair of fractions in increasing order, or NULL.
diopt.isqrange <- function(val) {

  if ((2 == length(val)) &&
      is.numeric(val[1]) && (0.0 <= val[1]) && (val[1] <= 1.0) &&
      is.numeric(val[2]) && (0.0 <= val[2]) && (val[2] <= 1.0) &&
      (val[1] < val[2])) {
    val
  } else {
    NULL
  }
}

# Return the value val if a list with acceptable detector/test parameters
# (peak.*, flat.*, alpha.*), or NULL if any member is invalid.  Returned
# values may be changed by the tests (b.v. string completion).
diopt.isparam <- function(val) {

  if (is.list(val) &&
      !any(is.na(match(names(val), names(diopt.dflt)))) &&
      all( sapply(names(val),
                  function(tag) {
                    grepl("^peak\\.", tag) || grepl("^flat\\.", tag) ||
                    grepl("^alpha\\.", tag)
                  }) )) {
    param <- lapply(seq_along(val),
                    function(i) {
                      validate.diopt(names(val)[i], val[[i]])
                    })
    if (!any(sapply(param, is.null))) {
      names(param) <- names(val)
      param
    } else {
      NULL
    }
  } else {
    NULL
  }
}

# Return the value val if a known palette, either default R set from
# palette.pals or, if with initial prefix 'hcl:<name>', from hcl.pals.  val
# must be a string and may be in lower case; matching is done as in the R
# functions.  Return NULL if the value is invalid, including more than one.
diopt.ispalette <- function(val) {

  if ((1 == length(val)) && is.character(val)) {
    # This regexp is used by both palette.pals() and hcl.pals().
    re <- "[-_\\,()\\ \\.]"
    val <- tolower(val)
    if ("hcl:" == substr(val, 1, 4)) {
      pals <- hcl.pals()
      val <- substring(val, 5)
      prefix <- "hcl:"
    } else {
      pals <- palette.pals()
      prefix <- ""
    }
    pos <- match(gsub(re, "", val), tolower(gsub(re, "", pals)), nomatch=0L)
    if (0 < pos) {
      paste0(prefix, pals[pos])
    } else {
      NULL
    }
  } else {
    NULL
  }
}

# Return the value val if a valid index into the palette, which will be
# generated with 8 colors, ie. val must be 1 /tm 8.  Convert value to an
# integer.  Return NULL if invalid, including more than one.
diopt.iscolorID <- function(val) {

  if ((1 == length(val)) && is.numeric(val) && (1 <= val) && (val <= 8)) {
    as.integer(round(val))
  } else {
    NULL
  }
}

# Return the value val coerced to a logical TRUE/FALSE, or NULL if that
# cannot be done.
diopt.isbool <- function(val) {

  if ((1 == length(val)) && !is.na(as.logical(val))) {
    as.logical(val)
  } else {
    NULL
  }
}

# detectors.R:
# R side interface to feature detectors.  These are not specific to the
# spacing module; they can be used for any data.
#
# c 2024-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
# -
#


##### Public Interface

# Finds runs in data x, with fuzzy relative comparison where the difference
# from all points within the run to its start (not chaining along) is
# within the fraction feps of their average, ie. |(x2-x1)/((x1+x2)/2)| < feps.
# If feps is 0 then values must match exactly.  Infinities match each other
# but not finite values and NA and NaN values are skipped.  x can also be
# integer, logical, or string values in which only exact matches are supported
# (but still skipping NAs, unlike rle).  Returns a general vector with elements
#   $runs     lengths of fuzzy runs, length of x, 0 if no run starts there
#   $nskip    number of skipped elements
#   $stats    ['nrun'] number of runs, ['maxrun'] longest run
# The runs and nskip vectors are set up to scan along the original data.
#   i <- if (0 == r$runs[1]) { r$nskip[1] + 1 } else { 1 }
#   i <- i + r$runs[i] + r$nskip[i]
# Use runs.as.rle to convert the result to an rle encoding.
# Typical values are 0.001 for feps.
find.runs <- function(x, feps) {

  if (!is.vector(x)) {
    x <- as.vector(x)
  }
  if (0 == length(x)) {
    stop("vector has no data")
  }
  
  .Call("C_find_runs", x, feps, PACKAGE="Dimodal")
}


# Find local minima and maxima in data x, eliminating small peaks if 
# they are less than the fraction fht of the total range of the data or if
# the height to an adjacent minimum is less than the fraction frelht of their
# average value.  Values within fhtie are treated as equal (a fuzzy run) and
# considered located in the middle of the mesa.  x must be an integer or real
# vector.  The support of a maximum is the range of points until their values
# drop to the fraction fhsupp of the peak height to that side, similar to FWHM
# where fhsupp would be 0.5.  The range is NA for minima and one-sided peaks.
# Return a data frame assigned to class Dipeak with extrema in rows and columns
#   $pos       position of extremum in x
#   $ismax     TRUE if a maximum, FALSE if a minimum
#   $valsd     data at extremum standardized by stdev
#   $lht       peak height to left minimum, NA at minimum or first peak
#   $rht       peak height to right minimum, NA at minimum or last peak
#   $lminID    position of minimum to left, NA if first/last or at minimum
#   $rminID    position of minimum to right, NA if first/last or at minimum
#   $lsuppID   first index of support range
#   $rsuppID   last index (incl.) of support range
# Typical values are 0.05 for fht, 0.15 for frelht, and 0.001 for fhtie.
# The data frame will be empty (0 rows) if the data do not support extrema:
# there are too few, or the values are constant, or are not finite.
find.peaks <- function(x, fht, frelht, fhtie, fhsupp) {

  if (!is.vector(x)) {
    x <- as.vector(x)
  }
  if (0 == length(x)) {
    return(mockup.Dipeak())
  }

  pk <- .Call("C_find_peaks", x, fht, frelht, fhtie, fhsupp, PACKAGE="Dimodal")
  df <- as.data.frame(pk)
  class(df) <- c("Dipeak", class(df))
  df
}


# Find local flats in data x, as defined by the fripple value as a fraction
# of the range of x centered at each point.  Flats must have a length at least
# minlen (absolute) or a fraction fminlen as a fraction of the number of data
# points, whichever is larger.  To each side of the seed flats may contain
# noutlier points outside the ripple spec, although the start or endpoint will
# always be in range.  They may overlap.  Data must be real or integer and
# non-finite values will be ignored.  Returns a data frame with features in
# rows and columns
#   $src       point in x that seeded the flat, a unique identifier
#   $stID      starting index in x of each flat
#   $endID     ending index in x of each flat
#   $len       length of flat
#   $srcval    data value at src (center of flat height)
#   $ht        height of flat in x
#   $htsd      standardized height (by stdev) of flat
# Typical values are 0.05 for fripple, 30 for minlen, 0.05 for rellen, and
# 1 for noutlier.
find.flats <- function(x, fripple, minlen, fminlen, noutlier) {

  # Final timing [ms] based on sub-set of asteroid axes, for Diw 150 window.
  # This is using identify_flats_cover for both.
  #   # samples    scan     tree    speed-up
  #     1k            1        1        1x
  #     5k            5        8        0.6x
  #    10k           31       17      1.8x
  #    50k          827      105      7.9x
  #   100k         5250      294     17.9x
  #   250k        33214      969     34.3x
  #   500k       142768     1934     73.8x
  #   750k       336697     4370     77.0x
  if (!is.vector(x)) {
    x <- as.vector(x)
  }
  if (sum(is.finite(x)) < 2) {
    return(mockup.Diflat())
  }

  ft <- .Call("C_find_flats", x, fripple, minlen, fminlen, noutlier, 0L,
              PACKAGE="Dimodal")
  df <- as.data.frame(ft)
  class(df) <- c("Diflat", class(df))
  df
}

# Find intervals without a significant slope, inverting the interval spacing
# test used in the modehunt package.  Data x must be real or integer, and
# non-finite values, including NA and NaN, will be removed; the indices will
# refer to the sorted raw data after removal.  alpha is the significance
# level of the test, between 0 and 1, and correct, a boolean, if TRUE adjusts
# the test against short intervals.  Returns a data frame with one row per
# section and columns
#   $stID      starting index in the sorted x of each section
#   $endID     ending index (incl.) of each section
# Typical values are 0.95 for alpha and TRUE for correct.
find.level.sections <- function(x, alpha, correct) {

  if (!is.vector(x)) {
    x <- as.vector(x)
  }

  # finmap will store the indices in the original data we're processing,
  # which we need when mapping back from the level sections.
  finmap <- which(is.finite(x))
  x <- sort(x[finmap])
  if (is.integer(x[1L])) {
    x <- as.numeric(x)
  }

  if (length(x) < 3) {
    stop("must have at least three points for level sections")
  }

  lvl <- .Call("C_find_level_sections", x, alpha, correct, PACKAGE="Dimodal")

  data.frame( stID=finmap[lvl[["stID"]]], endID=finmap[lvl[["endID"]]] )
}

# Dimodal.R:
# Use spacing to detect modality, with local peaks corresponding to changes
# and flats to modes.  Detect and evaluate features in spacing after low-pass
# filtering with height or length or aspect ratio models or excursion/
# bootstrap simulations, in interval spacing as excursions or in statistics
# of signed runs or longest or permutation simulation.
#
# c 2024-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
#



#### Public Interface

# Perform a modality analysis on data vector x using low-pass spacing if 
# Diopt('analysis') includes 'lp', interval spacing with 'diw';
# more than one method can be chosen.
# opt sets the # Diopt analysis parameters.  The return value is a list of
# class 'Dimodal' with elements added per the analyze.* functions.
Dimodal <- function(x, opt=Diopt()) {

  if ("Diopt" != class(opt)) {
    opt <- Diopt(opt)
  }

  if (get("._TIMEIT", envir=._timer.env)) { setup.timing() }

  res <- list(data=gen.data(x, opt$analysis, opt), opt=opt)
  attributes(res$data) <- c(attributes(res$data),
                            list(data.name=deparse1(substitute(x))))

  if ("lp" %in% opt$analysis) {
    res <- c(res, analyze.lp(res$data, res$opt))
  }
  if ("diw" %in% opt$analysis) {
    res <- c(res, analyze.diw(res$data, res$opt))
  }

  class(res) <- "Dimodal"
  res
}


#### Implementation

# Perform feature detection and significance determination on the low-pass
# spacing stored in gen.data matrix d, using Diopt options opt.  Returns a
# list with elements
#   $lp.peaks      local extrema (position, height, signficance)
#   $lp.flats      local flats (position, significance)
# The tests run and added to the results depend on those stored in the
# lp.tests option.  Adds columns $p[peak|flat] and $naccept for a summary of
# each result.
analyze.lp <- function(d, opt) {

  if (is.null(d) || (0 == nrow(d)) || !("lp" %in% rownames(d))) {
    return(list(lp.peaks=mockup.Dipeak(), lp.flats=mockup.Diflat()))
  }

  if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LP") }

  # Local overrides.
  # It seems that sapply(opt$lp.param, function(p) { ... }) strips the
  # element name from p when passing to the function, so we must
  # address by position along the list.  But we don't have to screen for NULL.
  for (i in seq_along(opt$lp.param)) {
    opt[[names(opt$lp.param)[i]]] <- opt$lp.param[[i]]
  }

  a <- attributes(d)
  lpdi <- d["lp",a$lp.stID:a$lp.endID]
  # The model tests want the fractional window, not the absolute a$wlp.
  fwlp <- a$lp.window
  if (1 <= fwlp) {
    # - 1 b/c spacing not raw data.
    fwlp <- fwlp / (ncol(d) - 1)
  }

  if (("ht" %in% opt$lp.tests) || ("pkexcur" %in% opt$lp.tests)) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPPEAK") }
    pk <- find.peaks(lpdi, opt$peak.fht, opt$peak.frelht, opt$peak.fhtie,
                     opt$peak.fhsupp)
    pk <- shiftID.place(pk, a$lp.stID, d["xmid",])
    hts <- pmax(pk$lht, pk$rht)
    nsig <- rep(0L, nrow(pk))
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPPEAK") }

    if ("ht" %in% opt$lp.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPPKHT") }
      # -1 because length is spacing.
      tst <- Dipeak.test(hts, ncol(d)-1, fwlp, a$lp.kernel, FALSE)
      pht <- tst$p.value
      pk <- cbind(pk, pht=pht)
      if (0 < length(pht)) {
        nsig <- nsig + as.integer(is.finite(pht) & (pht <= opt$alpha.ht))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPPKHT") }
    } else {
      pht <- rep(NA, nrow(pk))
    }

    if ("pkexcur" %in% opt$lp.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPPKEXCUR") }
      b <- build.excursion(d, "lp", pk$lsuppID, pk$rsuppID, opt$excur.ntop)
      tst <- Diexcurht.test(b$ht, b$ndraw, b$xbase, opt$excur.nrep, TRUE,FALSE,
                            13*opt$excur.seed)
      pexcur <- tst$p.value
      pk <- cbind(pk, hexcur=tst$statistic, pexcur=pexcur)
      if (0 < length(pexcur)) {
        nsig <- nsig + as.integer(is.finite(pexcur) &
                                  (pexcur <= opt$alpha.pkexcur.lp))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPPKEXCUR") }
    } else {
      pexcur <- rep(NA, nrow(pk))
    }

    pk <- cbind(pk, ppeak=pmin(pht, pexcur, na.rm=TRUE), naccept=nsig)
    # The cbind's don't preserve the class, so re-insert.
    if (!("Dipeak" %in% class(pk))) {
      class(pk) <- c("Dipeak", class(pk))
    }
  } else {
    pk <- mockup.Dipeak()
  }

  if (("len" %in% opt$lp.tests) || ("ftexcur" %in% opt$lp.tests)) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPFLAT") }
    ft <- find.flats(lpdi, opt$flat.fripple, opt$flat.minlen, opt$flat.fminlen,
                     opt$flat.noutlier)
    ft <- shiftID.place(ft, a$lp.stID, d["xmid",])
    nsig <- rep(0L, nrow(ft))
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPFLAT") }

    if ("len" %in% opt$lp.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPFTLEN") }
      # -1 because length is spacing.
      tst <- Diflat.test(ft$len, ncol(d)-1, fwlp, a$lp.kernel,
                         opt$flat.distrib, FALSE)
      plen = tst$p.value
      ft <- cbind(ft, plen=plen)
      if (0 < length(plen)) {
        nsig <- nsig + as.integer(is.finite(plen) & (plen <= opt$alpha.len))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPFTLEN") }
    } else {
      plen <- rep(NA, nrow(ft))
    }

    if ("ftexcur" %in% opt$lp.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_LPFTEXCUR") }
      b <- build.excursion(d, "lp", ft$stID, ft$endID, opt$excur.ntop)
      tst <- Diexcurht.test(b$ht, b$ndraw, b$xbase, opt$excur.nrep, FALSE,TRUE,
                            17*opt$excur.seed)
      fexcur <- tst$p.value
      ft <- cbind(ft, hexcur=tst$statistic, pexcur=fexcur)
      if (0 < length(fexcur)) {
        nsig <- nsig + as.integer(is.finite(fexcur) &
                                  (fexcur <= opt$alpha.ftexcur.lp))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LPFTEXCUR") }
    } else {
      fexcur <- rep(NA, nrow(ft))
    }

    ft <- cbind(ft, pflat=pmin(plen, fexcur, na.rm=TRUE), naccept=nsig)
    if (!("Diflat" %in% class(ft))) {
      class(ft) <- c("Diflat", class(ft))
    }
  } else {
    ft <- mockup.Diflat()
  }

  attr(pk, "source") <- "LP"
  attr(ft, "source") <- "LP"
  
  if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_LP") }

  return(list(lp.peaks=pk, lp.flats=ft))
}

# Perform feature detection and significance determination on the interval
# spacing stored in gen.data matrix d, using Diopt options opt.  Returns a
# list with elements
#   $diw.peaks     local extrema (position, height, significance)
#   $diw.flats     local flats (position, significance)
# The tests run and added to the results depend on those stored in the
# diw.tests option.  Adds columns $p[peak|flat] and $naccept for a summary of
# each result.
analyze.diw <- function(d, opt) {

  if (is.null(d) || (0 == nrow(d)) || !("Diw" %in% rownames(d))) {
    return(list(diw.peaks=mockup.Dipeak(), diw.flats=mockup.Diflat()))
  }

  if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIW") }

  for (i in seq_along(opt$diw.param)) {
    opt[[names(opt$diw.param)[[i]]]] <- opt$diw.param[[i]]
  }

  a <- attributes(d)
  diw <- d["Diw",a$diw.stID:ncol(d)]

  if (("pkexcur" %in% opt$diw.tests) || ("runht" %in% opt$diw.tests) ||
      ("nrun" %in% opt$diw.tests) || ("maxrun" %in% opt$diw.tests)) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWPEAK") }
    pk <- find.peaks(diw, opt$peak.fht, opt$peak.frelht, opt$peak.fhtie,
                     opt$peak.fhsupp)
    pk <- shiftID.place(pk, a$diw.stID, d["xmid",])
    nsig <- rep(0L, nrow(pk))
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWPEAK") }
                       
    if ("nrun" %in% opt$diw.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWNRUN") }
      # + 1 because the signed difference starts one point later.
      # No feps to the test because values are via sign.
      tst <- Dinrun.test(d["signed",], pk$lminID+1, pk$rminID, 0, TRUE)
      pkr <- tst$p.value
      pk <- cbind(pk, nrun=tst$statistic, pnrun=pkr)
      if (0 < length(pkr)) {
        nsig <- nsig + as.integer(is.finite(pkr) & (pkr <= opt$alpha.nrun))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWNRUN") }
    } else {
      pkr <- rep(NA, nrow(pk))
    }

    if ("runlen" %in% opt$diw.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWRUNLEN") }
      # + 1 because the signed difference starts one point later.
      # 0 for feps because of the limited values from sign.
      tst <- Dirunlen.test(d["signed",], pk$lminID+1, pk$rminID, 0)
      prl <- tst$p.value
      pk <- cbind(pk, runlen=tst$statistic, prunlen=prl)
      if (0 < length(prl)) {
        nsig <- nsig + as.integer(is.finite(prl) & (prl <= opt$alpha.runlen))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWRUNLEN") }
    } else {
      prl <- rep(NA, nrow(pk))
    }

    if ("runht" %in% opt$diw.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWRUNHT") }
      b <- build.permutation(d, pk$lminID, pk$rminID)
      tst <- sapply(seq_along(b$ht),
                    function(i) {
                      # Only one permutation test so no need to scale seed.
                      res <- Dipermht.test(b$ht[i], b$xbase[[i]],
                                           opt$perm.nrep, FALSE, opt$perm.seed)
                      c(ht=res$statistic, pval=res$p.value)
                    })
      if (0 == length(b$ht)) {
        pperm <- NULL
      } else {
        pperm <- tst["pval",]
        pk <- cbind(pk, runht=tst["ht",], prunht=pperm)
        if (0 < length(pperm)) {
          nsig <- nsig + as.integer(is.finite(pperm) &
                                    (pperm <= opt$alpha.runht))
        }
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWRUNHT") }
    } else {
      pperm <- rep(NA, nrow(pk))
    }

    if ("pkexcur" %in% opt$diw.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWPKEXCUR") }
      b <- build.excursion(d, "Diw", pk$lsuppID, pk$rsuppID, opt$excur.ntop)
      tst <- Diexcurht.test(b$ht, b$ndraw, b$xbase, opt$excur.nrep, TRUE,FALSE,
                            19*opt$excur.seed)
      pexcur <- tst$p.value
      pk <- cbind(pk, hexcur=tst$statistic, pexcur=pexcur)
      if (0 < length(pexcur)) {
        nsig <- nsig + as.integer(is.finite(pexcur) &
                                  (pexcur <= opt$alpha.pkexcur.diw))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWPKEXCUR") }
    } else {
      pexcur <- rep(NA, nrow(pk))
    }

    pk <- cbind(pk, ppeak=pmin(pkr, prl, pperm, pexcur, na.rm=TRUE),
                naccept=nsig)
    # The cbind's don't preserve the class, so re-insert.
    if (!("Dipeak" %in% class(pk))) {
      class(pk) <- c("Dipeak", class(pk))
    }
  } else {
    pk <- mockup.Dipeak()
  }

  if ("ftexcur" %in% opt$diw.tests) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWFLAT") }
    ft <- find.flats(diw, opt$flat.fripple, opt$flat.minlen, opt$flat.fminlen,
                     opt$flat.noutlier)
    ft <- shiftID.place(ft, a$diw.stID, d["xmid",])
    nsig <- rep(0L, nrow(ft))
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWFLAT") }
    
    if ("ftexcur" %in% opt$diw.tests) {
      if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_DIWFTEXCUR") }
      b <- build.excursion(d, "Diw", ft$stID, ft$endID, opt$excur.ntop)
      tst <- Diexcurht.test(b$ht, b$ndraw, b$xbase, opt$excur.nrep, FALSE,TRUE,
                            23*opt$excur.seed)
      fexcur <- tst$p.value
      ft <- cbind(ft, hexcur=tst$statistic, pexcur=fexcur)
      if (0 < length(fexcur)) {
        nsig <- nsig + as.integer(is.finite(fexcur) &
                                  (fexcur <= opt$alpha.ftexcur.diw))
      }
      if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIWFTEXCUR") }
    } else {
      fexcur <- rep(NA, nrow(ft))
    }

    ft <- cbind(ft, pflat=pmin(fexcur, na.rm=TRUE), naccept=nsig)
    if (!("Diflat" %in% class(ft))) {
      class(ft) <- c("Diflat", class(ft))
    }
  } else {
    ft <- mockup.Diflat()
  }

  attr(pk, "source") <- "Diw"
  attr(ft, "source") <- "Diw"

  if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_DIW") }
  
  return(list(diw.peaks=pk, diw.flats=ft))
}


### Data generation

# Build all data for the analysis from data x and Diopt list opt for the
# analysis type(s) in method.  Non-finite values are automatically dropped.
# Returns a matrix of length(x) with rows:
#   'x'        raw data
#   'xsort'    sorted data
#   'xmid'     mid-distribution function or piece-wise linear CDF b/t jumps
#   'Di'       spacing, lead column (index 1) is NA
# If method includes 'lp' adds
#   'lp'       low-pass spacing, lead/trailing columns will be NA
# If method includes 'diw' adds
#   'Diw'      interval spacing, leading columns will be NA
#   'signed'   signed difference of Diw, one more leading column NA
# If wdiw is the actual interval size (not fractional per Diopt), then the
# first wdiw columns in the Diw row are NA, and the first wdiw+1 in signed.
# If wlp is the actual low-pass kernel size, then the first wlp/2 columns
# of lp are NA, rounded up, and the last wlp/2 columns rounded down are NA.
# If method includes 'lp' the matrix has attributes added:
#   'lp.kernel', 'lp.window' copied from opt
#   'wlp'      actual size of kernel (integer)
#   'lp.stID'  first column with non-NA data in lp row
#   'lp.endID' last column with non-NA data in lp row
# If method includes 'diw' the matrix has attributes added:
#   'diw.window' copied from opt
#   'wdiw'     actual interval (integer)
#   'diw.stID' first column with non-NA data in Diw row, = wdiw+1
#              interval data always goes up to last column
gen.data <- function(x, method, opt) {

  # The low-pass filter has a long run time on large data sets: 0.6 s for 50k
  # points, 66 s for 500k, 262 s for 1M, ie. O(n^2).  We added the method
  # argument to avoid running this on such sets if not needed.

  if (!is.numeric(x)) {
    stop("original data must be numeric")
  }
  x <- x[is.finite(x)]
  nx <- length(x)
  if (nx < 2) {
    warning("not generating spacing - too little data")
  }

  if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_XSORT") }
  
  xsrt <- sort(x)

  if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_XSORT") }
  if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_XMID") }

  xmid <- midquantile(xsrt, type=opt$data.midq)

  if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_XMID") }
  if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_XDI") }

  di <- c(NA, diff(xsrt))

  if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_XDI") }

  mdi <- matrix(c(x, xsrt, xmid, di), nrow=4, byrow=TRUE,
                dimnames=list(c("x", "xsort", "xmid", "Di"), NULL))

  miw <- NULL
  aiw <- NULL
  if ("diw" %in% method) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_XDIW") }
    
    if (opt$diw.window < 1.0) {
      wdiw <- round(opt$diw.window * nx)
    } else {
      wdiw <- round(opt$diw.window)
    }
    if (wdiw < 1) {
      warning("not generating interval spacing - interval too small")
    } else if (nx <= (wdiw + 1)) {
      warning("not generating interval spacing - interval does not fit data")
    } else {
      diw <- c(rep(NA, wdiw), diff(xsrt, lag=wdiw))

      # There can be problems with machine accuracy here, see data test case 7.
      # Clip all very small differences to 0 before taking the sign.  The 
      # factor of 10 is very loose, only 4 would be needed for the test case.
      ddiw <- diff(diw)
      ddiw[abs(ddiw) < (10 * .Machine$double.eps)] <- 0
      sdiw <- c(NA, sign(ddiw))

      miw <- matrix(c(diw, sdiw), nrow=2, byrow=TRUE,
                    dimnames=list(c("Diw", "signed"), NULL))
      aiw <- list(diw.window=opt$diw.window, wdiw=wdiw, diw.stID=wdiw+1)
    }
    
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_XDIW") }
  }

  mlp <- NULL
  alp <- NULL
  if ("lp" %in% method) {
    if (get("._TIMEIT", envir=._timer.env)) { start.timer("TIMER_XLP") }
    
    kernel <- fir.kernel(di[-1L], opt)
    wlp <- length(kernel)
    if (wlp < 3) {
      warning("not generating low-pass spacing - filter kernel too small")
    } else if (nx < (wlp + 1)) {
      warning("not generating low-pass spacing - not enough data for kernel")
    } else {
      u <- filter(di, rev(kernel), method="convolution", sides=2)

      stID <- 1 + ((wlp + 1) %/% 2)
      endID <- nx - (wlp - stID) - 1
    
      mlp <- matrix(u, nrow=1, byrow=TRUE, dimnames=list(c("lp"), NULL))
      alp <- list(lp.kernel=opt$lp.kernel, lp.window=opt$lp.window,
                wlp=wlp, lp.stID=stID, lp.endID=endID)
    }
    
    if (get("._TIMEIT", envir=._timer.env)) {  stop.timer("TIMER_XLP") }
  }

  m <- rbind(mdi, mlp, miw)
  attributes(m) <- c(attributes(m), alp, aiw)
  class(m) <- c("Didata", class(m))

  m
}

# Return the kernel for the low-pass filter specified in Diopt opt.  The
# data x is needed to determine the window size if a fraction.
fir.kernel <- function(x, opt) {

  if (opt$lp.window < 1.0) {
    wlp <- round(opt$lp.window * length(x))
  } else {
    wlp <- round(opt$lp.window)
  }

  if (wlp < 3) {
    return(NULL)
  }

  filter <- match.arg(tolower(opt$lp.kernel), fir.names)

  # Filter definitions in Discrete-Time Signal Porcessing ch 7.4 have 0
  # in first and last position - except Hamming and Kaiser - so we've
  # effectively increased the window width by 2 and then deleted these.

  if (("triangular" == filter) || ("bartlett" == filter)) {
    wlp <- wlp + 1
    kernel <- sapply(1:(wlp-1),
                     function(i) { ifelse(i<(wlp/2), 2*i/wlp, 2-(2*i/wlp)) })
  } else if ("kaiser" == filter) {
    beta <- fir.kaiser.beta(0.01)
    alpha <- (wlp - 1) / 2
    kernel <- besselI(beta * sqrt(1 - ((((0:(wlp-1)) - alpha)/alpha)^2)), 0) /
              besselI(beta, 0)
  } else if (("gaussian" == filter) || ("normal" == filter)) {
    # Per Klette (before eq. (2.18)) a three sigma drop-off means the kernel
    # width is 6*sigma - 1, ie. s=1 => diameter 5/radius 2, s=2 => d 11/r 5.
    sigmasq <- 2 * ((wlp + 1) / 6)^2
    wseq <- (1:wlp) - (wlp / 2)
    kernel <- exp(-((((wlp+1)/2) - (1:wlp))^2)/sigmasq) / sqrt(sigmasq * pi)
  } else {
    if ("hamming" == filter) {
      f <- pi * (0:(wlp-1)) / (wlp - 1)
    } else if (("hanning" == filter) || ("blackman" == filter)) {
      f <- pi * (1:wlp) / (wlp + 1)
    } else {
      stop("internal error - uncaught unsupported filter")
    }
    a <- fir.coscoef(filter)
    kernel <- rep(a[1], length(f))
    for (i in 2:length(a)) {
      kernel <- kernel + (a[i] * cos(2*(i-1)*f))
    }
  }

  if (abs(sum(kernel)) < 1e-6) {
    kernel
  } else {
    kernel / sum(kernel)
  }
}

# a is a vector of coefficients for the cosine-based FIR filter, or the
# name of the filter.  If the latter, return the corresponding vector; if
# the former, just return it.  Filters that don't use cosine terms cause
# an error.
fir.coscoef <- function(a) {

  if (is.numeric(a)) {
    a
  } else {
    filter <- match.arg(a, fir.names)
    if ("hanning" == filter) {
      c(0.5, -0.5)
    } else if ("hamming" == filter) {
      # Discrete-Time Signal Processing gives c(0.54, -0.46).  Via Harris
      # window filters page this is an approx of c(25/46, -21/46), which
      # elimates the first sidelobe.  Another option is c(0.53836, -0.46164)
      # which is equiripple.
      # c(0.54, -0.46)
      # c(0.53836, -0.46164)
      c(25/46, -21/46)
    } else if ("blackman" == filter) {
      # Harris values are exact, Discrete-Time SP uses the approximation
      # c(0.42, -0.5, 0.08).
      c(7938/18608, -9240/18608, 1430/18608)
    } else {
      stop(sprintf("unsupported filter '%s'", a))
    }
  }
}
  
# Return the Kaiser beta parameter for a stopband fripple (as a fraction).
fir.kaiser.beta <- function(fripple) {

  A <- -20 * log10(fripple)
  if (A < 21) {
    beta <- 0
  } else if (50 < A) {
    beta <- 0.1102 * (A - 8.7)
  } else {
    beta <- (0.5842 * ((A - 21) ^ 0.4)) + (0.07886 * (A - 21))
  }
  beta
}

# Construct the information needed for an excursion test on the row in
# Didata d.  Features extend between stID and endID (may be vectors) and
# the base of the excursion draws will be the difference of the data,
# ignoring the ntop largest changes if they are within the first or last
# ntop/2 data.  Returns a list with elements
#   $ht       the feature height
#   $ndraw    the feature width, or number of samples per excursion draw
#   $xbase    a vector with the differential signal to build excursion signals
# The height is taken over the endpoints, and therefore reflects the support
# if they are used.
build.excursion <- function(d, row, stID, endID, ntop) {

  if (length(stID) != length(endID)) {
    stop("internal error - feature bounds do not have same length")
  }

  if ("lp" == row) {
    xbase <- diff(d[row,attr(d,"lp.stID"):attr(d,"lp.endID")])
    xoff <- attr(d,"lp.stID") - 1
  } else if ("Diw" == row) {
    xbase <- diff(d[row,attr(d,"diw.stID"):ncol(d)])
    xoff <- attr(d,"diw.stID") - 1
  } else {
    xbase <- diff(d[row,])
    xbase <- xbase[is.finite(xbase)]
    xoff <- 0
  }
  
  ntop <- min(round(0.1 * length(xbase)), ntop)
  if (0 < ntop) {
    o <- order(abs(diff(xbase)), decreasing=TRUE)
    drop <- o[1:ntop]
    ldrop <- drop[drop <= (ntop/2)]
    # + 1 to compensate for the diff, deleting the upper point.
    # But not for the left drop to delete the lower point.
    rdrop <- drop[(length(o)-(ntop/2)+1) <= drop] + 1
    if ((0 < length(ldrop)) || (0 < length(rdrop))) {
      xbase <- xbase[-c(ldrop, rdrop)]
    }

    # Need to move endpoints out of dropped region.
    if (0 < length(ldrop)) {
      # +1 so endpoint after largest ldrop.
      stID <- pmax(stID, max(ldrop)+xoff+1)
      endID <- pmax(endID, max(ldrop)+xoff+1)
    }
    if (0 < length(rdrop)) {
      # -1 so endpoint is smaller than smallest rdrop.
      stID <- pmin(stID, min(rdrop)+xoff-1)
      endID <- pmin(endID, min(rdrop)+xoff-1)
    }
  }

  hts <- sapply(seq_along(stID),
                function(i) {
                  if (!is.finite(stID[i]) || !is.finite(endID[i])) {
                    NA_real_
                  } else {
                    diff(range(d[row,stID[i]:endID[i]]))
                  }
                })

  # Normally would add + 1, but the excursion is based on the diff.
  ndraw <- endID - stID

  list(ht=hts, ndraw=ndraw, xbase=xbase)
}

# Construct the information needed for a permutation test on the runs in
# Didata d for features with endpoints stID, endID (may be vectors).  The
# base data for the draw will be the delta along each run, ie. its length
# times its value (with 0 values mapped to 1).  Returns a list with elements
#   $ht       the feature height
#   $xbase    a list of the run deltas for each feature
build.permutation <- function(d, stID, endID) {

  if (length(stID) != length(endID)) {
    stop("internal error - feature bounds do not have the same length")
  }

  # We work in the signed row, so starting indices are offset by 1.
  stID <- stID + 1

  xsign <- lapply(seq_along(stID),
                 function(i) {
                   if (!is.finite(stID[i]) || !is.finite(endID[i])) {
                     NULL
                   } else {
                     d["signed",stID[i]:endID[i]]
                   }
                 })

  hts <- sapply(seq_along(stID),
                function(i) {
                  if (!is.finite(stID[i]) || !is.finite(endID[i])) {
                    NA_real_
                  } else {
                    diff(range(cumsum(xsign[[i]])))
                  }
                })

  xbase <- lapply(seq_along(stID),
                  function(i) {
                    if (!is.finite(stID[i]) || !is.finite(endID[i])) {
                      NULL
                    } else {
                      r <- rle(xsign[[i]])
                      # Note that we lose the length when the value is 0.
                      # This is OK for the permutation test as a placeholder
                      # but would not work for the runs tests.
                      r$lengths * r$values
                    }
                  })

  list(ht=hts, xbase=xbase)
}


#### Timing Analysis

# Undocumented performance tracking of the main steps in Dimodal.
# Creates a set of clocks TIMER_* for the main data preparation, low-pass
# spacing, interval spacing, and changepoint steps.  All functions must
# be called as Dimodal:::<fn> since they are not exported.  Use track.timing()
# to start or stop the overall tracking, clear.timing() to reset the clocks,
# dump.timing() to print tables of each step averaged over all calls to
# Dimodal() since the last clearing, and get.timers() to return the times
# in a matrix.  start.timer() and stop.timer() are used internally around
# each analysis step.

# Usage Guide:
#   track.timing() -         Track or stop performance tracking of Dimodal()
#   setup.timing() -         Prepares for timing if never done before,
#                            otherwise adds an entry for a new run.
#   clear.timing() -         Removes all history of timing and sets up for a
#                            new run.
#   dump.timing()  -         Print a table with the average timing (in ms)
#                            over however many runs have been timed.  Will
#                            only print the subset of the analysis if parts
#                            are skipped (ex. no low-pass).
#   get.timers()   -         Return a a matrix with the timing, columns are
#                            timer names (see setup.timing) and one row per run
#   start.timer("TIMER_*") - Register the clock time (well, elapsed process
#                            time) for the indicated timer.  A list of timers
#                            is found in setup.timing.
#   stop.timer("TIMER_*")  - Stores the elapsed time spent in the indicated
#                            timer for the current run.

# As with Diopt, we need a new environment because globals get locked.
if (!exists("._timer.env")) {
  ._timer.env <- new.env(hash=FALSE)
  assign("._TIMEIT", FALSE, envir=._timer.env)
}

# Allow timing analysis of Dimodal steps if onoff is TRUE, or stop it if FALSE.
track.timing <- function(onoff) {
  assign("._TIMEIT", onoff, envir=._timer.env)
}

# Prepare for timing Dimodal.  Modifies the ._timer.env global with the
# variables needed for timing Dimodal, including creating the TIMER_*
# identifiers for [start|stop].timer.
setup.timing <- function() {

  # Timers for low-pass spacing analysis.
  assign("TIMER_LP",         1L, envir=._timer.env)
  assign("TIMER_LPPEAK",     2L, envir=._timer.env)
  assign("TIMER_LPPKHT",     3L, envir=._timer.env)
  assign("TIMER_LPPKEXCUR",  4L, envir=._timer.env)
  assign("TIMER_LPFLAT",     5L, envir=._timer.env)
  assign("TIMER_LPFTLEN",    6L, envir=._timer.env)
  assign("TIMER_LPFTEXCUR",  7L, envir=._timer.env)

  # Timers for interval spacing analysis.
  assign("TIMER_DIW",        8L, envir=._timer.env)
  assign("TIMER_DIWPEAK",    9L, envir=._timer.env)
  assign("TIMER_DIWNRUN",    10L, envir=._timer.env)
  assign("TIMER_DIWRUNLEN",  11L, envir=._timer.env)
  assign("TIMER_DIWRUNHT",   12L, envir=._timer.env)
  assign("TIMER_DIWPKEXCUR", 13L, envir=._timer.env)
  assign("TIMER_DIWFLAT",    14L, envir=._timer.env)
  assign("TIMER_DIWFTEXCUR", 15L, envir=._timer.env)

  # Timers for changepoint analysis.
  assign("TIMER_CPT",   16L, envir=._timer.env)

  # Timers for data preparation.
  assign("TIMER_XSORT", 17L, envir=._timer.env)
  assign("TIMER_XMID",  18L, envir=._timer.env)
  assign("TIMER_XDI",   19L, envir=._timer.env)
  assign("TIMER_XDIW",  20L, envir=._timer.env)
  assign("TIMER_XLP",   21L, envir=._timer.env)

  # The timings matrix contains the time spent for each column, a matrix with
  # #(TIMER_*) columns and one row per timing run.
  if (!exists("timings", envir=._timer.env, inherits=FALSE)) {
    assign("timings", matrix(rep(NA_real_, 21), nrow=1), envir=._timer.env)
  } else {
    assign("timings",
           rbind(get("timings", envir=._timer.env),
                 matrix(rep(NA_real_, 21), nrow=1)),
           envir=._timer.env)
  }

  # The tst vector contains the start time for each timer.
  assign("tst", rep(NA_real_, 21), envir=._timer.env)
}

# Reset all timers.  Deletes any history, then re-creates the setup.
# Modifies the ._timer.env contents.
clear.timing <- function() {

  rm(list=ls(._timer.env), envir=._timer.env)
  setup.timing()
}

# Print a table with the average of all timers.
dump.timing <- function() {

  dt <- get("timings", envir=._timer.env)
  avgdt <- apply(dt, 2, mean, na.rm=TRUE)
  ntime <- apply(dt, 1, function(r) { !all(is.na(r)) })
  if (0 == sum(ntime)) {
    cat("\nTiming Analysis\n")
    cat("  none performed\n\n")
    return()
  } else {
    cat("\nTiming Analysis   from ", sum(ntime), " runs, times in [ms]\n")
  }

  dataID <- c("TIMER_XSORT", "TIMER_XDI", "TIMER_XDIW", "TIMER_XLP",
              "TIMER_XMID")
  datacol <- sapply(dataID, get, envir=._timer.env)
  have.data <- !all(is.na(avgdt[datacol]))
  xnames <- c("sort", "spacing", "interval spacing", "low-pass filtering",
              "mid-quantile")

  lpID <- c("TIMER_LP", "TIMER_LPPEAK", "TIMER_LPPKHT", "TIMER_LPPKEXCUR",
            "TIMER_LPFLAT", "TIMER_LPFTLEN", "TIMER_LPFTEXCUR")
  lpcol <- sapply(lpID, get, envir=._timer.env)
  have.lp <- !all(is.na(avgdt[lpcol]))
  lpnames <- c("analysis", "peak detection", "peak height test",
               "peak excursion test", "flat detection", "flat length test",
               "flat excursion test")

  diwID <- c("TIMER_DIW", "TIMER_DIWPEAK", "TIMER_DIWNRUN", 
             "TIMER_DIWRUNLEN", "TIMER_DIWRUNHT", "TIMER_DIWPKEXCUR",
             "TIMER_DIWFLAT", "TIMER_DIWFTEXCUR")
  diwcol <- sapply(diwID, get, envir=._timer.env)
  have.diw <- !all(is.na(avgdt[diwcol]))
  diwnames <- c("analysis", "peak detection", "run count test",
                "run length test", "run height (perm) test",
                "peak excursion test", "flat detection", "flat excursion test")

  cptID <- c("TIMER_CPT")
  cptcol <- sapply(cptID, get, envir=._timer.env)
  have.cpt <- !all(is.na(avgdt[cptcol]))
  cptnames <- c("analysis")

  wdt <- ceiling(log10(max(dt, na.rm=TRUE)))
  wname <- max(nchar(xnames[!is.na(avgdt[datacol])]),
               nchar(lpnames[!is.na(avgdt[lpcol])]),
               nchar(diwnames[!is.na(avgdt[diwcol])]),
               nchar(cptnames[!is.na(avgdt[cptcol])]))

  if (have.data) {
    cat("  Data Preparation\n")
    for (i in seq_along(datacol)) {
      if (!is.na(avgdt[datacol[i]])) {
        cat(sprintf("    %-*s   %*.0f\n",
                    wname, xnames[i], wdt, avgdt[datacol[i]]))
      }
    }
  }
  
  if (have.lp) {
    cat("  Low-Pass Spacing\n")
    for (i in seq_along(lpcol)) {
      if (!is.na(avgdt[lpcol[i]])) {
        cat(sprintf("    %-*s   %*.0f\n",
                    wname, lpnames[i], wdt, avgdt[lpcol[i]]))
      }
    }
  }
  
  if (have.diw) {
    cat("  Interval Spacing\n")
    for (i in seq_along(diwcol)) {
      if (!is.na(avgdt[diwcol[i]])) {
        cat(sprintf("    %-*s   %*.0f\n",
                    wname, diwnames[i], wdt, avgdt[diwcol[i]]))
      }
    }
  }
  
  if (have.cpt) {
    cat("  Changepoints\n")
    for (i in seq_along(cptcol)) {
      if (!is.na(avgdt[cptcol[i]])) {
        cat(sprintf("    %-*s   %*.0f\n",
                    wname, cptnames[i], wdt, avgdt[cptcol[i]]))
      }
    }
  }

  cat("\n")
}

# Return the timing since the last clear.  This is a matrix with TIMER_*
# in columns and one row per Dimodal run.
get.timers <- function() {

  t <- get("timings", envir=._timer.env)
  # Strip out any incomplete placeholder rows.
  t <- t[!apply(t, 1, function(x) { all(is.na(x)) }),]
  # Add timer names to matrix.
  nm <- ls(envir=._timer.env)[grepl("TIMER", ls(envir=._timer.env), fixed=TRUE)]
  nm <- nm[order(sapply(nm, get, envir=._timer.env))]
  colnames(t) <- nm
  
  t
}

# Register the start time for one of the TIMER_* identifiers tid registered
# in setup.timing.  Modifies the ._timer.env contents.
start.timer <- function(tid) {

  tst <- get("tst", envir=._timer.env)
  tid <- get(tid, envir=._timer.env)
  tst[tid] <- proc.time()['elapsed']
  assign("tst", tst, envir=._timer.env)
}

# Store the timing in ms of the TIMER_* identifier tid listed in setup.timing
# in the history.  Modifies the ._timer.env contents.
# If start.timer has never been called for this tid, the registered time will
# be NA.
stop.timer <- function(tid) {

  tend <- proc.time()['elapsed']
  tst <- get('tst', envir=._timer.env)
  tid <- get(tid, envir=._timer.env)
  dt <- get('timings', envir=._timer.env)
  dt[nrow(dt),tid] <- (tend - tst[tid]) * 1000
  assign('timings', dt, envir=._timer.env)
}

# utility.R:
# A grab-bag of useful helper functions for the Dimodal results.
#
# c 2025-2026 Greg Kreider, Primordial Machine Vision Systems, Inc.

## To Do:
# - 
#

##### Public Interface

# Estimate the quantiles q of data x using piecewise linear segments to
# approximate the quantile function for discrete/tied data, with type
# determining the choice of endpoint (1,3 at middle of jumps in x; 2,3 at
# middle of jumps in quantile/index ; 4 between middle of runs, 0
# auto-determine using 3 for quantized data, 4 else).  The default q is at
# the indices of x and re-maps any tied values.  feps is the relative
# tolerance for matching values as per find.runs.  Returns the estimate,
# a vector the same length  as x.
midquantile <- function(x, q=((1:length(x))-1)/(length(x)-1),
                        type=0, feps=0.0) {

  # The default type is best for non-continuous data, as it runs through
  # single data points.  If there are ties, 3 or 1 would be better.  1 is
  # recommended in the literature but can lie entirely to one side of the
  # data; see test case 1.  3 generates a smaller difference between the
  # actual and approximated data than 1 does.  We use a simple heuristic -
  # if the number of unique values is 10% of the data = quantized - to
  # decide.  signif rounds numeric inaccuracies if x is the spacing.
  if (0 == type) {
    if ((10 * length(unique(signif(x)))) < length(x)) {
      type <- 3
    } else {
      type <- 4
    }
  }

  x <- sort(x)
  s <- .Call("C_midq", x, as.integer(type), feps, PACKAGE="Dimodal")
  # The default q shifts indices to 0-based, including max index/count.
  .Call("C_eval_midq", s$x, s$q, q, PACKAGE="Dimodal")
}

# Convert the find.runs result runs from data x into an rle structure, a
# list with elements
#   $lengths   integer vector with the length of each run
#   $values    corresponding values from x, of the same type as x
# Additionally it adds
#   $nskip     integer vector counting non-finite values within each run
runs.as.rle <- function(runs, x) {

  if (length(runs$runs) != length(x)) {
    stop(paste0("runs generated for data of length ", length(runs$runs),
                " but x has ", length(x), " points"))
  }

  sel <- runs$runs != 0
  lens <- runs$runs[sel]
  vals <- x[sel]
  nskip <- runs$nskip[sel]

  structure(list(lengths=lens, values=vals, nskip=nskip), class="rle")
}


# The Dipeak structure contains minima and peaks that cannot be tested because
# they are at the edge of the data, with a minimum to only one side.  Returns
# only those rows of peak result pk with valid maxima.  Returned value will
# have zero rows if there are no valid peaks, or if pk is not a Dipeak object.
select.peaks <- function(pk) {

  if (!inherits(pk, "Dipeak") || (nrow(pk) <= 2) ||
      !("ismax" %in% colnames(pk))) {
    mockup.Dipeak()
  } else {
    pk[pk$ismax & c(FALSE, rep(TRUE, nrow(pk)-2), FALSE),]
  }
}


# Shift any interval spacing feature positions in Dimodal result m to align
# with the raw data.  For peaks this is the 'pos' column, for flats
# the 'src', 'stID', and 'endID' columns - the original data values have
# already been shifted.   Returns the modified x, unless it does not contain
# both spacings or peaks or flats, in which case return x unchanged.
center.diw <- function(m) {

  a <- attributes(m$data)
  if ("diw.window" %in% names(a)) {
    # This can differ from the shift in print.Didata if both the LP and Diw
    # windows are odd, but we don't require m have low-pass results.
    shft <- -(a$wdiw %/% 2)
    if (("diw.peaks" %in% names(m)) && (0 < nrow(m$diw.peaks))) {
      m$diw.peaks$pos <- pmax(1, m$diw.peaks$pos + shft)
    }
    if (("diw.flats" %in% names(m)) && (0 < nrow(m$diw.flats))) {
      m$diw.flats$src <- pmax(1, m$diw.flats$src + shft)
      m$diw.flats$stID <- pmax(1, m$diw.flats$stID + shft)
      m$diw.flats$endID <- pmax(1, m$diw.flats$endID + shft)
    }
  }

  m
}

# Match low-pass and interval features in Dimodal results m.  Peaks must be
# within near points of each other, or if near is a fraction less than 1
# as that fraction of the data length.  The overlap of flats must be at least
# the fraction foverlap of the length of both.  Returns a list with four
# maps of rows in the low-pass and interval spacing features, using nomatch
# (coerced to integer) if there is no corresponding peak/flat.  These are
#   $peak.lp2diw      $peak.diw2lp
#   $flat.lp2diw      $flat.diw2lp
# Also print the matches unless quiet is TRUE.
match.features <- function(m, near=10, foverlap=0.70, nomatch=NA_integer_,
                           quiet=FALSE) {

  if (near <= 0) {
    stop("near must be positive")
  } else if (near < 1.0) {
    near <- round(near * ncol(m$data))
  }

  if ((foverlap <= 0.0) || (1.0 < foverlap)) {
    stop("foverlap not between 0 (excl.) and 1")
  }

  morig <- m
  m <- center.diw(m)

  nomatch <- as.integer(nomatch)
  
  if (("lp.peaks" %in% names(m)) && ("diw.peaks" %in% names(m))) {
    if (is.finite(nomatch) && is.integer(nomatch) && (1 <= nomatch) &&
        ((nomatch <= nrow(m$lp.peaks)) || (nomatch <= nrow(m$diw.peaks)))) {
      stop("nomatch value may conflict with row numbers in result")
    }
    
    lp <- select.peaks(m$lp.peaks)
    diw <- select.peaks(m$diw.peaks)

    if ((0 == nrow(lp)) || (0 == nrow(diw))) {
      pk.lp <- rep(nomatch, nrow(lp))
      pk.diw <- rep(nomatch, nrow(diw))
    } else {
      pk.lp <- sapply(m$lp.peaks$pos,
                      function(p) {
                        if (p %in% lp$pos) {
                          sep <- abs(p - diw$pos)
                          map <- which.min(sep)
                          if (sep[map] <= near) {
                            which(m$diw.peaks$pos == diw$pos[map])
                          } else {
                            nomatch
                          }
                        } else {
                          nomatch
                        }
                      })
      pk.diw <- match(1:nrow(m$diw.peaks), pk.lp, nomatch=nomatch)
    }

    if (!quiet) {
      print_peakmatch(m$lp.peaks, morig$diw.peaks, pk.lp, near)
    }
  }

  if (("lp.flats" %in% names(m)) && ("diw.flats" %in% names(m))) {
    if ((0 == nrow(m$lp.flats)) || (0 == nrow(m$diw.flats))) {
      ft.lp <- rep(nomatch, nrow(m$lp.flats))
      ft.diw <- rep(nomatch, nrow(m$diw.flats))
    } else {
      if (is.finite(nomatch) && is.integer(nomatch) && (1 <= nomatch) &&
          ((nomatch <= nrow(m$lp.flats)) || (nomatch <= nrow(m$diw.flats)))) {
        stop("nomatch value may conflict with row numbers in result")
      }
      ft.lp <- sapply(1:nrow(m$lp.flats),
                      function(i) {
                        olap <- pmin(m$lp.flats$endID[i], m$diw.flats$endID) -
                                pmax(m$lp.flats$stID[i], m$diw.flats$stID) + 1
                        map <- which.max(olap)
                        longest <- max(m$lp.flats$len[i], m$diw.flats$len[map])
                        if (foverlap <= (olap[map] / longest)) {
                          map
                        } else {
                          nomatch
                        }
                      })
      ft.diw <- match(1:nrow(m$diw.flats), ft.lp, nomatch=nomatch)
    }

    if (!quiet) {
      print_flatmatch(m$lp.flats, morig$diw.flats, ft.lp, foverlap)
    }
  }

  if (!quiet) { cat("\n") }

  list(peak.lp2diw=pk.lp, peak.diw2lp=pk.diw,
       flat.lp2diw=ft.lp, flat.diw2lp=ft.diw)
}

# Move the position/endpoints of peaks or flats feat by offset-1 data points
# to align with the spacing grid (use the 'lp.stID' or 'diw.stID' attributes
# from the gen.data matrix).  Add columns 'x' (for peaks), 'xst, 'xend' to
# the matrix with the position mapped back into the original data using the
# mid-distribution xmid, per the discrete quantile correction of
# Ma/Genton/Parzen.  Returns the modified feat.
shiftID.place <- function(feat, offset, xmid) {

  if (0 == nrow(feat)) {
    return(feat)
  }

  # The points are located on-grid so no need to interpolate in xmid.

  if ("pos" %in% colnames(feat)) {
    feat$pos <- feat$pos + offset - 1
    feat$x <- xmid[feat$pos]
  }
  if ("lminID" %in% colnames(feat)) {
    feat$lminID <- feat$lminID + offset - 1
  }
  if ("rminID" %in% colnames(feat)) {
    feat$rminID <- feat$rminID + offset - 1
  }
  if ("lsuppID" %in% colnames(feat)) {
    feat$lsuppID <- feat$lsuppID + offset - 1
  }
  if ("rsuppID" %in% colnames(feat)) {
    feat$rsuppID <- feat$rsuppID + offset - 1
  }
  if ("src" %in% colnames(feat)) {
    feat$src <- feat$src + offset - 1
  }
  if ("stID" %in% colnames(feat)) {
    feat$stID <- feat$stID + offset - 1
    feat$xst <- xmid[feat$stID]
  }
  if ("endID" %in% colnames(feat)) {
    feat$endID <- feat$endID + offset - 1
    feat$xend <- xmid[feat$endID]
  }

  feat
}


##### Implementation

# Print matching peaks in the Dipeak low-pass data lp and interval spacing diw,
# using the row map lp2diw between the two (NA entries for non-matching rows).
# near is the separation between between maxima to 'match'.
print_peakmatch <- function(lp, diw, lp2diw, near) {

  cat("\n  matching peaks within ", near, " points after centering intervals\n")
  if ((0 == length(lp2diw)) || !any(is.finite(lp2diw))) {
    cat("    none\n")
    return()
  }

  dgt <- Diopt()$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }

  sel.lp <- which(is.finite(lp2diw) &
                  (0 < lp2diw) & (lp2diw <= length(diw$pos)))
  sel.diw <- lp2diw[sel.lp]
      
  spos.lp <- format(lp$pos[sel.lp])
  spos.diw <- format(diw$pos[sel.diw])
  wpos <- max(nchar(spos.lp), nchar(spos.diw), nchar("pos"))

  sx.lp <- format(lp$x[sel.lp], digits=dgt)
  sx.diw <- format(diw$x[sel.diw], digits=dgt)
  wx <- max(nchar(sx.lp), nchar(sx.diw), nchar("raw"))

  spass.lp <- format(lp$naccept[sel.lp])
  spass.diw <- format(diw$naccept[sel.diw])
  wpass <- max(nchar(spass.lp), nchar(spass.diw), nchar("#pass"))
  
  wcol <- wpos + wx + wpass + 8
  cat(paste0("   ", format("low-pass", width=wcol, justify="c"), "   ",
             "   ", trimws(format("interval", width=wcol, justify="c"), "right"),
             "\n"))
  shdr <- sprintf("   %*s    %*s    %*s",
                  wpos, "pos", wx, "raw", wpass, "#pass")
  cat(paste0(shdr, "   ", shdr, "\n"))
  for (i in seq_along(sel.lp)) {
    slp <- sprintf("   %*s   (%*s)   %*s",
                   wpos, spos.lp[i], wx, sx.lp[i], wpass, spass.lp[i])
    sdiw <- sprintf("   %*s   (%*s)   %*s",
                    wpos, spos.diw[i], wx, sx.diw[i], wpass, spass.diw[i])
    cat(paste0(slp, "   ", sdiw, "\n"))
  }
}

# Print matching flats in the Dflat low-pass data lp and interval spacing diw,
# using the row map lp2diw between the two (NA entries for non-matching rows).
# foverlap is the minimum fraction of each flat's length that is common.
print_flatmatch <- function(lp, diw, lp2diw, foverlap) {

  cat(sprintf("\n  matching flats each overlapping by %.2f after centering intervals\n",
              foverlap))
  if ((0 == length(lp2diw)) || !any(is.finite(lp2diw))) {
    cat("    none\n")
    return()
  }

  dgt <- Diopt()$digits
  if (0 == dgt) {
    dgt <- options()$digits
  }
  
  sel.lp <- which(is.finite(lp2diw))
  sel.diw <- lp2diw[sel.lp]

  sst.lp <- format(lp$stID[sel.lp])
  send.lp <- format(lp$endID[sel.lp])
  spos.lp <- sapply(seq_along(sst.lp),
                    function(i) { paste0(sst.lp[i], " - ", send.lp[i]) })
  sst.diw <- format(diw$stID[sel.diw])
  send.diw <- format(diw$endID[sel.diw])
  spos.diw <- sapply(seq_along(sst.diw),
                     function(i) { paste0(sst.diw[i], " - ", send.diw[i]) })
  wpos <- max(nchar(spos.lp), nchar(spos.diw), nchar("pos"))

  sxst.lp <- format(lp$xst[sel.lp], digits=dgt)
  sxend.lp <- format(lp$xend[sel.lp], digits=dgt)
  sx.lp <- sapply(seq_along(sxst.lp),
                  function(i) { paste0("(", sxst.lp[i], " ", sxend.lp[i], ")") })
  sxst.diw <- format(diw$xst[sel.diw], digits=dgt)
  sxend.diw <- format(diw$xend[sel.diw], digits=dgt)
  sx.diw <- sapply(seq_along(sxst.diw),
                   function(i) { paste0("(", sxst.diw[i], " ", sxend.diw[i], ")") })
  wx <- max(nchar(sx.lp), nchar(sx.diw), nchar("raw"))
            
  spass.lp <- format(lp$naccept[sel.lp])
  spass.diw <- format(diw$naccept[sel.diw])
  wpass <- max(nchar(spass.lp), nchar(spass.diw), nchar("#pass"))

  wcol <- wpos + wx + wpass + 8
  cat(paste0("  ", format("low-pass", width=wcol, justify="c"),
             "  ", trimws(format("interval", width=wcol, justify="c"), "right"),
             "\n"))
  shdr <- sprintf("   %s  %s  %*s",
                  format("pos", width=wpos, justify="c"),
                  format("raw", width=wx, justify="c"), wpass, "#pass")
  cat(paste0(shdr, "   ", shdr, "\n"))
  for (i in seq_along(sel.lp)) {
    slp <- sprintf("   %*s  %*s  %*s",
                   wpos, spos.lp[i], wx, sx.lp[i], wpass, spass.lp[i])
    sdiw <- sprintf("   %*s  %*s  %*s",
                    wpos, spos.diw[i], wx, sx.diw[i], wpass, spass.diw[i])
    cat(paste0(slp, "   ", sdiw, "\n"))
  }
}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{find.peaks}
\alias{find.peaks}
\title{
Local minima and maxima detector.
}
\description{
Locate local minima and maxima within data, removing small peaks or shallow
minima.
}

\usage{
find.peaks(x, fht, frelht, fhtie, fhsupp)
}

\arguments{
\item{x}{
  a vector of real or integer values
}
\item{fht}{
  minimum absolute peak height as fraction (0 - 1 excl.) of data range
}
\item{frelht}{
  minimum relative peak height as fraction (0 - 1 excl.) of average value
}
\item{fhtie}{
  relative fraction (0 - 1 excl.) of values to consider as same
}
\item{fhsupp}{
  fraction (0 - 1 incl.) of max-min heights for support
}
}

\details{
The detector begins by using \code{find.runs} to identify sequences of nearly
equal values, in case there are flat tops at peaks or in valleys at minima.
It treats runs as a single point when marking all triples as local minima
(x[i] > x[j] < x[k]) or maxima (x[i] < x[j] > x[k]).  It determines the
smallest difference between a maximum and its adjacent minima, both as an
absolute value and as a fraction of their average,
\code{|x[j] - x[i or k]| / ((|x[j]| + |x[i or k]|)/2)}, for \code{i} and
\code{k} the minima to the left and right of the maximum \code{j}.  Working
from the smallest absolute difference to the largest, the detector checks if
either the absolute height is below the fraction \code{fht} of the data's
range or if the relative height is below \code{frelht}.  If so, the peak is
merged into its neighbor by deleting it and the minimum and re-calculating
the height of the neighbor.

The first and last data point are always extrema, a maximum if they are larger
than their neighbors.

We may want to limit the extent of a peak between two minima, for example
to avoid one that lies in the middle of a long flat and that makes the peak
appear much wider than it is.  Let the support of a peak be the points whose
value drops less from the peak than the fraction \code{fhsupp} of the
feature's height.  If 1.0, the support extends to the minima.  If 0 the
support is only the local maximum.  The Full Width at Half Maximum (FWHM)
measure would use set fhsupp=0.5.  a larger value, for example 0.9, will back
away from a minimum in a flat without distorting the feature for testing.

Although the order of checking and merging peaks can affect the final set,
and the process is deterministic, we do not document what this order is.
If there are multiple local extrema with the same value, we do not guarantee
which will be selected, although the result will be stable over repeated calls.

The arguments correspond to \code{Diopt} options \opt{peak.fht},
\opt{peak.frelht}, \opt{peak.fhtie}, and \opt{peak.fhsupp}.  The caller
provides them; this function does not access \code{Diopt} directly.

This function is not exported from the Dimodal package, but may be useful on
its own for any signal.  In Dimodal it is followed by a call to
\code{shiftID.place} to move indices to the original data grid.

The detector is O(\code{n log n}) in time and O(\code{n}) in memory.
}

\value{
find.peaks returns a \code{"Dipeak"} object.  If the data is bad
(constant or has two or fewer points or is all NA) then the result will be
empty and have zero rows.
}

\seealso{
 \code{\link{find.runs}},
 \code{\link{Dipeak}},
 \code{\link{Diopt}},
 \code{\link{shiftID.place}}
}

\keyword{Dimodal}
\keyword{peak detector}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Ditest}
\alias{Ditest}
\alias{print.Ditest}
\alias{summary.Ditest}
\title{
Class Methods for Ditest Objects
}
\description{
Common functions to handle \opt{Ditest} class results from the Dimodal
feature tests.
}

\usage{
\method{print}{Ditest}(x, \dots)
\method{summary}{Ditest}(object, \dots)
}

\arguments{
\item{x}{an object of class \opt{Ditest}}
\item{object}{an object of class \opt{Ditest}}
\item{\dots}{extra arguments, ignored for all methods}
}

\details{
The values of the individual tests specify the contents of the object.
These functions provide a common formatting of the result.  The print
methods show the test applied and all parameters, as well as the result.
The summary methods show the tested statistic and the resulting probability.
}

\value{
the argument \code{x} or \code{object}, invisible
}

\seealso{
  \code{\link{Dipeak.test}},
  \code{\link{Diflat.test}},
  \code{\link{Dinrun.test}},
  \code{\link{Dirunlen.test}},
  \code{\link{Diexcurht.test}},
  \code{\link{Dipermht.test}}
}

\examples{
ptst <- Dipeak.test(1:4, 250, 0.15, 'bartlett')
ptst
## Summary results.
summary(ptst)
}

\keyword{Ditest}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{find.flats}
\alias{find.flats}
\title{
Local flat detector.
}
\description{
Locate flat sections, specified by the ripple parameter, within data.
}

\usage{
find.flats(x, fripple, minlen, fminlen, noutlier)
}

\arguments{
\item{x}{
  a vector of real or integer values
}
\item{fripple}{
  height of flat as fraction (0 -- 1 excl.) of data range
}
\item{minlen}{
  minimum length of flat, in data points (absolute)
}
\item{fminlen}{
  minimum length of flat, as fraction (0 -- 1 excl.) of data length (relative)
}
\item{noutlier}{
  number of points outside the ripple range allowed to either side of source
}
}

\details{
The detector scans every point in \code{x} to find the contiguous range of
indices that lie within the ripple centered at the point's value.  The
algorithm allows noutlier points to each side of the source point (so the
total can be twice this), but not as the first or last points.  In other
words, the endpoints of the flat will be within the ripple specification.
Flats must have a length of the larger of \code{minlen} or the relative
\code{fminlen} of \code{length(x)}, counting that portion not covered by a
longer flat.  Flats may therefore overlap, but only if the extension meets
the length requirement.  Flats are reported for their original extent and
not the exposed portion.

The algorithm will ignore non-finite values in \code{x}, but the endpoints
of the features apply to the original, unscreened data.

The arguments correspond to options \opt{flat.fripple}, \opt{flat.minlen},
\opt{flat.fminlen}, and \opt{flat.noutlier}.  These are provided when
called within Dimodal; this function does not access \code{Diopt} directly.

This function is not exported from the Dimodal package, but may be useful on
its own for any signal.  Within Dmodal it is followed by a call to
\code{shiftID.place} to move indices to the original data grid.

The detector switches its scan algorithms based on the data size.  The
base approach, which scans outward from each source point, is
O(\code{n Lflat}) in time and O(\code{n}) in memory, where Lflat is the
average length of all ranges.  This is usually a substantial fraction of the
data, a third to half, so the algorithm is essentially O(\code{n^2}).  For
larger data sets it switches to a segtree-based search that has a high
overhead.  The time complexity is not known, and the memory consumption is
still O(\code{n}).  The second approach runs faster with more than
10 thousand data points, and is an order of magnitude quicker for
100 thousand points.
}

\value{
find.flats returns a \code{"Diflat"} object.  If there are no flats in the
data then the result will be empty and have zero rows.
}

\seealso{
 \code{\link{Diflat}},
 \code{\link{Diopt}},
 \code{\link{shiftID.place}}
}

\keyword{Dimodal}
\keyword{flat detector}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{find.runs}
\alias{find.runs}
\title{
Fuzzy run detector
}
\description{
A runs length detector that handles nearly equal real values and skips
invalid values.
}

\usage{
find.runs(x, feps)
}

\arguments{
\item{x}{
  a vector
}
\item{feps}{
  fractional (0 - 1 excl.) relative difference to treat values as same
}
}

\details{
This runs finder looks for sequences of real values that almost match,
considering them as a run of same values.  Two points match if the
difference between their values as a fraction of their average is less than
the threshold: |x[i] - x[j]| / ((|x[i]| + |x[j]|)/2) < feps.  For each
point in \code{x} the detector scans forward until a point fails to be close,
with the run taken over the matching interval.  The base of the comparison is
always the first point in the run, and not a chain of adjacent values.  The
detector treats infinities as equal to each other but not to finite values
and skips over any NA or NaN values while counting them.  Values within
double precision tolerance of each other always match, irrespective of the
relative threshold.

Note that this relative difference depends on the data not averaging close
to zero, especially at zero-crossing points, and is sensitive to a constant
shift.  The first is not a problem for spacing, which is always positive,
but may require shifting other types of data.  That would also help with the
second issue.

For integer, logical, and character data, values must match exactly and feps
is not used.

Other data types are not supported.

Although not exported from the Dimodal package, this detector may be useful
outside the spacing analysis for any signal.  Within Dimodal it is called
by the run count and length tests and internally within the C code.

The detector returns two vectors, each with the same length as the original
data.  \code{runs} stores the length of the run starting at the data point
and \code{nskip} the number of skipped elements within the run.  To process
the runs, start at index \code{i=runs[1L]}, or if that is zero
\code{i=skip[1L]}.  The next run starts at \code{i + runs[i] + nskip[i]},
until i goes beyond the length of the data.

The detector is O(\code{n}) in both time and memory, albeit with several
passes through the data.
}

\value{
find.runs returns a list with elements:
\item{\code{runs}}{
  an integer vector of length x with the length of
  the run starting at each point or 0 if the point does not start a run
}
\item{\code{nskip}}{
  an integer vector of length x with the number of
  skipped data points within the run
}
\item{\code{stats}}{
  a vector of three integer values, \code{nrun} the total number of runs,
  \code{maxrun} the longest run, and \code{nx} the length of the original
  data and of \code{runs} and \code{stats}
}
}

\note{
Use the utility function \code{runs.as.rle} to convert the returned value to
a \opt{rle} encoding.
}

\seealso{
 \code{\link{rle}},
 \code{\link{runs.as.rle}}
}

\keyword{Dimodal}
\keyword{runs}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dipeak}
\alias{Dipeak}
\alias{print.Dipeak}
\alias{summary.Dipeak}
\alias{mark.Dipeak}
\title{
Class methods for Dipeak Objects
}
\description{
Common functions to handle \code{"Dipeak"} class results, which store
information about local extrema and their probability.
}

\usage{
\method{print}{Dipeak}(x, \dots)
\method{summary}{Dipeak}(object, \dots)
\method{mark}{Dipeak}(x, hist=NULL, opt=Diopt(), \dots)
}

\arguments{
\item{x}{ an object of class \code{"Dipeak"} }
\item{object}{ an object of class \code{"Dipeak"} }
\item{hist}{ the histogram being marked or NULL for a low-pass or interval
  spacing graph }
\item{opt}{ local version of options to guide plots }
\item{\dots}{ extra arguments, ignored for all methods
}
}

\details{
A \opt{Dipeak} object is a data frame that describes local extrema, including
their position in the low-pass or interval spacing and equivalent raw value,
their height and width, and any tests that have run to judge the significance
of the feature.  The basic information comes from the \code{find.peaks}
detector, and the \code{Dimodal} analysis adds the feature analysis.  The
class inherits from \opt{data.frame}.

The print method shows the position (index) and raw value of each maximum,
the minimum to either side, and if different the support.  It then has
two tables with the statistic (test value) and probability for any tests
that have been run, as well as the acceptance levels.  These come from the
options \opt{alpha.ht} for the model test, \opt{alpha.pkexcur.lp} and
\opt{alpha.pkexcur.diw} for the excursion test, and \opt{alpha.nrun} and
\opt{alpha.runlen} and \opt{alpha.runht} for the runs tests.  If the
option \opt{mark.alpha} is TRUE then probabilities at or below the
acceptance level are underlined.  Set the option FALSE if your terminal
does not support ANSI escape codes, or see the \pkg{crayon} package for the
extensive tests needed to determine this automatically.  These options are
set at the global level and cannot be overridden for a single call to print.
The pass column after the overall peak probability will contain the number
of accepted/passing tests, if any.  Raw values and statistics are printed
with option \opt{digits} significant digits.  If the option is set to
zero then the standard \code{options()$digits} is used.

The summary method lists the position and raw value of each maximum, the
overall probability or best result from all tests, and a list of the tests
that pass per the \code{Diopt} acceptance levels.

Peaks and minima are marked in a spacing graph by vertical lines.  Maxima
are drawn with dashes and are made thick if the peak passes any test.
Minima are drawn with dotted lines.  The marks are colored per the option
\opt{colID.peak} of the current palette.  Histogram annotations include
only lines for peaks, drawn thick if significant.  They run from the quantile
axis on the right to the distribution function curve, then down to the data
value.  Marking the histogram assumes that Dimodal has added the \code{x}
column with the peak's original data value and has shifted positions.

\code{select.peaks} is a utility function that extracts only those rows with
valid peaks, which have a minima to each side.
}

\value{
The basic data frame returned from \code{find.peaks} includes the columns
 \item{\code{pos}}{ index in data of local extremum }
 \item{\code{ismax}}{ TRUE if a local maximum, FALSE if a minimum }
 \item{\code{valsd}}{ data at extremum, standardized by \code{sd(x)} }
 \item{\code{lht}}{ difference in valsd between maximum and minimum to its
   left, NA if no minimum or if a minimum}
 \item{\code{rht}}{ difference in valsd between maximum and minimum to its
   right, NA if no minimum or if a minimum }
 \item{\code{lminID}}{ position/index of minimum to the left, NA if not a
   valid maximum }
 \item{\code{rminID}}{ position/index of minimum to the right, NA if not a
   valid maximum }
 \item{\code{lsuppID}}{ left index of support, NA if not a valid maximum }
 \item{\code{rsuppID}}{ right index (incl.) of support, NA if not a valid
   maximum \nl }
The indices are located in the data passed to the detector.  Valid maxima
have minima on both sides.  There will always be two extrema at the first
and last point.

\code{Dimodal} adds several columns.  It shifts all indices to the original
data, compensating for the invalid points skipped during low-pass filtering
or interval spacing.  Mapping indices back to the raw data and summarizing
the tests, it adds
 \item{\code{x}}{ the location of the peak in the original data }
 \item{\code{ppeak}}{ the best probability of all tests run or NA if none
   or if not a maximum with minima to both sides }
 \item{\code{naccept}}{ number of tests passing their acceptance level,
   0 if none or ppeak is NA
   \nl }
x can be generated by calling the utility function \code{shiftID.place} on
the find.peaks result.  Dimodal also adds an attribute \opt{source} with
value "LP" or "Diw" depending on the filter used.

For low-pass testing the results may include
 \item{\code{pht}}{ the probability of the feature per the height model,
   using the larger of \code{lht} and \code{rht} }
 \item{\code{hexcur}}{ the feature height used for the excursion test }
 \item{\code{pexcur}}{ the probability of the peak using bootstrap sampling
   of the difference of the low-pass spacing
   \nl }

For interval spacing tests the results may include those based on runs in its
signed difference, between the bounding minima
 \item{\code{nrun}}{ the number of runs }
 \item{\code{pnrun}}{ the probability of nrun using the Kaplansky-Riordan
   test }
 \item{\code{runlen}}{ the length of the longest run }
 \item{\code{prunlen}}{ the probability of runlen using a Markov
   chain model of the entire data sample }
 \item{\code{runht}}{ the feature height in the signed difference }
 \item{\code{prunht}}{ the probability of the peak with height runht
   using a permutation test of the runs }
 \item{\code{hexcur}}{ the feature height in the interval spacing between
   minima }
 \item{\code{pexcur}}{ the probability of the peak with height hexcur
   using bootstrap sampling from the interval spacing of the whole data
   \nl }

The methods return the object, invisibly.
}

\seealso{
  \code{\link{find.peaks}},
  \code{\link{Dimodal}},
  \code{\link{Diopt}},
  \code{\link{data.frame}},
  \code{\link{select.peaks}}
  \code{\link{shiftID.place}}
}

\examples{
m <- Dimodal(faithful$eruptions, Diopt.local(diw.window=16))
summary(m$lp.peaks)
m$diw.peaks
}

\keyword{Dipeak}
\keyword{Dimodal}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal Excursion and Permutation Tests}
\alias{Diexcurht.test}
\alias{Dipermht.test}
\title{
Feature significance test using permutation or bootstrap (excursion) sampling.
}
\description{
Repeatedly draw samples from a base set of differential values, with or
without replacement, and regenerate the signal, measuring its height or
range.  Compare the feature height against this distribution to estimate
its probability.
}

\usage{
Diexcurht.test(ht, ndraw, xbase, nexcur, is.peak, lower.tail=TRUE, seed=0)
Dipermht.test(ht, xbase, nperm, lower.tail=TRUE, seed=0)
}

\arguments{
\item{ht}{
  one or more feature heights
}
\item{ndraw}{
  the size of the feature, a vector as long as ht or a single value
}
\item{xbase}{
  a vector of values to draw from, differential values that can be
  summed to reconstruct the signal
}
\item{nexcur}{
  the number of times to repeat the draw
}
\item{nperm}{
  the number of times to permute the draw
}
\item{is.peak}{
  a boolean, TRUE to calculate height for peaks, FALSE for flats
}
\item{lower.tail}{
  a boolean, TRUE to count the number of simulated heights greater than ht,
  FALSE the number less than or equal to
}
\item{seed}{
  if positive, value to set the RNG seed before any sampling
}
}

\details{
The excursion test determines the distribution of a feature's height by
sampling from a set of differential data and doing a cumulative sum.  The
reconstructed feature height for a peak is the maximum sum above the lower
of the first or last point, or the total range for flats, using \code{is.peak}
to choose.  As a bootstrap excursion this can be used with the difference
of the low-pass or interval spacing for either peaks or flats.  The
resolution of the probability returned from the test is \code{1/nexcur}.

The permutation test shuffles the values in \code{xbase} before re-summing
them to simulate the signal.  The test automatically uses the peak height
definition.  The base set should be the length of the runs within the
feature, positive or negative or zero (ignoring length).  If generated
using \code{rle}, multiply \code{$lengths} by \code{$values}.  Permutations
are restricted so that adjacent values do not have the same sign, because
this would form a longer run.  This eliminates a large fraction of the
possible permutations, 90\% or more.  If 5\% of the possible number of
permutations, the factorial of the length of \code{xbase}, is less than the
requested sample size, the test will exhaustively check all possibilities
rather than sampling.  For nperm=5000 this happens at eight runs.

These tests are general and do not depend on the type of feature or spacing
being analyzed.  The caller, in this library the \code{Dimodal} function,
is responsible for identifying features and their height and setting up
the base set and draw size.  NA or NaN heights will generate NA
probabilities, as do 0 heights and draw sizes that are too small (minimum
3 for excursions, 2 for permutations).  All base values must be finite.
Bad argument values will raise errors.

If more than one height is provided the tests will evaluate them against
one set of draws, unless the draw sizes also vary, in which case pairs of
the arguments will be used.

A FALSE \code{lower.tail} is appropriate for peaks, TRUE for flats.  The
probability can have a large uncertainty if the tests generate tied values
matching the height.  This is mostly seen with the runs permutations, which
generate discrete heights.  The count uses half the ties, which is a simple
form of mid-distribution correction.

The RNG seed, if not zero, is added to the draw counter and used to set up
the random number generator before sampling, so that the results are
repeatable.  A value of zero leaves the setup alone.

The \code{nexcur} argument corresponds to option \opt{excur.nrep} and
\code{seed} to \opt{excur.seed}.  Dimodal uses \opt{excur.ntop} when
generating xbase for excursions.  The equivalents for the permutation test
are \opt{perm.nrep} and \opt{perm.seed}.  The default number of excursions
is higher than for permutations, based on an evaluation of the stability on
artificial and real data.  The distribution of the excursion heights is slow
to settle, and has a large confidence interval; by 15,000 repetitions the
median height has settled to within half a percent and the 90\% confidence
interval is two percent of that median.  Fewer repetitions may suffice,
although below 5,000 you may notice the variation in the test probabilities.
The permutation test is more stable and fewer repetitions are needed, so that
count has been given its own option.  Internally Dimodal will adjust the
excursion seed differently for each of the peak and flat tests, so that the
results will be repeatable but different sequences will be used for each.
The probabilities will be evaluated against \opt{alpha.pkexcur.[lp|diw]},
\opt{alpha.runht}, and \opt{alpha.ftexcur.[lp|diw]}.

Each draw runs in \code{O(ndraw)} time and memory, where ndraw is the size of
the xbase set for the permutations.  The permutation test takes noticeably
longer with more than two symbols (ie. if there are ties) because a different
sampling strategy is needed to guarantee the non-adjacency condition.  Unless
the package has been compiled to use the R sampling function we use the PCG
random number generator from Melissa O'Neill for the excursion draws because
it is much faster than the built-in sampling while offering greater bit depth.
}

\value{
Both functions return a list of class \opt{Ditest} with elements
\item{method}{ a string describing the test }
\item{statfn}{ function used to evaluate significance level/probability }
\item{statistic}{ what is tested, the height of the feature }
\item{statname}{ text string describing the statistic }
\item{parameter}{ other arguments provided to the function, named as such }
\item{p.value}{ probability of feature }
\item{alternative}{ a string describing the direction of the test vs. the
  null distribution }
\item{critval}{ critical values of the height, at quantiles 0.005, 0.01,
  0.05, and 0.10, or 1-these if lower.tail is FALSE }
\item{xbase}{ a copy of the argument
  \nl }

\code{statistic} and \code{p.value} will have the length of the \code{ht}
argument.  NA or 0 heights or draws will generate NA for the probabilities.
}

\seealso{
\code{\link{Dimodal},
\link{Diopt},
\link{rle}},
\code{\link{.Random.seed}}
}

\references{
\url{https://pcg-random.org} for the PCG random number generator.
}

\examples{
## Recommended number of excursions/permutations is larger.
## 2000 reduces the run-time.
set.seed(2) ; x <- round(runif(50,-1,1) * 10) / 10
Diexcurht.test(6.5, 30, x, 2000, TRUE, lower.tail=FALSE, seed=3)
xrun <- rle(sign(diff(x)))$values * rle(sign(diff(x)))$lengths
Dipermht.test(6.5, xrun, 2000, FALSE, 3)
}

\keyword{Dimodal}
\keyword{modality}
\keyword{low-pass spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{find.level.sections}
\alias{find.level.sections}
\title{
Detector for intervals without significant slope
}
\description{
An inversion of the \pkg{modehunt} package's test for sloping sections.
}
\usage{
find.level.sections(x, alpha, correct)
}
\arguments{
\item{x}{
  a vector of real or integer data
}
\item{alpha}{
  the significance level of the level test, between 0 and 1 excl.
}
\item{correct}{
  a boolean whether to bias the test against short sections
}
}

\details{
The \pkg{modehunt} test for sloping sections sums interval spacings of
different widths at each point in x, scaling them by a score function to
get a test statistic.  Comparing the test statistic to a critical value
taken from 100 thousand uniform samples decides if the data has non-zero
slope at the point.  The algorithm combines these into sloping segments and
returns the longest common subset.

This detector performs the same test to find the sloping sections, but
processes them differently to identify the shortest non-sloping section
beginning at each point.  These sections may overlap.

Typical parameter values are 0.95 for \code{alpha} and TRUE for
\code{correct}.  Larger values of alpha will produce fewer sections.
They will generally span all of the data.

The test uses a model of the critical value for the test statistic,
to avoid the simulated draws made in \pkg{modehunt}.  The model uses the
length of the data and the significance level, and has been generated for
lengths from 50 to 5000 and alpha from 0.00001 to 0.99999.  Accuracy outside
this range cannot be guaranteed.

The level detector was developed to find flats without (explicitly) filtering
the data, but the results were poor.  Many sections are short and
concentrated near the start and end of the sorted data.  The detector is
sensitive to variations in the data and the division is noisy.  The sections
that are longer do not identify single modes, and tend to span inter-modal
transitions.  Any flats in the spacing are a subset of the level sections,
in count and in length.  There is also a bias towards sections at the start
of the data rather than the end due to how the algorithm incrementally
chooses the starting point of the next section.  In fact, the detector places
a minimum length of 5 on intervals at the end to prevent fragmenting the last.
The detector cannot be used on its own to locate modes.

The analysis is O(\code{n^2}) in time and O(\code{n}) in memory.  If the
data contains integers they are converted to reals.  Other data types are
not supported.  The data is sorted internally.

Although not exported from the Dimodal package, this detector may be useful
outside the spacing analysis for any signal.
}

\value{
find.level.sections returns a data frame with one row per section and
columns
 \item{\code{stID}}{ the index in the sorted x of the section's starting
   endpoint }
 \item{\code{endID}}{ the index in the sorted x of the last endpoint, incl.
   \nl }
The indices refer to the sorted input after removing non-finite values.
They must be mapped back to \code{x} with the mid-distribution function.
There is no way to judge the significance of any section.
}

\seealso{
 \code{modeHunting} in the \pkg{modehunt} package
}

\keyword{Dimodal}
\keyword{runs}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal Runs Tests}
\alias{Dinrun.test}
\alias{Dirunlen.test}
\title{
Tests of runs within subsets of data
}
\description{
Check the significance of a subsequence of data based on the number of runs or
the length of the longest.  These tests work on any data with a limited number
of symbols.  They are used within the Dimodal package to test peaks in the
signed difference of the interval spacing, which has values -1, 0, and +1.
}

\usage{
Dinrun.test(x, stID, endID, feps, lower.tail=TRUE)
Dirunlen.test(x, stID, endID, feps)
}

\arguments{
\item{x}{
  a vector with a limited number of distinct values, including numeric,
  integer, character, factor, or logical
}
\item{stID}{
  a scalar or vector of data indices of the start indices within x of
  the features
}
\item{endID}{
  a scalar or vector of data indices of the end indices (incl.) of the
  features
}
\item{feps}{
  closeness of tied real values, as per \code{find.runs}, and ignored for
  other types
}
\item{lower.tail}{
  a boolean, if TRUE the test returns the probability that the run count is
  not more than the observed, if FALSE is more than
}
}

\details{
\code{Dinrun.test} compares the number of runs within each sequence defined
by the endpoints to the expected, based on a combinatorial counting of all
possible sequences of the symbols.  The number of runs is distributed
normally, with an expected value and variance that depends on the number of
each symbol.  Wolf and Wolfowitz derived formulas for these values in the
case of two symbols, and Kaplansky and Riordan generalized this to
arbitrary sets of symbols.  This test implements that general version.  Its
value is the probability of getting the actual number of runs or fewer, i.e.
the lower tail.

Filtering introduces correlation between symbols, which we can account for
by using a Markov chain model.  The length of a run can be estimated by
separating the symbol transition matrix into two parts, the diagonal which
generates a matching symbol and the off-diagonal elements which switch to
another.  A new Markov chain modeling a run uses these two sub-matrices with
the advancing steps placed on the diagonal of the run's transition matrix
and the resetting steps in the first column.  An absorbing state, or identity
matrix, captures all runs longer than that being tested.  The run length
probability follows from the chance of entering this absorbing state after
stepping the new chain over the length of the feature.  This amounts to a
recursion with the sub-matrices followed weighting by the steady-state or
symbol's stationary state vector to sum over all possible starting symbols.
We assume the symbol's Markov chain has order one.

These tests use \code{find.runs} to determine the number of runs and longest
within the endpoints, and ignores any NA and NaN values in \code{x}.  Its
\code{feps} argument is used to consider which real values are the same.  NA
and NaN start or end indices generate NA statistics and probabilities, so
that the features from \code{Dipeak} and \code{Diflat} can be passed
directly.  Note that if doing so, and passing the difference of the spacing
as \code{x}, \code{stID} should be increased by 1 to account for the point
lost in the difference.  Numeric indices are rounded to integers, and other
types or values that are out of bounds to the vector raise errors.  If the
runs are based on a discrete set of values, as they are in Dimodal, then
feps can be set to 0, otherwise the option \opt{peak.fhtie} could be used.

The runs statistic test involves an O(\code{n}) scan of each sequence to
count the number of symbols within; there is therefore little advantage to
calling this with vectors of indices, rather than one making separate calls
one feature at a time.  The longest runs test requires estimating the
transition matrix over all data, which is O(\code{n}), and this overhead
can be shared by calling it once for all features.  The actual recursion
is expensive, involving O(\code{2 l^2 n}) matrix multiplications and
additions, where \code{l} is the longest run length and \code{n} the
sequence length; the matrix size equals the number of symbols (3x3 for the
signed difference).

The probabilities should be evaluated against options \opt{alpha.nrun} and
\opt{alpha.runlen} for the minimum passing level.
}

\value{
\code{Dinrun.test} and \code{Dirunlen.test} return lists of class
\opt{Ditest} with elements
\item{method}{a string describing the test}
\item{statfn}{function used to evaluate significance level/probability}
\item{statistic}{what is tested, the number of runs or maximum length}
\item{statname}{text string describing the statistic}
\item{parameter}{other distribution arguments, for the runs test \code{Erun}
  the expected count and \code{Vrun} its variance, for the maximum run length
  \code{featlen} the feature length}
\item{p.value}{probability of feature}
\item{tmat}{for the run length test, the transition matrix of the Markov chain}
\item{wt}{for the run length test, the weight applied to starting to each state
  \nl}

\code{statistic}, \code{parameter}, and \code{p.value} are vectors with the
length of the stID and endID vectors.
}

\references{
A. Wald, J. Wolfowitz (1940),
On a test whether two samples are from the same population.
\emph{The Annals of Mathematical Statistics} 11, pp. 147--162.

I. Kaplansky and J. Riordan (1945),
Multiple matching and runs by the symbolic method,
\emph{The Annals of Mathematical Statistics}, 16, pp. 272--277.
}

\seealso{
 \code{\link{find.runs}}
}

\examples{
## The inner diff generates the spacing, the outer the signed difference.
xrun <- sign(diff( diff(sort( iris$Petal.Width )) ))
## No epsilon needed for signed values.
Dinrun.test(xrun, 1, length(xrun), 0)
Dirunlen.test(xrun, 1, length(xrun), 0)
}

\keyword{Dimodal}
\keyword{runs}
\keyword{Markov Chain}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal Model Tests}
\alias{Dipeak.test}
\alias{Dipeak.critval}
\alias{Diflat.test}
\alias{Diflat.critval}
\title{
Significance models of features in the low-pass spacing.
}
\description{
Return the probability of the characteristic feature value, the height of a
peak or length of a flat, using parametric models developed for the low-pass
spacing, or determine the feature value at some significance level.
}

\usage{
Dipeak.test(ht, n, flp, filter, lower.tail=TRUE)
Dipeak.critval(pval, n, flp, filter)
Diflat.test(len, n, flp, filter, basedist, lower.tail=TRUE)
Diflat.critval(pval, n, flp, filter, basedist)
}

\arguments{
\item{ht}{
  difference(s) between the standardized data value at the peak and
  deepest minimum to either side
}
\item{len}{
  length(s) of flat in data points
}
\item{pval}{
  the significance level(s) to find the corresponding height or length,
  the quantile of the feature value is \code{1-pval}
}
\item{n}{
  number of data points before filtering
}
\item{flp}{
  the size of the FIR kernel, either as a fraction of \code{n} or as an
  integer
}
\item{filter}{
  the FIR kernel used to smooth the spacing
}
\item{basedist}{
  for the flat models, the distribution used to generate the length quantiles
}
\item{lower.tail}{
  a boolean, if TRUE the test returns the probability the null distribution
  is less than or equal to the feature value, if FALSE greater than
}
}

\details{
The test functions convert the feature value into a quantile or significance
level based on null distribution models.  The critval functions do the
opposite.  The models are parametric because they are built on draws of
specifically chosen variates and the size of features that appear after
low-pass filtering the data.  The features depend on the size of the
draw \code{n} and the smoothing done, set by the Finite Impulse Response
(FIR) \code{filter} and the size \code{flp} of the kernel.  Implicitly they
depend on the feature detectors, but variations in the parameters controlling
those have neither been studied nor incorporated in the model.

The peak height model comes from draws of an asymmetric Weibull variate with
scale 2 and shape 4, which proved to give reasonable, conservative quantiles
against other distributions.  The preferred filter uses a Kaiser kernel.
The other filters available, the Bartlett or triangular (synonyms), Hanning,
Hamming, Gaussian or normal (synonyms), and Blackman kernel, are handled by
scaling the Kaiser model.  The filter size is typically expressed as a
fraction of the draw size, with \code{flp=0.15} a good default; spans in
data points are also accepted.  Smaller kernels will produce rougher data
with more peaks and fewer flats and can be tolerated if the spacing is
already smooth, as happens with very large data sets.  The test height for
the model is scaled by the standard deviation of the total signal.

The peak test models the distribution of heights with an inverse Gaussian,
a.k.a. Wald distribution.  The height is corrected for the filter and its
size, and the inverse Gaussian location and scale parameters depend on the
data and filter sizes.  These values are provided in the returned list.

The flat length model varies much more with the parametric distribution
chosen as the base, and the recommended \code{basedist}, a logistic variate,
is a compromise.  Models for normal or Gaussian (synonyms), Gumbel, and
Weibull distributions are also available, but there is little overlap
between the quantiles of lengths within them; the logistic falls in the
middle.  The Weibull variant is more liberal, accepting lengths that are
two-thirds those needed to pass at the same level as the logistic.  The
Gumbel lengths are four-thirds longer.  The filter type, size, and draw
size are the same as for the peak height model.  Unlike the peak model,
different filters require different models internally.

The length distribution varies smoothly with the data size and filter, and
the flat model can calculate the probability directly without going through
a distribution function.

The models come from simulations over the ranges \code{n = 50 \dots 500} and
\code{flp = 0.05 \dots 0.5}, measuring quantiles between \code{q = 0.90 \dots
0.99999}.  They fit the critical values within 5\% over most of these values,
degrading to 10\% at the edges.  The spread in the reported probability also
increases at the edges of the parameter space.  In particular, data sets of
less than 60 points or windows larger than 30\% are less trustworthy, as are
quantiles beyond 0.9999.  The models will generate a warning under these
conditions and a tighter significance level should be used to judged the
results.  For data sizes much beyond 500, it is better to switch to the
normal or Weibull base distribution when testing flats.

Bad values passed for the draw and LP kernel sizes will raise errors.  The
filter name will default to Kaiser if the argument does not match a supported
kernel or if it is a bad value (NA, empty, or non-character).  The base
distribution similarly defaults to the logistic.  The arguments correspond
to options \opt{lp.kernel}, \opt{lp.window} or \opt{diw.window}, and
\opt{flat.distrib}.  The probabilities should be evaluated against
\opt{alpha.ht} and \opt{alpha.len} for the minimum passing level.

All four functions can take vectors as their first argument, which are
evaluated one by one for the given filter and draw set-up.
}

\value{
\code{Dipeak.test} and \code{Diflat.test} return lists of class \opt{Ditest}
with elements
\item{method}{a string describing the test}
\item{statfn}{function used to evaluate significance level/probability}
\item{statistic}{what is tested, the height of the peak or length of the flat}
\item{statname}{text string describing the statistic}
\item{parameter}{distributional arguments, for the peak the corrected height
  \code{corrht} and the \code{mu} and \code{lambda} for the inverse Gaussian;
  omitted for flats}
\item{p.value}{probability of feature}
\item{alternative}{a string describing the direction of the test vs. the
  null distribution}
\item{model}{parameters for the feature model, \code{n} the data size,
  \code{flp} the low-pass filter size as a fraction of \code{n},
  \code{filter} the low-pass kernel, and for flats \code{basedist} the
  distribution used to build the model
  \nl }

\code{statistic} and \code{p.value} will have the length of the \code{ht}
or \code{len} argument.  NA and NaN values in the first argument will
propagate to \code{p.value}, NULL produces an empty vector, and non-numeric
values an NA.  If \code{pval} is less than 0 or greater than 1 the
\code{p.value} is NaN.
}

\seealso{
 \code{\link{Dimodal}},
 \code{\link{Diopt}},
 \code{\link{find.peaks}},
 \code{\link{find.flats}}
}

\examples{
pval <- Dipeak.test(0.25*(1:16), 200, 0.15,'kaiser', lower.tail=FALSE)
pval$p.value
## Recovers pval.
Dipeak.critval(pval$p.value, 200, 0.15,'kaiser')

pval <- Diflat.test(10*(1:12), 200, 0.15,'kaiser', 'logistic', lower.tail=FALSE)
pval$p.value
Diflat.critval(pval$p.value, 200, 0.15,'kaiser', 'logistic')
}

\keyword{Dimodal}
\keyword{modality}
\keyword{low-pass spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}
\newcommand{\optcr}{\ifelse{html}{}{\ifelse{latex}{\sspace \cr}{\cr}}}

\name{Diopt}
\alias{Diopt}
\alias{Diopt.local}
\title{
Options for Dimodal package
}
\description{
Analysis, test, and display parameters for the Dimodal package.  Modeled on
the \code{par} options storage.
}
\usage{
Diopt(...)
Diopt.local(...)
}
\arguments{
\item{...}{
access or change or override options
}
}
\details{
Diopt provides a global database of options for the Dimodal analysis.  The
function can be called in five ways:
\describe{
\item{\code{Diopt()}}{
  returns the current options set as a list of tag=value pairs
}
\item{\code{Diopt(NULL)}}{
  resets all options to the package defaults, returning values before reset
}
\item{\code{Diopt(tag1=val1, tag2=val2, ...)}}{
  checks the option values and stores those that are valid, returning in a
  list the old values of the tags or NULL if a value is rejected
}
\item{\code{Diopt(list(tag1=val1, tag2=val2, ...))}}{
  checks the option values and stores those that are valid, returning in a
  list the old values of the tags or NULL if a value is rejected
}
\item{\code{Diopt("tag1", "tag2", ...)}}{
  returns the requested option(s) in a list
}
}
The function does sanity checking on the passed values for each option, and
is preferred to directly modifying the elements in the returned list.  Like
\code{par}, \code{Diopt(Diopt(tag1=val1, tag2=val2))} will restore the
original state.

Diopt.local returns all current values of the options overridden by the
tag=value pairs in the argument, if the values are valid.  Like Diopt, the
arguments can be wrapped in an unnamed list.  The options stored do not
change; this is for a local, one-off edit to the values.  Do not use
Diopt when setting an \code{opt} argument for other functions, which expect
the full list of values and not just those changed.

String options are partially matched.  Fraction values are between 0 and 1,
exclusive.
}
\value{
As described in Details.
}

\section{Parameters: Data Preparation}{
\describe{
\item{\code{analysis} <string>}{ \optcr
  which type of spacing to analyze, one or both of "lp" and "diw"
  (case-insensitive) for using a low-pass filter or interval spacing.
}
\item{\code{data.midq} <integer with valid midquantile type>}{ \optcr
  approximation method to convert quantiles or indices to original data,
  if 0 determines automatically
}
}
}

\section{Parameters: Low-Pass Filter Setup}{
\describe{
\item{\code{lp.kernel} <string>}{ \optcr
  filter type, one of "kaiser", "triangular" or "bartlett" (synonyms),
  "hanning", "hamming", "gaussian" or "normal" (synonyms), "blackman"
  (case-insensitive)
}
\item{\code{lp.window} <fraction between 0 and 1 or positive integer>}{ \optcr
  kernel size, either as fraction of the data or in points
}
\item{\code{lp.tests} <string>}{ \optcr
  tests to run on detected features, one or more of "ht", "pkexcur", "len",
  and "ftexcur" (case-insensitive)
}
\item{\code{lp.param} <list>}{ \optcr
  list of tag=value pairs of detector or test or acceptance parameters that
  will be overridden for the low-pass spacing checks
}
}
}

\section{Parameters: Interval Spacing Setup}{
\describe{
\item{\code{diw.window} <fraction or positive integer>}{ \optcr
  interval size, either as a fraction of the data or in points
}
\item{\code{diw.tests} <string>}{ \optcr
  tests to run on detected features, one or more of "pkexcur", "runht",
  "nrun", "runlen", "ftexcur" (case-insensitive)
}
\item{\code{diw.param} <list>}{ \optcr
  list of tag=value pairs of detector or test or acceptance parameters to
  override for the interval spacing checks
}
}
}

\section{Parameters: Local Extrema Detector}{
\describe{
\item{\code{peak.fht} <fraction>}{ \optcr
  minimum peak height as a fraction of the signal's range
}
\item{\code{peak.frelht} <fraction>}{ \optcr
  minimum peak height as a fraction of the average with the adjacent minima
}
\item{\code{peak.fhttie} <fraction>}{ \optcr
  maximum relative difference to treat values around extrema as the same
}
\item{\code{peak.fhsupp} <fraction>}{ \optcr
  fraction of peak height to determine the support range, used for excursion
  test (ex. 0.5 is the Full Width at Half Maximum, 1.0 extends to the minima)
}
}
}

\section{Parameters: Flat Detector}{
\describe{
\item{\code{flat.fripple} <fraction>}{ \optcr
  ripple specification as fraction of the signal's range
}
\item{\code{flat.minlen} <positive integer incl. 0>}{ \optcr
  shortest length of flats in data points
}
\item{\code{flat.fminlen} <fraction>}{ \optcr
  shortest length of flats as fraction of the data
}
\item{\code{flat.noutlier} <positive integer incl. 0>}{ \optcr
  number of outliers beyond the ripple allowed between source of flat and
  each endpoint
}
\item{\code{flat.distrib} <string>}{ \optcr
  null distribution model for length model test, one of "logistic",
  "weibull", "normal" or "gaussian" (synonyms), or
  "gumbel" (case-insensitive)
}
}
}

\section{Parameters: Excursion/Permutation Tests}{
\describe{
\item{\code{excur.nrep} <positive integer>}{ \optcr
  number of trial draws for peak and flat excursion tests
}
\item{\code{excur.ntop} <positive integer incl. 0>}{ \optcr
  number of points at start and end of data to ignore if they are the largest,
  to avoid adding the large tails of the spacing to the draw pool; can compare
  the first and last spacings against its standard deviation
}
\item{\code{excur.seed} <positive integer incl. 0>}{ \optcr
  RNG seed for each excursion test, 0 to not set
}
\item{\code{perm.nrep} <positive integer>}{ \optcr
  number of trial draws for run permutation test
}
\item{\code{perm.seed} <positive integer incl. 0>}{ \optcr
  RNG seed for each permutation test, 0 to not set
}
}
}

\section{Parameters: Test Significance (Acceptance) Levels}{
\describe{
\item{\code{alpha.ht} <fraction>}{ \optcr
  acceptance level of the peak height model test
}
\item{\code{alpha.pkexcur.lp} <fraction>}{ \optcr
  acceptance level of the peak height excursion (bootstrap) test for features
  found in the low-pass spacing
}
\item{\code{alpha.pkexcur.diw} <fraction>}{ \optcr
  acceptance level of the peak height excursion (bootstrap) test for features
  found in the interval spacing
}
\item{\code{alpha.len} <fraction>}{ \optcr
  acceptance level of the flat length model test
}
\item{\code{alpha.ftexcur.lp} <fraction>}{ \optcr
  acceptance level of the flat height excursion (bootstrap) test for features
  found in the low-pass spacing
}
\item{\code{alpha.ftexcur.diw} <fraction>}{ \optcr
  acceptance level of the flat height excursion (bootstrap) test for features
  found in the interval spacing
}
\item{\code{alpha.runht} <fraction>}{ \optcr
  acceptance level of the runs permutation test
}
\item{\code{alpha.nrun} <fraction>}{ \optcr
  acceptance level of the runs count statistics test
}
\item{\code{alpha.runlen} <fraction>}{ \optcr
  acceptance level of the longest run test
}
}
}

\section{Parameters: Other}{
\describe{
\item{\code{track.maxwindow} <fraction or positive integer>}{ \optcr
  maximum low-pass or interval window size for Ditrack variations
}
}
}

\section{Parameters: Display}{
\describe{
\item{\code{palette} <string>}{ \optcr
  the color palette to use when plotting the spacing or histogram and marking
  features, either a name from \code{palette.pals} or if starting with the
  string "hcl:" from \code{hcl.pals}, where the matching of the name follows
  the same rules as the palette functions
}
\item{\code{colID.data} <integer 1 t/m 8>}{ \optcr
  index in the palette (of the 8 colors generated) for drawing the data or its
  spacing
}
\item{\code{colID.filter} <integer 1 t/m 8>}{ \optcr
  index in the palette for drawing the LP or interval spacing
}
\item{\code{colID.hist} <integer 1 t/m 8>}{ \optcr
  index in the palette for drawing the bars in the data histogram
}
\item{\code{colID.cdf} <integer 1 t/m 8>}{ \optcr
  index in the palette for drawing the distribution atop the histogram,
  and the decile axes to align the spacing index with the data
}
\item{\code{colID.peak} <integer 1 t/m 8>}{ \optcr
  index in the palette when marking the peaks with vertical dashed or dotted
  lines
}
\item{\code{colID.flat} <integer 1 t/m 8>}{ \optcr
  index in the palette when marking the flats with boxes around the feature
}
\item{\code{digits} <positive integer incl. 0>}{ \optcr
  number of significant digits to use when printing raw values, if 0 then
  use \code{options("digits")}
}
\item{\code{mark.alpha} <logical>}{ \optcr
  TRUE to use ANSI underlining escape codes when printing probabilities to
	compare them against the alpha levels, FALSE to not add any formatting
}
\item{\code{mark.flat} <string>}{ \optcr
  how to indicate flats in a plot, one of "box" or "bar" (case-insensitive)
}
}
}

\seealso{
\code{\link{Dimodal}}, \code{\link{palette.pals}}, \code{\link{hcl.pals}}
}

\examples{
## Use the triangular kernel spanning 10\% of the data for the
## low-pass analysis.  Set the interval to 30 and override the
## ripple spec in the interval spacing but not for the low-pass.
Diopt(lp.kernel='bartlett', lp.window=0.10, diw.window=30, diw.param=list(flat.fripple=0.02))
## Diopt()$lp.kernel also accesses the new value.
Diopt('lp.kernel', 'diw.window')
## Temporarily override the kernel, without changing default.
Diopt.local(list(lp.kernel='Hanning'))$lp.kernel
Diopt()$lp.kernel
## Reset.
Diopt(NULL)
}

\keyword{Dimodal}
\keyword{modality}
\keyword{spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal}
\alias{Dimodal}
\alias{print.Dimodal}
\alias{summary.Dimodal}
\alias{plot.Dimodal}
\title{
Detect modality in the spacing of data.
}

\description{
\pkg{Dimodal} studies the modality of data using its spacing.  The presence
of peaks or local increases in it indicates the data is multi-modal and
locates the anti-modes.  Flats or consistent spacing cover the modes.
\pkg{Dimodal} finds these features after smoothing the spacing by low-pass
filtering, which supports discrete or heavily quantized data, or in the
interval spacing.  Several tests, using parametric models, runs, and
bootstrap sampling, evaluate these features.
}

\usage{
Dimodal(x, opt=Diopt())
\method{print}{Dimodal}(x, feature=c('peaks', 'flats'), \dots)
\method{summary}{Dimodal}(object, feature=c('peaks', 'flats'), \dots)
\method{plot}{Dimodal}(x, show=c('lp', 'histogram', 'diw'),
     feature=c('peaks', 'flats'), opt=Diopt(), \dots)
}

\arguments{
\item{x}{ for \code{Dimodal} the (numeric) data vector to analyze;
  for the methods an object of class \opt{Dimodal}
}
\item{object}{ an object of class \opt{Dimodal} }
\item{opt}{ local version of options to guide analysis }
\item{feature}{
  display only the indicated feature(s) in all methods that were run, or
  for plots mark only them in the graph
}
\item{show}{
  plot the low-pass spacing, a histogram of the raw data, and/or the
  interval spacing, in separate graphs in the order given
}
\item{\dots}{ extra arguments, ignored for all methods }
}

\details{
Changes in the spacing of data can indicate a change in its modality, and
\code{Dimodal} is a general interface to feature detectors and tests to
evaluate such changes.  Spacing, the difference between consecutive order
statistics or the delta after sorting the data, takes on a `U' form,
increasing rapidly in the tails and remaining stable in the center (for
single-sided variates it forms half the U; uniform variates have constant
spacing).  The transition between modes is marked by local increases in the
spacing while the center of modes see stable values.  Dimodal therefore
looks for local maxima or peaks in the spacing, or locally flat regions.

The spacing, designated \code{Di}, is often very noisy, and may be quantized
to a few values if the data is discrete or taken with limited precision.
Smoothing is necessary, which Dimodal can do either by apply a low-pass
(\code{lp}) filter or by taking the difference over more than one order
statistic.  The latter is called the interval spacing \code{Diw} and is
generated as a difference with lag; it is equivalent to a running mean or
rectangular filter of the raw spacing.  The recommended low-pass filter is
a Kaiser kernel, which offers good high-frequency suppression and main lobe
width; other available filters are the Bartlett or triangular (synonyms),
Hanning, Hamming, Gaussian or normal (synonyms), and Blackman.  Filtering
is done by convolving the data with the filter's kernel, rather than moving
to the Fourier domain.  Points at the start and finish that are partially
covered by the kernel or interval are set to NA and attributes attached to
the data give the valid range.  Indexing from the two spacings is different.
The low-pass kernel is centered, with partial overlaps at both ends.  The
interval spacing is defined as trailing from the upper index, which runs
to the end of the data, so the partial overlap occurs only at the start.
This will be seen in the position of the smoothed curves when plotting
results and the shift in indices needed to align the two schemes will be
printed with the data summary.  The raw values corresponding to a feature
automatically compensate for the difference.

The feature detectors \code{find.peaks} and \code{find.flats} have
separate help pages describing their algorithms and the parameters that
control their analysis.  These features are local and therefore not only
indicate whether data may be multi-modal, but provide the location of the
modes and the transitions between them.

Dimodal uses three main strategies to evaluate the features.  First, the
models tests are \code{Dipeak.test} and \code{Diflat.test}, with critical
values at a significance level also available.  These models are based on
simulations of the peak heights and flat lengths in a univariate null
distribution and offer a parametric assessment of their significance.  They
are less conservative than other modality detectors.  Second, the
bootstrap test is \code{Diexcurht.test}.  The bootstrap simulates the
features drawing from a pool of the difference of the spacing, estimating
their probability without assuming any underlying distribution.  Finally,
the runs tests are \code{Dinrun.test}, \code{Dirunlen.test}, and
\code{Dipermht.test}.  Quantizing the filtered spacing into a few levels
by taking the sign of the difference (in other words, if the signal is
increasing, decreasing, or constant) allows us to consider runs in the
symbols.  We can test how many there are, or the longest, or if a
permutation of them recreates the feature.

A fourth strategy, using changepoint detectors on the raw spacing to
detect transitions between modes and anti-modes, is not included in
this version of Dimodal.  See the package help page or DESCRIPTION file
for the location of the full version.

The bootstrap test extends a peak to its support, defined by the
\opt{peak.fhsupp} option, a fraction of the peak's height.  A value of 0.9
is enough to back the away from minima placed in a long flat while not
distorting the peak's width if the minima are well-defined.  0.5
corresponds to Full Width at Half Maximum (FWHM), and 1.0 extends the
peak to the minima.

The analysis of each feature is gathered into separate S3 class objects
which support printing and marking plots.
The generic functions on the Dimodal
result route to these objects if they are selected by the \code{features}
argument.  A plot may contain the filtered spacing or interval spacing plus
a histogram of the raw data, with features annotated on each.  It uses layout
to create a row of the shown graphs, as specified by the \code{show}
argument.  The histogram annotations will come from the first, leftmost,
spacing shown.

The raw data must be numeric or integer.  Non-finite values, including NA,
will be dropped.

Dimodal needs a complete list of options for the \code{opt} argument.  Do
not make changes in the call, as \code{Diopt} will return only the changed
values.  Use \code{Diopt.local} instead.

The option \opt{analysis} controls which smoothed spacing to generate,
one or both of 'lp' and 'diw'.  If none of these are specified
the data will contain only the spacing and mid-quantile function,
without any features or their analysis.

Dimodal uses options \opt{lp.param} and \opt{diw.param} to override
the detector options for each method, and \opt{lp.tests} and
\opt{diw.tests} to determine which feature tests to carry out.  If
these are empty lists then the data will contain the smoothed spacing but
there will be no features.  While generating the data it uses options
\opt{lp.kernel} and \opt{lp.window} to set up the low-pass filter,
and \opt{diw.window} for the interval width.  It uses \opt{excur.ntop}
when creating the base set of draws for excursion tests.  Option
\opt{data.midq} determines the approximation method (\code{type} argument to
the \code{midquantile} function), when converting indices in the spacing back
to order statistics.

The default values of the detector options come from the development of the
low-pass models.  We do not know how different values will affect the models.
The interval spacing is much rougher than low-pass filtering, which may
require looser ripple and height parameters to find any flat, or reduce the
number of peaks.  The excursion tests will accommodate this.
}

\value{
A list assigned to class \opt{Dimodal} with elements
\item{data}{
  an object of class \opt{Didata} with all data used in the analysis
}
\item{lp.peaks}{
  an object of class \opt{Dipeak} capturing the local extrema in the
  low-pass spacing and their evaluation, with test results and raw data
  locations added to the features from \code{find.peaks}
}
\item{lp.flats}{
  an object of class \opt{Diflat} capturing the local flats in the
  low-pass spacing and their evaluation, with test results and raw data
  locations added to the features from \code{find.flats}
}
\item{diw.peaks}{
  an object of class \opt{Dipeak} containing the local extrema in the
  interval spacing and their evaluation, with test results and raw data
  locations added to the features from \code{find.peaks}
}
\item{diw.flats}{
  an object of class \opt{Diflat} capturing the local flats in the
  interval spacing and their evaluation, with test results and raw data
  locations added to the features from \code{find.flats}
}
\item{opt}{
  the list passed as the opt argument, per \code{\link{Diopt}}
  \nl }
These elements will have empty data structures if the analysis is not run.

Dimodal will automatically call \code{shiftID.place} on each detector's
results and will summarize the tests, as described with each data class.
Dimodal adds an attribute \opt{source} to each of the features, with
value LP, Diw, or Di.
}

\seealso{
 \code{\link{Diopt}} for the parameters controlling the analysis.
 
 \code{\link{find.peaks}, \link{find.flats}} for feature detection.
 
 \code{\link{Dipeak.test}, \link{Diflat.test}} for parametric models to
 evaluate the features, \code{\link{Diexcurht.test}} for a bootstrap test of
 feature significance, \code{\link{Dinrun.test}, \link{Dirunlen.test}} for
 tests of runs (here for sequences in the sign of the difference in the
 interval spacing), and \code{\link{Dipermht.test}} for a permutation test
 of the runs making a feature.
 
 \code{\link{Didata}, \link{Dipeak}, \link{Diflat}} for the
 data structures generated by the feature detectors and their evaluation.
 
 \code{\link{center.diw}} to further shift the position of interval spacing
 features to the middle of the interval to align with low-pass features.
 
 \code{\link{match.features}} to identify common peaks and flats in both
 spacings.

 \code{\link{shiftID.place}} to move indices in either spacing to the
 original data grid and add the corresponding raw values.

 \code{\link{midquantile}} for the mid-quantile mapping from index to raw
 data.
}

\examples{
## The interval spacing is noisy with the default options, so require a
## larger peak height with a temporary value to Diopt.
oldopt <- Diopt(diw.param=list(peak.fht=0.125))
## Run the analysis.
m <- Dimodal(faithful$waiting)
## If printing the results, the interval spacing peaks have a probability
## just under 0.05 but fail the acceptance levels.
summary(m)
## Details about the peaks in both spacings.
print(m, feature="peaks")
## We find one peak in both spacings, but only the low-pass is significant.
match.features(m)
## Three plots side by side.  The limited resolution of the data is clear
## in the interval spacing.
dev.new(width=12, height=4) ; plot(m)
## Restore the old option values.  Diopt(NULL) returns to defaults.
oldopt <- Diopt(oldopt)
}

\keyword{Dimodal}
\keyword{modality}
\keyword{spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal-package}
\alias{Dimodal-package}
\alias{Dimodal-package}
\docType{package}

\title{ Detection of Multi-Modal Data Using Spacing }

\description{
The \pkg{Dimodal} package uses the spacing of data, or difference between
order statistics, to detect and locate modes or the transition between them.
Consistent spacing with stable values appears within a mode, while it
increases at the anti-modes.  The package contains parametric and
non-parametric tests of features detected within the spacing.
}

\section{Main Interface}{
The package has three top-level commands.
\itemize{
\item \code{\link{Dimodal}} runs the modality analysis.  It supports print,
  summary, and plot methods.
\item \code{\link{Diopt}} provides a persistent database of options
  controlling the feature detectors, tests, and display of results.
\item \code{\link{Ditrack}} displays the position and probability of features
  as filter sizes change, to help in selecting the best size for analysis.
  It has a plot method to graphically show the results.
}
}

\section{Feature Detectors}{
The package has five feature detectors.  They work with any data, not just
the spacing.
\itemize{
\item \code{\link{find.runs}} identifies fuzzy runs, sequences of nearly-equal
  numeric values or of equal discrete values or symbols.
\item \code{\link{find.peaks}} identifies local extrema, merging small minor
  peaks into larger.
\item \code{\link{find.flats}} identifies flat or consistent stretches of
  values.
\item \code{\link{find.level.sections}} is an inversion of the \pkg{modehunt}
  changepoint detector.
}
}

\section{Feature Tests}{
\code{Dimodal} includes three groups of tests to evaluate features.
\itemize{
\item \code{\link{Dipeak.test}} and \code{\link{Diflat.test}} are parametric
  models of the peak and flat distributions after low-pass filtering.
  \code{\link{Dipeak.critval}} and \code{\link{Diflat.critval}} provide
  critical values of the peak (height) and flat (length) for a significance
  level.
\item \code{\link{Dinrun.test}} and \code{\link{Dirunlen.test}} are runs-based
  tests (up, down, equal trends) performed on the signed difference of a
  signal.  They include the Kaplansky-Riordan test of the number of runs and
  a Markov chain model for the longest run.
\item \code{\link{Dipermht.test}} and \code{\link{Diexcurht.test}} are
  bootstrap tests simulating a feature from the actual data.  They include a
  permutation test of runs and a general excursion test from the difference
  of a signal.
}
}

\section{Data}{
The \code{\link{kirkwood}} dataset has the multi-modal distribution of
asteroid orbital radii, where modes identify families of asteroids within
the main belt and anti-modes the Kirkwood gaps cleared out by regular
perturbation from Jupiter.
}

\section{Classes}{
The return value of each command, feature detector, and test is given an
S3 class that supports printing, summarizing, and perhaps plotting.  Links
can be found in the return value section of the command.
}

\section{Utility Functions}{
The package includes several functions to help work with the results of the
analysis.
\itemize{
\item \code{\link{midquantile}} uses piecewise linear segments to convert
  quantiles of discrete or heavily quantized data back to data.
\item \code{\link{runs.as.rle}} converts the \code{\link{find.runs}} result
  to the \code{"rle"} class.
\item \code{\link{select.peaks}} returns just the local maxima from the
  extrema detected by \code{\link{find.peaks}}.
\item \code{\link{center.diw}} shifts the indices of features in the interval
  spacing, normally located at the end of the interval, into the center, to
  align with the low-pass features and actual data.
\item \code{\link{match.features}} uses distance and overlap criteria to
  identify features found in both the low-pass and interval spacing.
\item \code{\link{shiftID.place}} moves the results of
  \code{\link{find.peaks}} and \code{\link{find.flats}}
  to the original data grid and converts any indices
  into raw values using the midquantile approximation.
}
}

\section{Version}{
This version of the package does not include the changepoint analysis of
the spacing.  The full version is available at
\url{https://www.primachvis.com/data/Dimodal_latest.tar.gz}.
}

\author{
Greg Kreider.

The package compiles by default with the PCG random number generator, written
by Melissa O'Neill, for sampling during the excursion tests.

The \code{kirkwood} dataset is taken from the Lowell Observatory asteroid
ephemeris.
}

\keyword{Dimodal}
\keyword{modality}
\keyword{spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Dimodal Utility Functions}
\alias{midquantile}
\alias{runs.as.rle}
\alias{select.peaks}
\alias{center.diw}
\alias{match.features}
\alias{shiftID.place}
\title{
Dimodal Utility Functions
}
\description{
Miscellaneous functions for working with Dimodal results.
}

\usage{
midquantile(x, q=((1:length(x))-1)/(length(x)-1), type=0L, feps=0.0)
runs.as.rle(runs, x)
select.peaks(pk)
center.diw(m)
match.features(m, near=10, foverlap=0.70, nomatch=NA_integer_, quiet=FALSE)
shiftID.place(feat, offset, xmid)
}

\arguments{
\item{m}{
  a \opt{Dimodal} object returned from \code{Dimodal}
}
\item{x}{
  the original data, with the same length as the members of \code{runs}
}
\item{runs}{
  the list returned from \code{find.runs}
}
\item{pk}{
  a \opt{Dipeak} object
}
\item{near}{
  maximum distance in points between matching peaks, or as a fraction of the
  length of the original data
}
\item{foverlap}{
  minimum fraction of the length of either flat that the common segment must
  cover
}
\item{nomatch}{
  value to use when a feature has no match in the other spacing, treated as
  integer internally
}
\item{quiet}{
  a boolean, TRUE to only determine the matching, FALSE to also print the
  aligned features
}
\item{q}{
  quantile(s) for mid-quantile approximation, by default at the data indices
}
\item{type}{
  algorithm determining segments approximating x, an integer from 0 to 4 as
  described in Details
}
\item{feps}{
  tolerance for matching values, per \code{find.runs}
}
\item{feat}{
  a \opt{Dipeak} or \opt{Diflat} data frame
}
\item{offset}{
  an integer, the amount to shift position of peaks or points or endpoints
  of flats
}
\item{xmid}{
  a vector of interpolated quantiles to convert indices back to raw data,
  as stored in the \opt{Didata} 
}
}

\details{
The \code{midquantile} function approximates the quantile function by
replacing the steps of the \code{ecdf} distribution with piecewise linear
segments; see Ma, Genton, and Parzen (2011).  This creates a ramp over tied
or discrete values, giving a better estimate of the position of features,
especially when there are large gaps between modes and few or no data points
within them.  The function determines the segment endpoints and by default
evaluates them on the original data grid, scaling the vector indices to run
from 0 through 1.  It first converts the data to runs using \code{find.runs},
with \code{feps} defining ties.  Segmentation type 1 is the mid-distribution
function of Ma, with the data value at the ends of runs shifted to the middle
of the change.  Segmentation type 2 instead shifts the quantiles by half an
index, extending the step in the \code{ecdf}.  These two approaches can
create an envelope around the quantile function, with the type 1 offset from
the data at q = 0 and the type 2 at q = 1.  Segmentation type 3 combines both
shifts, interpolating on a half grid for both x and q.  It follows the
quantile function better, but does round off the curve at single data points.
In practice types 1 and 3 are close. Type 4 runs segments between the middle
of runs, or through the data points when there are none.  This reduces its
estimation error, but the strategy does assume that the step in data to
either side of the run is about the same.  If not, the other approximations
would move away from the center of the run.

Type 4 is best when the data has very few ties.  Use types 3 or 1 when there
are.  Type 0 will automatically select the strategy, using 3 when there are
ties and 4 when not.  It uses a simple check, whether the number of unique
values is a tenth of the data, to decide if there are enough ties.

Internally the function makes two calls C-side.
\code{.Call("C_midq", x, type, feps, PACKAGE="Dimodal")} returns a vector
with the piecewise linear segments, with \code{$x} the endpoints along the
data and \code{$q} along the quantiles.
\code{.Call("C_eval_midq", pts$x, pts$q, q, PACKAGE="Dimodal")} uses these
segments as the first two arguments and new quantiles as the third to
interpolate data values.

The \code{find.runs} returned value has two vectors with the length of the
data.  One has non-zero values at the start of runs, the other counts skipped
invalid points.  The \opt{rle} class is more compact, storing only the runs
and the data values at the start.  The \code{runs.as.rle} function does this
compaction.

\code{find.peaks} returns a data frame with not only maxima but also the
minima between them.  It includes maxima even if they are at the first or last
point, with minima to only one side.  \code{select.peaks} selects only
those peaks surrounded by minima.  It may return a \opt{Dipeak} object with
no rows.  \code{pk} need not include the modifications from \code{Dimodal};
select.peaks keeps all columns of its argument.

Indexing in interval spacing is at the end of the interval but the low-pass
filter is centered.  \code{center.diw} shifts the interval spacing features
to align with the data, including peak positions, flat source identifiers,
and flat start and end points.  Note that the raw value is already shifted
when set by Dimodal and will not change.

\code{match.features} aligns peaks and flats between the low-pass and
interval spacing.  It compares only valid maxima, as per \code{find.peaks},
and shifts interval spacing positions with \code{center.diw} before matching
them.  Peaks must lie within \code{near} points to match.  Flats must overlap,
and the common segment must be at least \code{foverlap} of the length of
either flat.  The function prints the position, raw value, and the number of
tests that have passed their acceptance level, unless \code{quiet} is TRUE.
The \code{nomatch} value is cast internally to an integer and cannot be
between 1 and the number of features in either spacing, to prevent conflicts.
NA, 0, and negative values are acceptable.

The \code{shiftID.place} function is used in Dimodal to modify the
placement of features, and is provided separately if the detectors
\code{find.peaks} or \code{find.flats} are called
directly.  It adds \code{offset} to the columns \opt{pos}, \opt{stID},
\opt{endID}, \opt{lminID}, \opt{rminID}, \opt{lsuppID}, \opt{rsuppID},
and \opt{src} if they exist in the features data frame to account for
values skipped during filtering.  Use the \opt{lp.stID} attribute for
low-pass features, \opt{diw.stID} for interval spacing, and 2 for
changepoints.  If pos, stID, or endID are in the data frame the function
also adds columns \opt{x}, \opt{xst}, and \opt{xend} respectively with
the original data value for the index by using the midquantile result
\code{xmid}.
}

\value{
\code{midquantile} returns a vector the same length as the data.  Quantiles
outside the range [0,1] return the first or last data point, even if this is
discontinuous with the values at 0 or 1.  In other words, the function does
not follow the piecewise linear segment outside the valid range, but clips
it.  NA or NaN quantiles propagate.

\code{runs.as.rle} returns a list of class \opt{rle} with members
\opt{lengths} and \opt{values}, as per the \code{rle} command.  It also
adds a member \opt{nskip} with the number of non-finite values in the data
within the run.

\code{select.peaks} returns a subset of the argument, possibly with zero
rows.  If the argument is not a \opt{Dipeak} object, it returns a dummy
empty object.

\code{center.diw} returns its argument with modified \code{diw.peaks} and
\code{diw.flats}, if they exist.

\code{match.features} returns a list with four elements.  \opt{peak.lp2diw}
is a vector with one element per row in lp.peaks whose value is the matching
row number in diw.peaks, or \code{nomatch} if there is no match or the
lp.peaks row is not a valid peak.  \opt{peak.diw2lp} is a similar map
from \code{diw.peaks} to \code{lp.peaks}.  \opt{flat.lp2diw} and
\opt{flat.diw2lp} are the equivalent maps for flats.

\code{shiftID.place} returns the modified feat data frame.
}

\references{
Y. Ma, M. Genton, E. Parzen (2011),
Asymptotic properties of sample quantiles of discrete distributions.
\emph{Ann Inst Stat Math} 63, pp. 227--243.
}

\seealso{
 \code{\link{Dimodal}},
 \code{\link{Didata}},
 \code{\link{find.runs}},
 \code{\link{rle}},
 \code{\link{find.peaks}},
 \code{\link{Dipeak}},
 \code{\link{find.flats}},
 \code{\link{Diflat}},
 \code{\link{ecdf}}
}

\examples{

m <- Dimodal(faithful$eruptions, Diopt.local(analysis=c('lp','diw')))
# How many peaks were found?  Use print.data.frame to see the full structure.
nrow(select.peaks(m$lp.peaks))
nrow(select.peaks(m$diw.peaks))
# Compare to m$diw.peaks.
m$diw.peaks
center.diw(m)$diw.peaks
# Flats do not match because the Diw feature only covers 50% of the LP.
match.features(m)

plot(sort(iris$Petal.Length))
lines(midquantile(iris$Petal.Length, type=1L), col='red')
lines(midquantile(iris$Petal.Length, type=2L), col='blue')
lines(midquantile(iris$Petal.Length, type=3L), col='green')
lines(midquantile(iris$Petal.Length, type=4L), col='orange')

# See the Dimodal.R source code for the use of shiftID.place.

# To simplify the runs in the signed difference of the interval spacing
# runs.as.rle(Dimodal:::find.runs(m$data['signed',], 0.01), m$data['signed',]) 
}

\keyword{Dimodal}
\keyword{runs}
\keyword{peaks}
\keyword{flats}
\keyword{interval spacing}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Ditrack}
\alias{Ditrack}
\alias{print.Ditrack}
\alias{summary.Ditrack}
\alias{plot.Ditrack}
\title{
Class methods for Ditrack Objects
}
\description{
Find peaks and flats that appear in the spacing as the size of the
low-pass filter or interval changes.  Support methods for the class
include printing and plotting the significant features.
}

\usage{
Ditrack(x, smooth=c('lp', 'diw'), opt=Diopt())
\method{print}{Ditrack}(x, \dots)
\method{summary}{Ditrack}(object, \dots)
\method{plot}{Ditrack}(x, feature=c('peaks', 'flats'), opt=Diopt(), \dots)
}

\arguments{
 \item{x}{ for \code{Ditrack}, the data to study; for the class methods,
   the value returned from \code{Ditrack} }
 \item{object}{ an object of class \code{"Ditrack"} }
 \item{smooth}{ which filtering to apply to the data, "lp" to look for
   features in the low-pass spacing or "diw" in the interval spacing }
 \item{opt}{ local version of options passed to \code{Dimodal} analysis }
 \item{feature}{ which feature(s) to plot, may include both }
 \item{\dots}{ extra arguments, ignored for all methods }
}

\details{
The mode tree and SiZer visualization tools track a feature's position and
some metric as the bandwidth of a kernel density estimate changes.  They
help find the best bandwidth for the trade-off between feature detectability
and smoothing.  The \code{Ditrack} class does the same for the spacing
analysis.  It varies the size of the low-pass filter kernel or interval
spacing and tracks the position and probability of peaks and flats per the
tests in the \code{Dimodal} analysis.

The \code{Ditrack} function varies the window size and gathers the
peak and flat results into matrices.

The tracker will overwrite the options \opt{analysis} and either
\opt{lp.window} or \opt{diw.window} from \code{Diopt}.  These should not
appear in the \opt{lp.param} or \opt{diw.param} overrides, which have
precedence.  It steps these sizes from 0.01 to the option
\opt{track.maxwindow}.

\code{plot.Ditrack} graphs one or both features, with the window size on the
vertical axis and the position along the horizontal.  Peaks and minima are
drawn as symbols, with dots for the maxima and dashes for the minima.  The
dots are filled in if any test passes its acceptance level.  Flats are drawn
with a line spanning the feature.  Both are color-coded according to the
overall probability, the best of all tests that were run.  The coloring is
fixed and does not use the \opt{palette} option.  There will always be a
minimum between two peaks.  Extrema at the edges of the data are not drawn.

\code{print.Ditrack} gives a table with the number of significant features
at the 0.01 and 0.05 level plus the number of features significant according
to the current acceptance levels, as the window size changes.

The summary method prints the ranges of window sizes for peaks and flats
that give features that pass the acceptance levels of any test that has
been run.
}

\value{
The tracker returns a list assigned to class \opt{Ditrack} with elements
 \item{\code{peaks}}{ a subset of the \opt{Dipeak} result with columns
   \opt{pos}, \opt{ismax}, \opt{ppeak}, and \opt{naccept}, plus an additional
   column \opt{winpct} giving the filter size as an integer percentage,
   keeping only rows with valid peaks }
 \item{\code{flats}}{ a subset of the \opt{Diflat} result taking columns
   \opt{stID}, \opt{endID}, \opt{pflat}, and \opt{naccept}, plus an
   additional column \opt{winpct} giving the filter size }
 \item{\code{nx}}{ the length of the data \code{x} argument }
 \item{\code{smooth}}{ the argument of the same name }
The \code{peaks} and \code{flats} matrices will have one row per feature,
be it a minimum or maximum from \code{Dipeak} or a flat.  There may be
many rows with the same \code{winpct}, or none for a given value.
}

\seealso{
  \code{\link{Dimodal}},
  \code{\link{Dipeak}},
  \code{\link{Diflat}}
}

\examples{
  ## Not run because of the run time.
  \donttest{ trk <- Ditrack(quakes[,3], 'lp') }
  ## Two plots side by side.
  \donttest{ dev.new(width=8,height=4) ; plot(trk) }
  \donttest{ trk }
}

\keyword{Dimodal}
\keyword{spacing}

\name{kirkwood}
\alias{kirkwood}
\docType{data}
\title{Asteroid Orbital Axis Showing Kirkwood Gaps}
\description{
The distance of asteroids from the sun is multi-modal, with modes
corresponding to asteroid families and anti-modes to the Kirkwood gaps.
}

\usage{
data(kirkwood)
}

\format{
\code{kirkwood} is a numeric vector of 2093 values representing the
semi-major axis of an asteroid in Astronomical Units (A.U.).
}

\details{
The axis is stored in column 18 of the original data file.  We keep only
those asteroids within the orbit of Jupiter, with an axis below 5 A.U.  We
further keep only those asteroids whose diameter is known or not blank, in
column 7 of the data file.
}

\source{
Lowell Observatory maintains the asteroid ephemeris website at
\url{https://asteroid.lowell.edu/astorb/}
The \code{kirkwood} dataset is a subset of the "astorb.txt.gz" static copy
of the ephemeris, downloaded on 1 May 25.  The full file contains more than
1.4 million entries, and is growing daily.  Lowell Observatory provides a
query builder for downloading up-to-date values at
\url{https://asteroid.lowell.edu/query-builder}

The ephemeris is supported by grants from NASA and the Lowell Observatory;
details are found on the website.  Further distribution of the data requires
acknowledging these funding sources.
}

\references{
N. Moskovitz, L. Wasserman, B. Burt, R. Schottland, E. Bowell, M. Bailen,
M. Granvik,
The \verb{astorb} database at Lowell Observatory,
\emph{Astronomy and Computing}, 41, Oct 2022, pp. 100661.
}

\examples{

## Load data.
data(kirkwood, package="Dimodal")

## Start from scratch.
## Normally you would not need to save the results of the call, but
## this clutters up the screen.
opt <- Diopt(NULL)

## Set up the analysis.
## This is a large data set so spacing in features is small but
## smooth, which allows tighter detector parameters.  Values depend
## on the range of the spacing and are found by trial and error.
## Set RNG seeds for repeatability.
## Using bars to mark flats is easier to read with dense data.
opt1 <- Diopt(peak.fht=0.015, flat.fripple=0.0075, analysis=c("lp", "diw"),
              excur.seed=3, perm.seed=5, mark.flat="bar")

## Use Ditrack to see where passing features (filled in dots) appear.
## Interval spacing is the same cut-off.  Wrapped in \donttest because of
## the run time.
\donttest{trk <- Ditrack(kirkwood, "lp")}
\donttest{dev.new(width=8,height=4) ; plot(trk)}
opt1 <- c(opt1, Diopt(lp.window=0.05, diw.window=0.05))

## Run analysis.
m <- Dimodal(kirkwood)

## Summarize the data and spacing.
m$data

## Print features that exist in both low-pass and interval spacing.
## Allow large distance between peaks because  of data set size.
mtch <- match.features(m, near=30)

## Gap distance (in AU), dropping the extra LP peak.
## There are gaps at 2.30, 2.48, 2.86, 2.93, and 3.06 AU.
a_gap <- (select.peaks(m$lp.peaks)$x[-3L] + select.peaks(m$diw.peaks)$x) / 2
a_gap

## Orbital resonance with Jupiter, which has orbital axis of
## 5.201 AU, from Kepler's third law.  The corresponding
## resonance is 3.40 (ratio 7:2 or 10:3), 3.04 (3:1),  2.46 (5:2),
## 2.36 (7:3), and 2.22 (9:4).  Simulations confirm all but the
## first and last.
resonance <- (5.201 / a_gap) ^ (3/2)
resonance

## The flats include families of asteroids, although other
## considerations must be taken into account to find true
## family members (composition of asteroid, orbital
## eccentricity and inclination).
## The second range at 2.76 AU includes the Ceres family,
## the third at 3.13 AU the Themis.  The first range at
## 2.64 AU includes the Eunomia and Prosperina families.
m$diw.flats

## 3 graphs side-by-side with results.
dev.new(width=12, height=4) ; plot(m)

## The full test results of the low-pass peaks.
m$lp.peaks

## In the Ditrack plot there are two peaks in the very lower right
## corner.  We need smaller filter kernels/intervals to find these.
## The parameters already set will still apply.
opt3 <- Diopt(lp.window=0.015, diw.window=0.015, peak.fht=0.025, peak.frelht=0.10)
## Running Dimodal with such small windows will generate warnings
## about the model tests being pushed out of bounds.
## This adds gaps at 1.93 (2:1) and 1.74 (7:4), both expected
## from simulations.  The 2:1 gap position shifts because the
## spacing is so large that there is data point to anchor the
## value and a second, smaller increase pulls the peak to 1.93.
m2 <- Dimodal(kirkwood) ; mtch <- match.features(m2, near=30)
dev.new(width=12, height=4) ; plot(m2)

## Restore default.
opt <- Diopt(opt)
}

\keyword{datasets}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Didata}
\alias{Didata}
\alias{print.Didata}
\alias{summary.Didata}
\title{
Class methods for Didata Objects
}
\description{
Common functions to handle \opt{Didata} class results, which store all
data used in the modality analysis.
}

\usage{
\method{print}{Didata}(x, \dots)
\method{summary}{Didata}(object, \dots)
}

\arguments{
\item{x}{ an object of class \opt{Didata} }
\item{object}{ an object of class \opt{Didata} }
\item{\dots}{ extra arguments, ignored for all methods }
}

\details{
Didata is a matrix that contains all data used in the modality analysis,
with contents that depends on which analyses are run.  The matrix has one
column for each raw data point.  Dimodal creates the Didata object
internally; there is no public function to generate it.  Row \opt{xmid}
is the interpolated raw value for each index, as calculated by
\code{midquantile(data["xsort",], type=#)}, where the algorithm type is
provided by the option \opt{data.midq}.  Row \opt{signed} added by
the interval spacing analysis is its signed difference, where differences
within an order of magnitude of \code{.Machine.double.eps} are treated as
ties to avoid numerical precision errors.

The print method shows the source of the data and any filters applied to
the spacing.  It then prints a table with the valid columns, range of
values, and standard deviation of the original data and spacing.  If
low-pass and interval spacing were needed by \code{Dimodal}, then information
about them is added to the table.  The method uses option \opt{digits} to
control the number of significant digits for raw values.  If it is set to
zero then the standard \code{options()$digits} is used.

The summary method prints just the setup information.
}

\value{
The basic data stored in the matrix contains the rows
 \item{\code{x}}{ the original data }
 \item{\code{xsort}}{ sorted data }
 \item{\code{xmid}}{ the mid-distribution from \code{midquantile} }
 \item{\code{Di}}{ the spacing; index 1 is NA
   \nl }

If low-pass spacing is required, \code{Dimodal} adds the row
 \item{\code{lp}}{ spacing after low-pass filtering
   \nl }
where the leading and trailing columns are set to NA when part of the
filter falls outside the data.

If interval spacing is needed, the matrix also contains rows
 \item{\code{Diw}}{ the interval spacing }
 \item{\code{signed}}{ the signed difference of the interval spacing,
   with values -1, 0, +1 $ \nl }
where the leading columns up to the width of the interval are set NA.  The
signed difference has an additional NA.

Several bits of metadata are stored as attributes with the matrix, in
addition to any generated by R.  The basic analysis provides
 \item{\code{data.name}}{ the source of the data in the call to \code{Dimodal}
   \nl }
Low-pass filtering adds
 \item{\code{lp.kernel}}{ the name of the filter }
 \item{\code{lp.window}}{ the window or size of the filter, as provided
   by \code{Diopt}, which is normally a fraction of the data size but
   may also be in data points }
 \item{\code{wlp}}{ the actual window width in data points }
 \item{\code{lp.stID}}{ the first column in the matrix with valid
   low-pass values }
 \item{\code{lp.endID}}{ the last column (inclusive) that is valid
   \nl }
Interval spacing adds
 \item{\code{diw.window}}{ the interval size, as provided by \code{Diopt},
   which is normally a fraction of the data size but may also be in data
   points }
 \item{\code{wdiw}}{ the interval in data points }
 \item{\code{diw.stID}}{ the first column in the matrix with valid interval
   spacing; the last valid column is the end of the matrix
   \nl }

Note that the \code{signed} row starts one column after the interval spacing,
at \code{diw.stID + 1}, and runs to the end of the matrix.

The methods return the object, invisibly.
}

\seealso{
  \code{\link{Dimodal}},
  \code{\link{Diopt}},
  \code{\link{midquantile}}
}

\examples{
m <- Dimodal(faithful$eruptions, Diopt.local(diw.window=16))
m$data
}

\keyword{Didata}
\keyword{Dimodal}

\newcommand{\opt}{\ifelse{latex}{\code{"#1"}}{\verb{"#1"}}}
\newcommand{\nl}{\ifelse{latex}{ }{\ifelse{html}{ }{ \cr}}}

\name{Diflat}
\alias{Diflat}
\alias{print.Diflat}
\alias{summary.Diflat}
\alias{mark.Diflat}
\title{
Class methods for Diflat Objects
}
\description{
Common functions to handle \code{"Diflat"} class results, which store
information about local flats and their probability.
}

\usage{
\method{print}{Diflat}(x, \dots)
\method{summary}{Diflat}(object, \dots)
\method{mark}{Diflat}(x, hist=NULL, opt=Diopt(), \dots)
}

\arguments{
\item{x}{ an object of class \code{"Diflat"} }
\item{object}{ an object of class \code{"Diflat"} }
\item{hist}{ the histogram being marked or NULL for a low-pass or interval
  spacing graph }
\item{opt}{ local version of options for color scheme }
\item{\dots}{ extra arguments, ignored for all methods }
}

\details{
A \opt{Diflat} object is a data frame that describes local flats,
including their position or span in the low-pass or interval spacing and
the equivalent raw values, their length, and any tests that have run to
judge the significance of the feature.  The basic information comes from
the \code{find.flats} detector.  The analysis in \code{Dimodal} moves the
positions to the original data and adds the test results.  The class
inherits from \opt{data.frame}.

The print method shows the endpoints (incl.) of the flat and the raw value
at each end, the statistics (test values) and separately the probability from
each.  The acceptance level of the tests are also given underneath.  These
come from the options \opt{alpha.len}, \opt{alpha.ftexcur.lp}, and
\opt{alpha.ftexcur.diw}.  If the option \opt{mark.alpha} is TRUE then
probabilities at or below the acceptance level are underlined.  Set the
option FALSE if your terminal does not support ANSI escape codes, or see the
\pkg{crayon} package for the extensive tests needed to determine this
automatically.  These options are set at the global level and cannot be
overridden for a single call to print; Diopt is called internally.  The pass
column after the overall flat probability will contain the number of
accepted/passing tests, if any.  Raw values and statistics are printed with
option \opt{digits} significant digits.  If the option is set to zero then
the default \code{options()$digits} is used.

The summary method gives the endpoints, the overall probability or best result
from all tests, and a list of the tests that pass per the \code{Diopt}
acceptance levels.

Flats are marked in a spacing graph according to the \opt{mark.flat} option,
either as a box around the spacing or as a bar above or below it.  The border
or line is thicker if the flat passes any test at the option acceptance
levels.  When annotating a histogram the flats are always drawn as bars at
the top of the graph.  The marks are colored per the option \opt{colID.flat}
index of the current palette.  Marking the histogram assumes that Dimodal
has added the \code{xst} and \code{xend} columns.
}

\value{
The basic data frame returned from \code{find.flats} contains the columns
 \item{\code{src}}{ the data point that triggered the flat, whose value is
   nearly centered in the height.  This is a unique identifier for the
   feature. }
 \item{\code{stID}}{ first data point in the flat }
 \item{\code{endID}}{ last data point (incl.) in the flat }
 \item{\code{len}}{ the length of the flat, or endID - stID + 1 }
 \item{\code{srcval}}{ the data value at src }
 \item{\code{ht}}{ the height of the flat }
 \item{\code{htsd}}{ the flat height standardized by the data's standard
   deviation, or ht / sd(x)
   \nl }
There is one row per flat, and the result is an empty data frame with zero
rows if there are no flats.  The indices of stID and endID are located in
the data passed to the detector, which excludes the filter's NA columns.

\code{Dimodal} adds several columns.  It shifts all indices to the original
data, adjusting for the filter or interval size.  Mapping indices back to
the raw data, it adds
 \item{\code{xst}}{ the location of the start point in the original data }
 \item{\code{xend}}{ the location of the end point (incl.) in the
   original data
   \nl }
This can be done by calling the utility function \code{shiftID.place} on the
\code{find.flats} result.  It also adds an attribute \opt{source} with value
"LP" or "Diw" depending on the filter used.  As a summary of the results it
adds \item{\code{pflat}}{ the best probability of all tests run, NA if no
test results are available }
 \item{\code{naccept}}{ number of tests passing their acceptance level,
   0 if none or pflat is NA
   \nl }

For low-pass spacing the test results may include
 \item{\code{plen}}{ the probability of the feature per the length model }
 \item{\code{hexcur}}{ the flat height used for the excursion test }
 \item{\code{pexcur}}{ fraction of bootstrap samples from the low-pass spacing
   generating smaller heights than hexcur
   \nl }
depending on what has been run.
   
The interval spacing omits the length model test and may perform the excursion
test on the difference of the interval spacing.
 \item{\code{hexcur}}{ the feature height used for the excursion test }
 \item{\code{pexcur}}{ fraction of bootstrap samples from the interval
   spacing that generate smaller heights than hexcur
   \nl }

The methods return the object, invisibly.
}

\seealso{
  \code{\link{find.flats}},
  \code{\link{Dimodal}},
  \code{\link{Diopt}},
  \code{\link{shiftID.place}},
  \code{\link{data.frame}}
}

\examples{
m <- Dimodal(faithful$eruptions, Diopt.local(diw.window=16))
m$lp.flats
summary(m$diw.flats)
}

\keyword{Diflat}
\keyword{Dimodal}
